Package: modelbpp
Title: Model BIC Posterior Probability
Version: 0.3.0
Authors@R: 
    c(person(given = "Shu Fai",
             family = "Cheung",
             role = c("aut", "cre"),
             email = "shufai.cheung@gmail.com",
             comment = c(ORCID = "0000-0002-9871-9448")),
      person(given = "Huiping",
             family = "Wu",
             role = c("aut")),
      person(given = "Shing On",
             family = "Leung",
             role = c("aut"),
             comment = c(ORCID = "0000-0002-0962-339X")),
      person(given = "Ivan Jacob Agaloos",
             family = "Pesigan",
             role = "ctb",
             email = "r.jeksterslab@gmail.com",
             comment = c(ORCID = "0000-0003-4818-8420")))
Description: Fits the neighboring models of a fitted
  structural equation model and assesses the model
  uncertainty of the fitted model based on BIC posterior
  probabilities (BPP), using the method presented in
  Wu, Cheung, and Leung (2020)
  <doi:10.1080/00273171.2019.1574546>. See
  Pesigan, Cheung, Wu, Chang, and Leung (2026)
  <doi:10.3758/s13428-025-02921-x> for an introduction to
  the package.
License: GPL (>= 3)
Encoding: UTF-8
Suggests: knitr, rmarkdown, tinytest
Depends: R (>= 4.0.0)
Imports: lavaan, parallel, pbapply, igraph, manymome
VignetteBuilder: knitr
URL: https://sfcheung.github.io/modelbpp/
BugReports: https://github.com/sfcheung/modelbpp/issues
LazyData: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-12 15:26:03 UTC; shufa
Author: Shu Fai Cheung [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9871-9448>),
  Huiping Wu [aut],
  Shing On Leung [aut] (ORCID: <https://orcid.org/0000-0002-0962-339X>),
  Ivan Jacob Agaloos Pesigan [ctb] (ORCID:
    <https://orcid.org/0000-0003-4818-8420>)
Maintainer: Shu Fai Cheung <shufai.cheung@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-12 23:40:02 UTC
Built: R 4.5.2; ; 2026-05-12 23:50:28 UTC; unix

Package: modelbpp
Title: Model BIC Posterior Probability
Version: 0.3.0
Authors@R: 
    c(person(given = "Shu Fai",
             family = "Cheung",
             role = c("aut", "cre"),
             email = "shufai.cheung@gmail.com",
             comment = c(ORCID = "0000-0002-9871-9448")),
      person(given = "Huiping",
             family = "Wu",
             role = c("aut")),
      person(given = "Shing On",
             family = "Leung",
             role = c("aut"),
             comment = c(ORCID = "0000-0002-0962-339X")),
      person(given = "Ivan Jacob Agaloos",
             family = "Pesigan",
             role = "ctb",
             email = "r.jeksterslab@gmail.com",
             comment = c(ORCID = "0000-0003-4818-8420")))
Description: Fits the neighboring models of a fitted
  structural equation model and assesses the model
  uncertainty of the fitted model based on BIC posterior
  probabilities (BPP), using the method presented in
  Wu, Cheung, and Leung (2020)
  <doi:10.1080/00273171.2019.1574546>. See
  Pesigan, Cheung, Wu, Chang, and Leung (2026)
  <doi:10.3758/s13428-025-02921-x> for an introduction to
  the package.
License: GPL (>= 3)
Encoding: UTF-8
Suggests: knitr, rmarkdown, tinytest
Depends: R (>= 4.0.0)
Imports: lavaan, parallel, pbapply, igraph, manymome
VignetteBuilder: knitr
URL: https://sfcheung.github.io/modelbpp/
BugReports: https://github.com/sfcheung/modelbpp/issues
LazyData: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-12 15:26:03 UTC; shufa
Author: Shu Fai Cheung [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9871-9448>),
  Huiping Wu [aut],
  Shing On Leung [aut] (ORCID: <https://orcid.org/0000-0002-0962-339X>),
  Ivan Jacob Agaloos Pesigan [ctb] (ORCID:
    <https://orcid.org/0000-0003-4818-8420>)
Maintainer: Shu Fai Cheung <shufai.cheung@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-12 23:40:02 UTC
Built: R 4.6.0; ; 2026-05-13 00:21:38 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(c,model_set)
S3method(c,partables)
S3method(plot,model_graph)
S3method(print,model_set)
S3method(print,partables)
S3method(print,sem_outs)
export(fit_many)
export(gen_models)
export(get_add)
export(get_drop)
export(get_partables)
export(identical_partables)
export(is_partable)
export(measurement_invariance_models)
export(min_prior)
export(model_graph)
export(model_set)
export(model_set_combined)
export(partables_drop)
export(same_variables)
export(to_partables)

<!-- badges: start -->
[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![Project Status: Active - The project has reached a stable, usable state and is being actively developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![CRAN status](https://www.r-pkg.org/badges/version/modelbpp?color=blue)](https://CRAN.R-project.org/package=modelbpp)
[![CRAN: Release Date](https://www.r-pkg.org/badges/last-release/modelbpp?color=blue)](https://cran.r-project.org/package=modelbpp)
[![CRAN RStudio mirror downloads](https://cranlogs.r-pkg.org/badges/grand-total/modelbpp?color=blue)](https://r-pkg.org/pkg/modelbpp)
[![Code size](https://img.shields.io/github/languages/code-size/sfcheung/modelbpp.svg)](https://github.com/sfcheung/modelbpp)
[![Last Commit at Main](https://img.shields.io/github/last-commit/sfcheung/modelbpp.svg)](https://github.com/sfcheung/modelbpp/commits/main)
[![R-CMD-check](https://github.com/sfcheung/modelbpp/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/sfcheung/modelbpp/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

# modelbpp: Model BIC Posterior Probability <img src="man/figures/logo.png" align="right" />

(Version 0.3.0 updated on 2026-05-12, [release history](https://sfcheung.github.io/modelbpp/news/index.html))

This package is for assessing model uncertainty in structural
equation modeling (SEM) by the BIC posterior
probabilities of the fitted model and its neighboring models,
based on the method presented in Wu, Cheung, and Leung (2020).
The package name, `modelbpp`,
stands for `model` `b`ayesian `p`osterior `p`robability. An
introduction to the package can be found in the following
article:

- Pesigan, I. J. A., Cheung, S. F., Wu, H., Chang, F., & Leung, S. O. (2026).
  How plausible is my model? Assessing model plausibility of
  structural equation models using Bayesian posterior probabilities (BPP).
  *Behavior Research Methods, 58*(3), Article 73.
  https://doi.org/10.3758/s13428-025-02921-x

# Homepage

For more information on this package, please visit its GitHub page:

[https://sfcheung.github.io/modelbpp/](https://sfcheung.github.io/modelbpp/)

A quick introduction on how to use this package
can be found in the Get-Started article (`vignette("modelbpp")`).

# Installation

The stable CRAN version can be installed by `install.packages()`:

```r
install.packages("modelbpp")
```

The latest developmental-but-stable version of this package can be installed by `remotes::install_github`:

```r
remotes::install_github("sfcheung/modelbpp")
```

# Issues

If you have any suggestions or found any bugs, please feel
free to open a GitHub issue. Thanks.

https://github.com/sfcheung/modelbpp/issues

# Reference(s)

Wu, H., Cheung, S. F., & Leung, S. O. (2020).
Simple use of BIC to assess model selection uncertainty:
An illustration using mediation and moderation models.
*Multivariate Behavioral Research*, *55*(1), 1--16.
https://doi.org/10.1080/00273171.2019.1574546

if (length(unclass(packageVersion("modelbpp"))[[1]]) == 4) {

# WIP

suppressMessages(library(lavaan))

mod <-
"
visual =~ x1 + x2 + x3
textual =~ x4 + x5 + x6
speed =~ x7 + x8 + x9
"
fit <- cfa(mod, HolzingerSwineford1939[HolzingerSwineford1939$school == "Grant-White", ])
suppressWarnings(out_pc <- model_set(fit,
                                     loadings_to_exclude = "none",
                                     parallel = FALSE,
                                     progress = FALSE))
expect_stdout(print(out_pc),
              "failed")
}

#' @title Models That Are More Restricted
#'
#' @description Generate a list of
#' models with one or more free parameter
#' dropped (fixed to zero).
#'
#' @details It generates a list of models
#' with one or more free parameters
#' dropped, that is, fixed to zero
#' (with degrees of freedom,
#' *df*, increases by one or more).
#'
#' All free parameters are included in
#' the pool of candidates, except for
#' those explicitly requested to be
#' kept.
#'
#' The models will be checked by `lavaan`
#' to make sure that the increase in
#' model degrees of freedom is of the
#' expected value.
#'
#' This function is called by
#' [model_set()] and usually users do
#' not need to call it. It is exported
#' for advanced users.
#'
#' @param sem_out The original model,
#' which is the output from an
#' structural equation modeling
#' function. Currently support
#' [lavaan::lavaan-class] objects only.
#'
#' @param must_drop A character vector
#' of parameters, named in
#' `lavaan::lavaan()` style (e.g.,
#' `"y ~ x"`), that must be included.
#' Default is `NULL`.
#'
#' @param must_not_drop A character
#' vector of parameters, named in
#' [lavaan::lavaan()] style (e.g.,
#' `"x1 ~~ x1"`), that must not be
#' included. Default is `NULL`.
#'
#' @param loadings_to_exclude Whether factor loadings
#' will be excluded. If `"single"`,
#' then only "single" loadings (an indicator
#' loads on only one latent factor)
#' will be excluded. If `"all"`, then
#' all factor loadings will be excluded.
#' If `"none"`, then no loadings will
#' be excluded. Be careful when using
#' `"none"` because the models may not
#' make sense. The settings in `must_drop`
#' and `must_not_drop` will override this
#' argument.
#'
#' @param df_change How many degrees
#' of freedom away in the list. All
#' models with *df* change less than
#' or equal to this number will be
#' included, taking into account
#' requirements set by other arguments.
#' Default is 1.
#'
#' @param model_id The identification
#' number of the starting model.
#' Default is `NA`, no identification
#' number.
#'
#' @param keep_correct_df_change Keep
#' only models with actual *df* change
#' equal to expected *df* change.
#'
#' @param remove_duplicated If `TRUE`,
#' the default, duplicated models are
#' removed.
#'
#' @param progress Whether a progress
#' bar will be displayed, implemented
#' by the `pbapply` package. Default
#' is `FALSE`.
#'
#' @return An object of the class
#' `partables`, a named list of parameter
#' tables, each of them to be used by
#' [lavaan::lavaan()] or [update()]
#' for fitting a model with the added
#' parameters.
#'
#' @seealso [print.partables()]
#'
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>
#'
#' @examples
#'
#' library(lavaan)
#'
#' dat <- dat_path_model
#' mod <-
#' "
#' x3 ~ a*x1 + b*x2
#' x4 ~ a*x1 + x2
#' ab := a*b
#' "
#'
#' fit <- sem(mod, dat_path_model, fixed.x = TRUE)
#' mod_to_drop <- get_drop(fit)
#' mod_to_drop
#'
#' @export

get_drop <- function(sem_out,
                     must_drop = NULL,
                     must_not_drop = NULL,
                     loadings_to_exclude = c("single", "none", "all"),
                     df_change = 1,
                     model_id = NA,
                     keep_correct_df_change = TRUE,
                     remove_duplicated = TRUE,
                     progress = FALSE
                    ) {
    if (missing(sem_out)) stop("sem_out is not supplied.")
    if (!inherits(sem_out, "lavaan")) {
        stop("sem_out is not a lavaan-class object.")
      }
    loadings_to_exclude <- match.arg(loadings_to_exclude)
    pt <- lavaan::parameterTable(sem_out)
    # ==== Remove all user-defined parameters unless constrained ====
    pt <- pt_remove_user_defined(pt, remove_constrained = FALSE)
    # ==== Exclude all parameters already constrained to be equal ====
    id_exclude_eq <- pt_remove_constrained_equal(pt, return_id = TRUE)
    # ==== Exclude the variances of exogenous variables ====
    id_exclude_exo_var <- pt_remove_exo_var(pt, return_id = TRUE)
    # ==== Exclude error variances of endogenous variables ====
    id_exclude_end_var <- pt_remove_end_var(pt, return_id = TRUE)
    # ==== Handle loadings ====
    # FALSE: To exclude from the list tot drop
    # TRUE: To include in the list to drop
    if (loadings_to_exclude == "single") {
      id_exclude_loadings <- pt_remove_all_single_loadings(
                        pt,
                        return_id = TRUE)
    }
    if (loadings_to_exclude == "all") {
      id_exclude_loadings <- pt_remove_all_loadings(
                        pt,
                        return_id = TRUE)
    }
    if (loadings_to_exclude == "none") {
      id_exclude_loadings <- rep(TRUE, nrow(pt))
    }
    # ==== Determine the candidate lists ====
    id_to_drop <- pt$free > 0
    id_to_drop <- id_to_drop & id_exclude_eq
    id_to_drop <- id_to_drop & id_exclude_exo_var
    id_to_drop <- id_to_drop & id_exclude_end_var
    id_to_drop <- id_to_drop & id_exclude_loadings
    # ==== must_drop: User specified parameters ====
    if (!is.null(must_drop)) {
        id_must_drop <- syntax_to_id(must_drop, ptable = pt)
        id_to_drop[id_must_drop] <- TRUE
      }
    # ==== must_not_drop: User specified parameters ====
    if (!is.null(must_not_drop)) {
        id_must_not_drop <- syntax_to_id(must_not_drop, ptable = pt)
        id_to_drop[id_must_not_drop] <- FALSE
      }
    # ==== any(id_to_drop) ====
    if (any(id_to_drop)) {
        # ==== Determine the sets of changes ====
        sets_to_gen <- lapply(seq_len(df_change),
                    function(x) {
                              utils::combn(which(id_to_drop), x, simplify = FALSE)
                            }
                  )
        sets_to_gen <- unlist(sets_to_gen, recursive = FALSE)
        df0 <- lavaan::fitMeasures(sem_out, "df")
        if (progress) {
            cat("\nGenerate", length(sets_to_gen), "more restrictive model(s):\n")
            op_old <- pbapply::pboptions(type = "timer")
            out <- pbapply::pblapply(sets_to_gen,
                                     gen_pt_drop,
                                     pt = pt,
                                     to = model_id,
                                     source_df = df0,
                                     sem_out = sem_out)
            pbapply::pboptions(op_old)
          } else {
            out <- lapply(sets_to_gen, gen_pt_drop, pt = pt, to = model_id,
                          source_df = df0, sem_out = sem_out)
          }
      } else {
        out <- list()
      }

    # ==== Keep tables with expected df only? ====
    if (keep_correct_df_change) {
        chk1 <- sapply(out, attr, which = "df_actual")
        chk2 <- sapply(out, attr, which = "df_expected")
        out <- out[chk1 == chk2]
      }

    out_names <- sapply(out, function(x) {
        paste("drop:",
              paste(attr(x, "parameters_dropped"), collapse = ";"))
      })
    names(out) <- out_names
    attr(out, "call") <- match.call()
    attr(out, "sem_out") <- sem_out
    class(out) <- c("partables", class(out))
    if (remove_duplicated) {
        out <- unique_models(out, progress = progress)
      }
    out
  }

#' @noRd

gen_pt_drop <- function(x, pt, to, source_df = NA, sem_out) {
    # Function to generate pt
    for (i in x) {
        pt[i, "free"] <- 0
        pt[i, "ustart"] <- 0
        pt[i, "est"] <- 0
      }
    p_to_drop <- sapply(x, function(x)
        paste0(pt[x, "lhs"], pt[x, "op"], pt[x, "rhs"])
      )
    p_to_drop_out <- lapply(x, function(x) {
        c(lhs = pt[x, "lhs"], op = pt[x, "op"], rhs = pt[x, "rhs"])
      })
    suppressWarnings(sem_out_update <- lavaan::update(sem_out,
                                     pt,
                                     do.fit = TRUE,
                                     optim.force.converged = TRUE,
                                     warn = FALSE,
                                     se = "none",
                                     baseline = FALSE,
                                     check.start = FALSE,
                                     check.post = FALSE,
                                     check.vcov = FALSE,
                                     control = list(max.iter = 1)))
    pt_update <- lavaan::parameterTable(sem_out_update)
    attr(pt_update, "parameters_dropped") <- p_to_drop
    attr(pt_update, "parameters_dropped_list") <- p_to_drop_out
    attr(pt_update, "ids_dropped") <- x
    attr(pt_update, "to") <- to
    attr(pt_update, "df_expected") <- unname(source_df) +
                               length(x)
    attr(pt_update, "df_actual") <- unname(lavaan::fitMeasures(sem_out_update,
                                                        fit.measures = "df"))
    pt_update
  }

suppressMessages(library(lavaan))

dat <- dat_path_model

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)

expect_error(fit_many(progress = FALSE))
expect_error(fit_many(mod, progress = FALSE))
expect_error(get_add())
expect_error(get_add(mod))
expect_error(get_drop())
expect_error(get_drop(mod))
expect_error(model_set(progress = FALSE))
expect_error(model_set(mod, progress = FALSE))

suppressMessages(library(lavaan))

dat <- dat_path_model

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod,
           dat_path_model,
           fixed.x = TRUE)
mod_to_add <- get_add(fit)
mod_to_drop <- get_drop(fit)
mod_to_fit <- c(mod_to_add, mod_to_drop)
out <- fit_many(mod_to_fit,
                fit,
                progress = FALSE)

expect_equivalent(
    out$change,
    c(1, 1, -1, -1)
  )

out <- fit_many(mod_to_fit,
                fit,
                original = "add: x4~x2",
                progress = FALSE)

expect_equivalent(
    out$change,
    c(0, 0, -2, -2)
  )

out <- fit_many(mod_to_fit,
                fit,
                original = NA,
                progress = FALSE)

expect_true(all(is.na(out$change)))

# Test Print

expect_stdout(print(out),
              "drop: x3~~x4",
              fixed = TRUE)
expect_stdout(print(out,
                    max_models = 2),
              "first 2",
              fixed = TRUE)

if (interactive()) {
    mod_to_add <- get_add(fit,
                          df_change = 2)
    mod_to_drop <- get_drop(fit,
                            df_change = 2)
    mod_to_fit <- c(mod_to_add, mod_to_drop)
    out <- fit_many(mod_to_fit,
                    fit,
                    parallel = TRUE,
                    ncores = 2,
                    progress = FALSE,
                    verbose = FALSE)
    expect_equivalent(
        out$change,
        c(1, 1, 2, -1, -1, -2),
        ignore_attr = TRUE
      )
  }

suppressMessages(library(lavaan))

dat <- dat_path_model

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1 + x2
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)
pt <- parameterTable(fit)
pt_no_user <- pt[pt$op != ":=", ]
mod_to_drop <- get_drop(fit)
fit_drop <- lapply(mod_to_drop, function(x) update(fit, x))
anova_drop <- lapply(fit_drop, function(x) anova(fit, x))

expect_true(
    all(names(mod_to_drop) %in% c("drop: x3~x2", "drop: x4~x2", "drop: x3~~x4")),
    info = "Parameters to drop as expected"
  )

expect_true(
    all(sapply(anova_drop, function(x) x["x", "Df diff"]) == 1),
    info = "All df differences are one"
  )

expect_true(
    all(sapply(mod_to_drop,
                modelbpp:::get_diff_drop, pt_no_user) == names(mod_to_drop)),
    info = "Generated difference matches the names"
  )

mod_to_drop <- get_drop(fit,
                        df_change = 2)
fit_drop <- lapply(mod_to_drop,
                   function(x) update(fit, x))
anova_drop <- lapply(fit_drop,
                     function(x) anova(fit, x))

expect_true(
    all(names(mod_to_drop) %in%
      c("drop: x3~x2", "drop: x4~x2", "drop: x3~~x4",
        "drop: x3~x2;x4~x2", "drop: x3~x2;x3~~x4", "drop: x4~x2;x3~~x4")),
    info = "Parameters to drop as expected"
  )

expect_true(
    all(sapply(anova_drop,
               function(x) x["x", "Df diff"]) ==
        c(1, 1, 1, 2, 2, 2)),
    info = "All df differences are one"
  )

expect_true(
    all(sapply(mod_to_drop,
               modelbpp:::get_diff_drop, pt_no_user) == names(mod_to_drop)),
    info = "Generated difference matches the names"
  )

# Test Print

expect_stdout(print(mod_to_drop),
              "`drop: x4~x2;x3~~x4`",
              fixed = TRUE)

suppressMessages(library(lavaan))
mod <-
"
fx1 =~ x1 + x2 + x3
fx2 =~ x4 + x5 + x6
fm1 =~ x7 + x8 + x9
fy2 =~ x10 + x11 + x12
fy3 =~ x13 + x14 + x15 + x11
fm1 ~ fx1 + fx2
fy2 ~ fm1
fy3 ~ fm1
"

fit <- sem(mod,
           dat_sem,
           warn = FALSE)

outa <- get_drop(fit, loadings_to_exclude = "all")
expect_true(!any(grepl("=", names(outa))))

outb <- get_drop(fit, loadings_to_exclude = "single")
expect_false("drop: fx1=~x2" %in% names(outb))
expect_true("drop: fy3=~x11" %in% names(outb))

outc <- get_drop(fit, loadings_to_exclude = "none")
expect_true("drop: fx1=~x2" %in% names(outc))

#' @title Helper Functions For `partables`-Class Objects
#'
#' @description For tasks such as comparing two parameter
#' tables inside a `partables`-class object.
#'
#' @name partables_helpers
NULL

#' @details
#'
#' The function [identical_partables()] compare two
#' `lavaan` parameter tables and see whether they
#' are identical. (Adapted from a similar function
#' in the package `semhelpinghands`).
#'
#' @param object1 A `lavaan` parameter
#' table or similar object.
#'
#' @param object2 A `lavaan` parameter
#' table or similar object.
#'
#' @return
#' The function [identical_partables()]
#' returns either `TRUE` or `FALSE`.
#'
#' @examples
#'
#' library(lavaan)
#' mod1 <-
#' "
#' x3 ~ x1
#' x2 ~ x4
#' "
#' mod2 <-
#' "
#' x2 ~ x4
#' x3 ~ x1
#' "
#' fit1 <- sem(mod1, dat_path_model)
#' fit2 <- sem(mod2, dat_path_model)
#' pt1 <- parameterTable(fit1)
#' pt2 <- parameterTable(fit2)
#' identical_partables(pt1, pt2)
#'
#' @rdname partables_helpers
#' @export

identical_partables <- function(object1,
                                object2) {
    if (!is_partable(object1) ||
        !is_partable(object2)) {
        stop("At least one of the objects is not a parameter table.")
      }
    if (!identical(nrow(object1),
                   nrow(object2))) {
        return(FALSE)
      }
    if (!identical(sum(object1$free > 0),
                   sum(object2$free > 0))) {
        return(FALSE)
      }
    if (!identical(sort(object1$op),
                   sort(object2$op))) {
        return(FALSE)
      }
    if (!identical(sort(union(object1$lhs, object1$rhs)),
                   sort(union(object2$lhs, object2$rhs)))) {
        return(FALSE)
      }
    # Drop nonessential columns
    object1$est <- NULL
    object2$est <- NULL
    object1$se <- NULL
    object2$se <- NULL
    pt1 <- fix_cov(object1)
    pt2 <- fix_cov(object2)

    tmp1 <- pt1[order(pt1$group,
                      pt1$op, pt1$lhs, pt1$rhs), ]
    tmp2 <- pt2[order(pt2$group,
                      pt2$op, pt2$lhs, pt2$rhs), ]
    rownames(tmp1) <- NULL
    rownames(tmp2) <- NULL
    if (!identical(tmp1[, c("group", "lhs", "op", "rhs")],
                   tmp2[, c("group", "lhs", "op", "rhs")])) {
        return(FALSE)
      }
    if (!identical(tmp1$free > 0,
                   tmp2$free > 0)) {
        return(FALSE)
      }
    if (!is.null(tmp1$label) &&
        !is.null(tmp2$label)) {
        if (!identical(tmp1$label,
                       tmp2$label)) {
            return(FALSE)
          }
      }
    if (!is.null(tmp1$exo) &&
        !is.null(tmp2$exo)) {
        if (!identical(tmp1$exo,
                       tmp2$exo)) {
            return(FALSE)
          }
      }
    if (!is.null(tmp1$ustart) &&
        !is.null(tmp2$ustart)) {
        tmp1u <- tmp1[which((tmp1$ustart > 0) & (tmp1$free == 0)), ]
        tmp2u <- tmp2[which((tmp2$ustart > 0) & (tmp2$free == 0)), ]
        if (!identical(tmp1u$start,
                       tmp2u$start)) {
            return(FALSE)
          }
      }
    return(TRUE)
  }

#' @noRd

check_lor <- function(object1, object2) {

  }

#' @param x An object to be checked.
#'
#' @details
#' The function [is_partable()] tries
#' to
#' check whether an object is "likely"
#' to be a parameter table that can be
#' used by [lavaan::lavaan()] and its
#' wrappers, such as [lavaan::sem()].
#'
#' Note that there is no guarantee the
#' the parameter table makes sense or
#' will not lead to error when fitted.
#' It can only check whether it has the
#' required columns.
#'
#' @return
#' The function [is_partable()]
#' returns either `TRUE` or `FALSE`.
#'
#' @examples
#'
#' is_partable(pt1)
#'
#' @rdname partables_helpers
#' @export

is_partable <- function(x) {
    if (!inherits(x, what = "data.frame")) {
        # stop("x is not a data-frame-like object.")
        return(FALSE)
      }
    # Must at least have these columns
    min_names <- c("id",
                   "lhs",
                   "op",
                   "rhs",
                   "free",
                   "ustart")
    if (!all(min_names %in% colnames(x))) {
        # stop("x does not have minimally required columns.")
        return(FALSE)
      }
    return(TRUE)
  }

#' @details
#' The function [same_variables()]
#' check whether all parameter labels
#' in a `partables`-class object use
#' the same observed variables.
#'
#' @return
#' The function [same_variables()]
#' returns either `TRUE` or `FALSE`.
#'
#' @examples
#'
#' out <- model_set(fit1,
#'                  fit_models = FALSE)
#' same_variables(get_partables(out))
#'
#'
#' @rdname partables_helpers
#' @export

same_variables <- function(x) {
    if (length(x) == 1) {
        return(TRUE)
      }
    out0 <- lapply(x, function(xx) {
                out <- tryCatch(lavaan::lavaan(xx))
                out
              })
    if (any(sapply(out0, inherits, what = "error"))) {
        stop("One or more parameter tables result in error.")
      }
    all_names <- lapply(out0,
                        lavaan::lavNames,
                        "ov")
    all_names0 <- all_names[[1]]
    for (xx in all_names[-1]) {
        if (!setequal(all_names0, xx)) {
            return(FALSE)
          }
      }
    return(TRUE)
  }


#' @details
#' The function [get_partables()]
#' extract the `partable` object from
#' a `model_set`-class object.
#'
#' @return
#' The function [get_partables()]
#' returns a `partables`-class
#' object.
#'
#' @examples
#'
#' out <- model_set(fit1,
#'                  fit_models = FALSE)
#' get_partables(out)
#'
#' @rdname partables_helpers
#' @export

get_partables <- function(x) {
    out <- x$models
    if (!inherits(out, what = "partables")) {
        return(NULL)
      }
    out
  }


#' @details
#' The function [to_partables()]
#' combine objects to create a
#' `partables`-class object.
#' The objects to be combined can be
#' a `lavaan`-class object (e.g.,
#' the output of [lavaan::sem()])
#' or a parameter table of `lavaan`.
#'
#' @return
#' The function [to_partables()]
#' returns a `partables`-class
#' object, created from the objects
#' supplied.
#'
#' @param ... The objects to be combined.
#'
#' @examples
#'
#' fit1 <- sem(mod1, dat_path_model)
#' fit2 <- sem(mod2, dat_path_model)
#' pt1 <- parameterTable(fit1)
#' pt2 <- parameterTable(fit2)
#'
#' to_partables(fit1, fit2)
#' to_partables(pt1, pt2)
#'
#' @rdname partables_helpers
#' @export

to_partables <- function(...) {
    ddd <- list(...)
    ddd_names <- sapply(as.list(match.call()[-1]), deparse)
    if (!is.null(names(ddd_names))) {
        i <- which(sapply(names(ddd_names), function(x) {nchar(x) > 0}))
        if (length(i) > 0) {
            ddd_names[i] <- names(ddd_names)[i]
            ddd_names <- unname(ddd_names)
          }
      }
    out <- lapply(ddd,
                  function(x) {
                      if (inherits(x, "lavaan")) {
                          return(lavaan::parameterTable(x))
                        }
                      if (is_partable(x)) {
                          return(x)
                        }
                      stop("At least one object is neither a lavaan object ",
                           "nor a parameter table.")
                    })
    names(out) <- ddd_names
    class(out) <- c("partables", class(out))
    out
  }


#' @noRd

fix_cov <- function(x) {
    i <- which(x$op == "~~")
    if (length(i) == 0) {
        return(x)
      }
    jj <- x[i, "rhs"] > x[i, "lhs"]
    ii <- i[jj]
    a <- x[ii, "lhs"]
    x[ii, "lhs"] <- x[ii, "rhs"]
    x[ii, "rhs"] <- a
    return(x)
  }

suppressMessages(library(lavaan))

# Test user models

if (interactive() &&
    length(unclass(packageVersion("modelbpp"))[[1]]) == 4) {

# Slow tests

dat <- dat_path_model

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)

mod2 <-
"
x2 ~ x1
x3 ~ x2
x4 ~ x3
"

mod3 <-
"
x2 ~ x4
x3 ~ x4
x1 ~ x2 + x3
"

fit2 <- sem(mod2, dat_path_model)
fit3 <- sem(mod3, dat_path_model)

out <- model_set(fit,
                 progress = FALSE)

pt_user <- c(out$models,
             user2 = parameterTable(fit2),
             user3 = fit3)

# pt_user[["user2"]] <-
# pt_user[["user3"]] <- parameterTable(fit3)
# pt_user

out_user <- model_set(sem_out = fit,
                      partables = pt_user,
                      progress = FALSE,
                      parallel = FALSE)
out_user_prior <- model_set(sem_out = fit,
                            partables = pt_user,
                            prior_sem_out = .50,
                            progress = FALSE,
                            parallel = FALSE)

expect_true(all(c("user2", "user3") %in% names(out_user_prior$models)),
            info = "Add user models")

out_user_prior2 <- model_set(sem_out = fit,
                             partables = pt_user,
                             prior_sem_out = c(original = .30,
                                               user2 = .20,
                                               user3 = .10),
                             progress = FALSE,
                             parallel = FALSE)
tmp <- out_user_prior2$prior
names(tmp) <- names(out_user_prior2$models)
expect_equal(unname(tmp[c("user2", "original", "user3")]),
             c(.20, .30, .10),
             info = "Add user models, with priors")

out_user_prior3 <- model_set(sem_out = fit,
                             partables = pt_user,
                             prior_sem_out = c(original = .31,
                                               `add: x4~x2` = .21,
                                               user2 = .10),
                             progress = FALSE,
                             parallel = FALSE)

tmp <- out_user_prior3$prior
names(tmp) <- names(out_user_prior3$models)
expect_equal(unname(tmp[c("add: x4~x2", "original")]),
             c(.21, .31),
             info = "Add user models, with priors and nonstandard names")

expect_stdout(print(out_user_prior3,
                    bpp_target = .95,
                    target_name = "add: x4~x2"),
              pattern = "Target Model: add: x4~x2",
              info = "Add user models, print method")

# User models with graph

suppressMessages(library(igraph))

mod <-
"
x2 ~ x3 + 0*x4
x1 ~ x3
"

fit <- sem(mod,
           dat_path_model,
           fixed.x = FALSE)
out <- model_set(fit,
                 fit_models = FALSE,
                 progress = FALSE,
                 parallel = FALSE)

mod2 <-
"
x2 ~ x1
x3 ~ x2
x4 ~ x3
"

mod3 <-
"
x2 ~ x4
x3 ~ x4
x1 ~ x2 + x3
"

fit2 <- sem(mod2, dat_path_model, fixed.x = FALSE)
fit3 <- sem(mod3, dat_path_model, fixed.x = FALSE)

mod4 <-
"
x2 ~ 0*x3 + 0*x4
x1 ~ 0*x3
"
fit4 <- sem(mod4, dat_path_model)

pt_user <- get_partables(out)

pt_user <- c(pt_user,
             user2 = fit2,
             user3 = fit3,
             user4 = fit4)

# pt_user[["user2"]] <- parameterTable(fit2)
# pt_user[["user3"]] <- parameterTable(fit3)
# pt_user[["user4"]] <- parameterTable(fit4)

out_user_prior5 <- model_set(sem_out = fit,
                             partables = pt_user,
                             prior_sem_out = c(original = .11,
                                               user2 = .50),
                             progress = FALSE,
                             parallel = FALSE)
g <- model_graph(out_user_prior5,
                 progress = FALSE)
# expect_warning(g <- model_graph(out_user_prior5),
#                pattern = "One or more")
expect_true(all(c("user2", "user4", "user3") %in%
                names(V(g))))
if (interactive()) {
    plot(g)
  }
}

suppressMessages(library(lavaan))

data(dat_path_model)
set.seed(8970431)
dat_path_model$group <- sample(c("Gp1", "Gp2"),
                               nrow(dat_path_model),
                               replace = TRUE)
dat_path_model2 <- dat_path_model
dat_path_model2[1, 2] <- NA

mod <-
"
x3 ~ x1 + x2
x4 ~ x1
"

fit1 <- sem(mod, dat_path_model)
pt1 <- parameterTable(fit1)
pt0 <- parameterTable(sem(mod))
pt2 <- lavaanify(mod)
pt3 <- pt1[c("id", "lhs", "op", "rhs",
             "free", "ustart")]
fit3 <- sem(pt3, dat_path_model)
coef(fit3)
coef(fit1)

expect_error(identical_partables(pt1, as.list(pt3)))
expect_error(identical_partables(pt3[1:3], pt1))

modb <-
"
x3 ~ x2 + x1
x4 ~ x1
"
fitb1 <- sem(modb, dat_path_model)

ptb1 <- parameterTable(fitb1)
ptb1
pt1

expect_true(identical_partables(pt1, ptb1))
expect_true(identical_partables(pt1, pt0))
expect_false(identical_partables(pt1, pt2))

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)
out <- model_set(fit,
                 fit_models = FALSE,
                 parallel = FALSE,
                 progress = FALSE)
pts <- out$models

modb <-
"
x3 ~ x1
x2 ~ a*x1
"

fitb <- sem(modb, dat_path_model, fixed.x = TRUE)
ptsb <- c(pts, parameterTable(fitb))

expect_true(same_variables(pts))
expect_false(same_variables(ptsb))

# Test to_partables

tmp <- to_partables(fitb, pt1, model3 = fit)
expect_equal(names(tmp), c("fitb", "pt1", "model3"))
expect_identical(tmp[[2]], pt1)

#' @title Print a `partables`-Class Object
#'
#' @description Print the content of
#' a `partables`-class object.
#'
#' @details
#' The print method for the output
#' of [gen_models()], [get_add()],
#'  and [get_drop()].
#'
#' @return
#' `x` is returned invisibly.
#' Called for its side effect.
#'
#' @param x A `partables`-class object.
#'
#' @param max_tables The maximum number
#' of models to be printed.
#' Default is 10.
#'
#' @param ...  Optional arguments.
#' Ignored.
#'
#'
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>
#'
#' @seealso [gen_models()], [get_add()],
#'  and [get_drop()].
#'
#' @examples
#'
#' library(lavaan)
#' dat <- dat_path_model
#' mod <-
#' "
#' x3 ~ a*x1 + b*x2
#' x4 ~ a*x1
#' ab := a*b
#' "
#' fit <- sem(mod, dat_path_model, fixed.x = TRUE)
#' mod_to_add <- get_add(fit)
#' mod_to_add
#' print(mod_to_add, max_tables = 1)
#' mod_to_drop <- get_drop(fit)
#' mod_to_drop
#' print(mod_to_drop, max_tables = 1)
#'
#' @export

print.partables <- function(x,
                            max_tables = 10,
                            ...) {
    x_n <- length(x)
    x_names <- names(x)
    x_1 <- x[[1]]
    x_call <- attr(x, "call")
    cat("\n")
    cat("Call:\n")
    print(x_call)
    cat("\n")
    cat("Number of parameter tables: ", x_n, "\n", sep = "")
    cat("\n")
    if (x_n > max_tables) {
        x_tmp <- x_names[seq_len(max_tables)]
        cat("The first", max_tables, "modification(s):\n")
      } else {
        x_tmp <- x_names
        cat("The modifications/models:\n")
      }
    cat(paste(x_tmp, collapse = "\n"), "\n")
    invisible(x)
  }

#' @noRd

unique_models <- function(partables,
                          original = "original",
                          progress = FALSE) {
    i_added <- added(partables)
    i_dropped <- dropped(partables)
    i_added <- lapply(i_added, sort)
    i_dropped <- lapply(i_dropped, sort)
    i_comb <- mapply(function(x, y) {list(x, y)},
                     x = i_added,
                     y = i_dropped,
                     SIMPLIFY = FALSE)
    i_null <- sapply(i_comb,
                      function(x) {
                          is.null(x[[1]]) && is.null(x[[2]])
                        }
                     )
    if (any(i_null)) {
        i_comb2 <- which(i_null)
      } else {
        i_comb2 <- NULL
      }
    if (!is.null(i_comb2)) {
        i_comb2 <- i_comb2[!duplicated(names(i_comb2))]
      }
    j <- !duplicated(i_comb)
    if (any(j)) {
        j_comb2 <- which(j)
      } else {
        j_comb2 <- NULL
      }
    out <- partables[unique(c(j_comb2, i_comb2))]
    out <- duplicated_by_pt(out,
                            original = original,
                            progress = progress)
    out
  }

#' @noRd

duplicated_by_pt <- function(partables_list,
                             original = "original",
                             progress = FALSE) {
    p <- length(partables_list)
    if (p <= 1) {
        return(partables_list)
      }
    p_names <- names(partables_list)
    chk <- rep(FALSE, p)
    if (progress) {
        pstar <- p * (p - 1) / 2
        cat("\nCheck for duplicated models (",
             p,
             " model[s] to check):\n",
             sep = "")
        pb <- utils::txtProgressBar(min = 0,
                                    max = pstar,
                                    width = 50,
                                    char = "+",
                                    style = 3)
      }
    k <- 0
    for (i in seq_len(p - 1)) {
        for (j in seq(from = i + 1, to = p)) {
            k <- k + 1
            if (progress) {
                utils::setTxtProgressBar(pb,
                                         value = k)
              }
            if (identical_partables(partables_list[[i]],
                                    partables_list[[j]])) {
                if (p_names[j] == original) {
                    chk[i] <- TRUE
                  } else {
                    chk[j] <- TRUE
                  }
              }
          }
      }
    if (progress) {
        close(pb)
      }
    out <- partables_list[!chk]
    out
  }

#' @noRd

added <- function(partables) {
    i1 <- sapply(partables, attr,
                 which = "parameters_added")

    i2 <- sapply(partables, attr,
                 which = "constraints_released")
    # i2 <- sapply(i2, function(x) {
    #           if (length(x) > 0) {
    #               strsplit(x[[1]],
    #                        split = ";",
    #                        fixed = TRUE)[[1]]
    #             } else {
    #               return(NULL)
    #             }
    #         })
    i <- mapply(c, i1, i2)
    i <- lapply(i, unlist)
    i
  }

#' @noRd

dropped <- function(partables) {
    i <- sapply(partables, attr,
                which = "parameters_dropped")
    # i <- sapply(i, function(x) {
    #           if (length(x) > 0) {
    #               strsplit(x[[1]],
    #                        split = ";",
    #                        fixed = TRUE)[[1]]
    #             } else {
    #               return(NULL)
    #             }
    #         })
    i <- lapply(i, unlist)
    i
  }

#' @title Generate a Graph of Models
#'
#' @description Generate an 'igraph' object from a
#' 'model_set' object.
#'
#' @details
#' It extracts the model list stored
#' in `object`, creates an adjacency
#' matrix, and then creates an `igraph`
#' object customized for visualizing
#' model relations.
#'
#' ## Construction of the Graph
#'
#' This is the default way to construct
#' the graph when the model set is
#' automatically by [model_set()].
#'
#' - Each model is connected by an
#'  arrow, pointing from one model to
#'  another model that
#'
#'    a. can be formed by adding one
#'       or more free parameter, or
#'
#'    b. can be formed by releasing one
#'       or more equality constraint
#'       between two
#'       parameters.
#'
#'    c. has nested relation with this
#'       model as determined by the
#'       method proposed by Bentler
#'       and Satorra (2010), if the
#'       models are not generated
#'       internally.
#'
#'  That is, it points to a model with
#'  more degrees of freedom (more
#'  complicated),
#'  and
#'  is nested within that model in either
#'  parameter sense or covariance sense.
#'
#' - By default, the size of the node
#'  for each model is scaled by its
#'  BIC posterior probability, if
#'  available. See *The Size of a Node*
#'  below.
#'
#' - If a model is designated as the
#'  original (target) model,
#'  than he original model, the models
#'  with
#'  more degrees of freedom than the
#'  original model,
#'  and the models with fewer degrees
#'  of freedom than the original models,
#'  are colored differently.
#'
#' - The default layout is the Sugiyama
#'  layout, with simpler models (models
#'  with fewer degrees of freedom) on
#'  the top. The lower a model is in
#'  the network, the more the degrees
#'  of freedom it has. This layout is
#'  suitable for showing the nested
#'  relations of the models. Models on
#'  the same level (layer) horizontally
#'  have the same model *df*.
#'
#' The output is an `igraph` object.
#' Users can customize it in any way
#' they want using functions from
#' the `igraph` package.
#'
#' If a model has no nested relation
#' with all other model, it will not
#' be connected to other models.
#'
#' If no model is named `original`
#' (default is `"original"`), then no
#' model is colored as the original model.
#'
#' ## User-Provided Models
#'
#' If `object` contained one or more
#' user-provided models which are not
#' generated automatically by
#' [model_set()] or similar functions
#' (e.g., [gen_models()]), then the
#' method by Bentler and Satorra (2010)
#' will be used to determine model
#' relations. Models connected by an
#' arrow has a nested relation based on
#' the NET method by Bentler and Satorra
#' (2010). An internal function inspired
#' by the `net` function from the
#' `semTools` package is used to
#' implement the NET method.
#'
#' ## The Size of a Node
#'
#' When a model is scaled by `x`,
#' which usually is the BIC posterior
#' probability, its size is determined
#' by:
#'
#' `max_size * (x - min(x))/(max(x) - min(x)) + min_size`
#'
#' @return
#' A `model_graph`-class object that
#' can be used as as an `igraph`-object,
#' with a plot method ([plot.model_graph()])
#' with settings
#' suitable for plotting a network
#' of models with BIC posterior probabilities
#' computed.
#'
#' @param object Must be a
#' `model_set`-class object for now.
#'
#' @param node_size_by_x Logical. Whether
#' node (vertex) sizes are determined
#' by a variable. Default is `TRUE`.
#' See `x` below on how size is
#' determined.
#'
#' @param x If not `NULL`, it should be
#' a numeric vector of length equal to
#' the number of models. The node sizes
#' will be proportional to the values
#' of `x`, offset by `min_size`. If
#' `NULL`, the default, the BIC
#' posterior probabilities
#' stored in `object` will be retrieved.
#'
#' @param node_size If `node_size_by_x`
#' is `FALSE`, this is the size for
#' all nodes.
#'
#' @param min_size The minimum size
#' of a node. Default is 5.
#'
#' @param max_size The maximum size
#' of a node. Default is 35.
#'
#' @param color_original The color
#' of node of the original model.
#' Default is `"lightblue"`.
#'
#' @param color_add The color of
#' the nodes of models formed by
#' adding one or more free parameters to
#' the original model.
#' Default is `"burlywood1"`.
#'
#' @param color_drop The color of
#' the nodes of models formed by
#' dropping one or more free parameters
#' from the original model.
#' Default is `"lightgreen"`.
#'
#' @param color_others The color
#' of other models not specified above.
#' Default is `"grey50"`.
#'
#' @param color_label The color of the
#' text labels of the nodes. Default
#' is `"black"`.
#'
#' @param node_label_size The size of
#' the labels of the nodes. Default is
#' 1.
#'
#' @param original String. The name
#' of the original model (target model).
#' Default is `"original"`.
#'
#' @param drop_redundant_direct_paths
#' Logical. Whether the redundant direct
#' path between two models. A direct path
#' is redundant if two models are also
#' connected through at least one
#' another model. Default is `TRUE`.
#'
#' @param label_arrow_by_df If `TRUE`,
#' then an arrow (edge) is always labelled
#' by the difference in model *df*s.
#' If `FALSE`, then no arrows are
#' labelled. If `NULL`, then arrows are
#' labelled when not all differences
#' in model *df*s are equal to one.
#' Default is `NULL`.
#'
#' @param arrow_label_size The size of
#' the labels of the arrows (edges), if labelled.
#' Default is 1.
#'
#' @param weight_arrows_by_df String.
#' Use if model *df*
#' differences are stored.
#' If `"inverse"`, larger the
#' difference in model *df*, *narrower*
#' an arrow. That is, more similar two
#' models are, thicker the arrow. If
#' `"normal"`, larger the difference
#' in model *df*, *wider* an arrow.
#' If `"none"`, then arrow width is
#' constant, set to `arrow_max_width`.
#' Default is `"inverse"`.
#'
#' @param arrow_min_width If
#' `weight_arrows_by_df` is not `"none"`,
#' this is the minimum width of an
#' arrow.
#'
#' @param arrow_max_width If
#' `weight_arrows_by_df` is not `"none"`,
#' this is the maximum width of an
#' arrow. If `weight_arrows_by_df` is
#' `"none"`, this is the width of all
#' arrows.
#'
#' @param progress Whether a progress
#' bar will be displayed for some
#' steps (e.g., checking for nested
#' relations). Default
#' is `TRUE`.
#'
#' @param short_names If `TRUE` and
#' short model names are stored,
#' they will be used as model labels.
#' Please print the object with
#' `short_names = TRUE` to find the
#' corresponding full model names.
#'
#' @param min_bpp_labelled If not `NULL`,
#' this is the minimum BPP for a model
#' to be labelled. Models with BPP less
#' than this value will not be labelled.
#' Useful when the number of models
#' is large.
#'
#' @param ... Optional arguments. Not
#' used for now.
#'
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>
#' The internal function for nesting
#' inspired by the `net` function
#' from the `semTools` package,
#' which was developed by
#' Terrence D. Jorgensen.
#'
#' @references
#' Bentler, P. M., & Satorra, A. (2010).
#' Testing model nesting and equivalence.
#' *Psychological Methods, 15*(2),
#' 111--123. \doi{10.1037/a0019625}
#' Asparouhov, T., & Muthén, B. (2019).
#' Nesting and Equivalence Testing for
#' Structural Equation Models.
#' *Structural Equation Modeling: A Multidisciplinary Journal, 26*(2),
#' 302--309. \doi{10.1080/10705511.2018.1513795}
#'
#' @examples
#'
#' library(lavaan)
#'
#' mod <-
#' "
#' m1 ~ x
#' y ~ m1
#' "
#'
#' fit <- sem(mod, dat_serial_4, fixed.x = TRUE)
#'
#' out <- model_set(fit)
#' out
#'
#' g <- model_graph(out)
#' plot(g)
#'
#' @export

model_graph <- function(object,
                        node_size_by_x = TRUE,
                        x = NULL,
                        node_size = 5,
                        min_size = 5,
                        max_size = 35,
                        color_original = "lightblue",
                        color_add = "burlywood1",
                        color_drop = "lightgreen",
                        color_others = "lightgrey",
                        color_label = "black",
                        node_label_size = 1,
                        original = "original",
                        drop_redundant_direct_paths = TRUE,
                        label_arrow_by_df = NULL,
                        arrow_label_size = 1,
                        weight_arrows_by_df = c("inverse", "normal", "none"),
                        arrow_min_width = .5,
                        arrow_max_width = 2,
                        progress = TRUE,
                        short_names = FALSE,
                        min_bpp_labelled = NULL,
                        ...) {
    if (!all(object$converged)) {
        stop("Not all models converged.")
      }
    user_models <- sapply(added(object$models), is.null) &
                   sapply(dropped(object$models), is.null)
    # TODO: Should always use models_network2()?
    if (sum(user_models, na.rm = TRUE) != 1) {
        # warning("One or more user models are present. ",
        #     "User model(s) will be plotted separately.")
        net_out <- models_network2(object,
                                   one_df_only = FALSE,
                                   progress = progress)
      } else {
        net_out <- models_network(object)
      }
    out <- igraph::graph_from_adjacency_matrix(net_out,
                                               mode = "directed",
                                               weight = "df")
    if (drop_redundant_direct_paths) {
        out <- delete_all_redundant_direct(out)
      }
    # Store Basic Info
    out <- doc_model_graph(object,
                           out)
    if (node_size_by_x) {
        if (is.null(x)) {
            if (inherits(object, "model_set")) {
                x <- object$bpp
              } else {
                stop("node_size_by_prob set to TRUE but",
                     "x cannot be extracted or not supplied.")
              }
          }
        node_size <- normalize_size(x,
                                    min_size = min_size,
                                    max_size = max_size)
      } else {
        node_size <- node_size
      }
    igraph::V(out)$size <- node_size
    p <- ncol(net_out)
    m_names <- colnames(net_out)
    i_original <- which(m_names == original)
    i_add <- grepl("^add: ", m_names)
    i_drop <- grepl("^drop: ", m_names)
    color_tmp <- rep(color_others, p)
    if (length(i_original) > 0) {
        color_tmp[i_original] <- color_original
      }
    color_tmp[i_add] <- color_add
    color_tmp[i_drop] <- color_drop
    igraph::V(out)$color <- color_tmp
    igraph::V(out)$frame.color <- "grey75"
    igraph::V(out)$label <- v_labels(m_names)
    igraph::V(out)$label.color <- color_label
    igraph::V(out)$label.cex <- node_label_size
    igraph::V(out)$label.family <- "sans"
    igraph::E(out)$arrow.size <- .75
    out <- label_by_df(out,
                       mode = label_arrow_by_df)
    if (!is.null(igraph::E(out)$df)) {
        weight_arrows_by_df <- match.arg(weight_arrows_by_df)
        out <- edge_weight(out,
                           mode = weight_arrows_by_df,
                           min_width = arrow_min_width,
                           max_width = arrow_max_width)
      }
    igraph::E(out)$label.cex <- arrow_label_size
    out <- igraph::add_layout_(out, igraph::with_sugiyama())
    out <- layer_by_df(out,
                       model_set_out = object)
    if (short_names) {
        if (is.character(object$short_names)) {
            tmp1 <- object$short_names
            tmp2 <- igraph::V(out)$name
            igraph::V(out)$full_name <- tmp2
            igraph::V(out)$label <- tmp1[tmp2]
          }
      }
    if (is.numeric(min_bpp_labelled)) {
        tmp2 <- igraph::V(out)$label
        tmp3 <- igraph::V(out)$name
        tmp1 <- (object$bpp < min_bpp_labelled)
        tmp1 <- tmp1[tmp3]
        tmp4 <- tmp2
        tmp4[tmp1] <- ""
        igraph::V(out)$label <- tmp4
      }
    class(out) <- c("model_graph", class(out))
    out
  }

#' @title Plot a Network of Models
#'
#' @description Plot a network of models
#' generated by `model_graph()`.
#'
#' @details This function is the plot
#' method of `model_graph` objects,
#' the output of
#' [model_graph()].
#'
#' For now, it simply passes the object
#' to `plot`-method of an `igraph` object.
#' This function
#' is created for possible customization
#' of the plot in the future.
#'
#' @param x The output of
#' [model_graph()]. (Named `x`
#' because it is required in the naming
#' of arguments of the `plot` generic
#' function.)
#'
#' @param ... Additional arguments,
#' passed to the `plot`-method of
#' an `igraph` object.
#'
#' @seealso [model_graph()]
#'
#' @return `NULL`. Called for its
#' side effect.
#'
#' @examples
#'
#' library(lavaan)
#'
#' dat <- dat_path_model
#'
#' mod <-
#' "
#' x3 ~ a*x1 + b*x2
#' x4 ~ a*x1
#' ab := a*b
#' "
#'
#' fit <- sem(mod, dat_path_model, fixed.x = TRUE)
#'
#' out <- model_set(fit)
#' out
#'
#' g <- model_graph(out)
#' plot(g)
#'
#' @export

plot.model_graph <- function(x,
                             ...) {
    NextMethod()
  }

suppressMessages(library(lavaan))
suppressMessages(library(igraph))

mod1 <-
"
m1 ~ x
m2 ~ m1
y ~ m2
"
fit1 <- sem(mod1, dat_serial_4_weak, fixed.x = FALSE)

mod2 <-
"
m1 ~ x
m2 ~ m1 + x
y ~ m2 + x
"
fit2 <- sem(mod2, dat_serial_4_weak, fixed.x = FALSE)

mod3 <-
"
m1 ~ x
m2 ~ m1 + x
y ~ m2 + x + m1
"
fit3 <- sem(mod3, dat_serial_4_weak, fixed.x = FALSE)

mod4 <-
"
m1 ~ x
m2 ~ m1 + x
y ~ m2 + m1
"
fit4 <- sem(mod4, dat_serial_4_weak, fixed.x = FALSE)

mod5 <-
"
m1 ~ x
m2 ~ m1
y ~ m2 + m1
"
fit5 <- sem(mod5, dat_serial_4_weak, fixed.x = FALSE)


fits <- list(fit1 = fit1,
             fit2 = fit2,
             fit3 = fit3,
             fit4 = fit4,
             fit5 = fit5)

out <- model_set(fits,
                 progress = FALSE,
                 prior_sem_out = c(fit5 = .05,
                                   fit1 = .05,
                                   fit3 = .40,
                                   fit2 = .20,
                                   fit4 = .28))
out
g <- model_graph(out,
                 progress = FALSE)
g2 <- model_graph(out,
                  label_arrow_by_df = FALSE,
                  progress = FALSE)
if (interactive()) {
plot(g)
plot(g2)
}

expect_false(is.null(E(g)$label))
expect_true(is.null(E(g2)$label))

g2 <- model_graph(out,
                  drop_redundant_direct_paths = FALSE,
                  progress = FALSE)
if (interactive()) {
plot(g2)
}

expect_false(are_adjacent(g, "fit1", "fit3"))
expect_true(are_adjacent(g2, "fit1", "fit3"))

suppressMessages(library(lavaan))
suppressMessages(library(igraph))

dat <- dat_path_model

mod1 <-
"
x3 ~ x1 + x2
x4 ~ x3
"
fit1 <- sem(mod1, dat, fixed.x = FALSE)

mod2 <-
"
x2 ~ x1
x3 ~ x2
x4 ~ x3
"
fit2 <- sem(mod2, dat, fixed.x = FALSE)

mod3 <-
"
x2 ~ x4
x3 ~ x4
x1 ~ x2 + x3
"
fit3 <- sem(mod3, dat, fixed.x = FALSE)

mod4 <-
"
x2 ~ x4
x3 ~ x4
x1 ~ x2 + x3 + x4
"
fit4 <- sem(mod4, dat, fixed.x = FALSE)

out <- model_set(sem_out = list(fit1 = fit1,
                                fit3 = fit3,
                                fit2 = fit2,
                                fit4 = fit4),
                 prior_sem_out = c(fit1 = .00005,
                                   fit2 = .60,
                                   fit4 = .20),
                 progress = FALSE,
                 parallel = FALSE)

fit_names <- c("fit1", "fit3", "fit2", "fit4")

expect_true(all(is.na(names(out$change))))
expect_equal(names(out$fit), fit_names)
expect_equal(names(out$models), fit_names)
g <- model_graph(out,
                 progress = FALSE)
expect_equal(V(g)$name, fit_names)

if (length(unclass(packageVersion("modelbpp"))[[1]]) == 4) {

suppressMessages(library(lavaan))

mod <-
"
visual =~ x1 + x2 + x3
textual =~ x4 + x5 + x6
speed =~ x7 + x8 + x9
"
fit <- cfa(mod, HolzingerSwineford1939[HolzingerSwineford1939$school == "Pasteur", ])
suppressWarnings(out_nc <- model_set(fit,
                    loadings_to_exclude = "none",
                    parallel = FALSE,
                    progress = FALSE))
expect_true(all(is.na(out_nc$bpp)))
expect_stdout(print(out_nc),
              "not converged")
expect_error(model_graph(out_nc))
}

suppressMessages(library(lavaan))
suppressMessages(library(igraph))

mod1 <-
"
m1 ~ x
m2 ~ m1
y ~ m2 + x
"
fit1 <- sem(mod1, dat_serial_4_weak, fixed.x = FALSE)

mod2 <-
"
x ~ m1 + m2
y ~ x
"
fit2 <- sem(mod2, dat_serial_4_weak, fixed.x = FALSE)

mod3 <-
"
x ~ m1 + m2
y ~ x + m2
"
fit3 <- sem(mod3, dat_serial_4_weak, fixed.x = FALSE)

mod4 <-
"
m1 ~ y
m2 ~ m1
x ~ m2 + y + m1
"
fit4 <- sem(mod4, dat_serial_4_weak, fixed.x = FALSE)

fits <- list(original = fit1,
             fit2 = fit2,
             fit3 = fit3,
             fit4 = fit4)
pts <- lapply(fits, parameterTable)
class(pts) <- c("partables", class(pts))

out1_df2 <- model_set(fit1,
                      partables = pts,
                      progress = FALSE,
                      prior_sem_out = c(original = .001,
                                        fit2 = .0005,
                                        fit4 = .97))
# Check short names
expect_stdout(print(out1_df2, short_names = TRUE),
              out1_df2$short_names[1])

g <- model_graph(out1_df2,
                 progress = FALSE,
                 short_names = TRUE)
expect_true(setequal(out1_df2$short_names,
                     V(g)$label))
if (interactive()) {
plot(g)
}
expect_equal(g$layout$layout[, 2],
             c(2, 2, 1, 1))

if (interactive() &&
    length(unclass(packageVersion("modelbpp"))[[1]]) == 4) {

    suppressMessages(library(lavaan))
    suppressMessages(library(igraph))

    mod <-
    "
    m1 ~ x
    m2 ~ m1
    y ~ m2
    "

    fit <- sem(mod,
              dat_serial_4_weak,
              fixed.x = TRUE)

    out1 <- model_set(fit,
                      df_change_add = 1,
                      df_change_drop = 1,
                      must_not_add = c("m1~~y",
                                      "m1~y"),
                      progress = FALSE,
                      parallel = FALSE)
    g1 <- model_graph(out1,
                      progress = FALSE)

    chk_g1 <- structure(c("drop: m1~x", "drop: m2~m1", "drop: y~m2", "original",
    "original", "original", "original", "original", "original", "add: m2~x",
    "add: y~m1", "add: y~x"), dim = c(6L, 2L))

    expect_identical(as_edgelist(g1),
                      chk_g1,
                      info = "model_graph: df diff = 1")

    out2 <- model_set(fit,
                      df_change_add = 2,
                      df_change_drop = 2,
                      must_not_add = c("m1~~y",
                                      "m1~y"),
                      progress = FALSE,
                      parallel = FALSE)
    g2 <- model_graph(out2,
                      progress = FALSE)

    chk_g2 <- structure(c("add: m2~x", "add: m2~x", "add: y~m1", "add: y~m1",
    "add: y~x", "add: y~x", "drop: m1~x", "drop: m2~m1", "drop: y~m2",
    "drop: m1~x;m2~m1", "drop: m1~x;m2~m1", "drop: m1~x;y~m2", "drop: m1~x;y~m2",
    "drop: m2~m1;y~m2", "drop: m2~m1;y~m2", "original", "original",
    "original", "add: m2~x;y~m1", "add: m2~x;y~x", "add: m2~x;y~m1",
    "add: y~m1;y~x", "add: m2~x;y~x", "add: y~m1;y~x", "original",
    "original", "original", "drop: m1~x", "drop: m2~m1", "drop: m1~x",
    "drop: y~m2", "drop: m2~m1", "drop: y~m2", "add: m2~x", "add: y~m1",
    "add: y~x"), dim = c(18L, 2L))

    expect_identical(as_edgelist(g2),
                      chk_g2,
                      info = "model_graph: df diff = 2")

    expect_true(is_igraph(g1),
                info = "model_graph")
    expect_true(is_igraph(g2),
                info = "model_graph")

    plot(g1)
    plot(g2)

    out3 <- model_set(fit,
                      prior_sem_out = .30,
                      df_change_add = 1,
                      df_change_drop = 1,
                      must_not_add = c("m1~~y",
                                       "m1~y"),
                      progress = FALSE,
                      parallel = FALSE)

    g3 <- model_graph(out3,
                      progress = FALSE)
    plot(g3)

}

# WIP

if (FALSE) {

    suppressMessages(library(lavaan))

    dat <- dat_sem

    mod <-
    "
    f1 =~ x1 + x2 + x3 + x4
    f2 =~ x5 + x6 + x7 + x8
    f3 =~ x9 + x10 + x11 + x12
    f4 =~ x13 + x14 + x15 + x16
    f2 ~ a * f1
    f3 ~ a * f2
    f4 ~ f3
    "

    fit <- sem(mod, dat_sem)
    pt <- parameterTable(fit)

    mod_to_drop <- get_drop(fit,
                          df_change = 3)
    length(mod_to_drop)
    head(mod_to_drop)
    names(mod_to_drop)
    attributes(mod_to_drop[[92]])

    tmp1 <- sapply(mod_to_drop, attr, "df_expected")
    tmp2 <- sapply(mod_to_drop, attr, "df_actual")
    all.equal(tmp1, tmp2)

    # Test Print

    mod_to_drop
    print(mod_to_drop, max_tables = 25)
  }

#' @title Print a `model_set`-Class Object
#'
#' @description Print the content of
#' a `model_set`-class object.
#'
#' @details
#' It is the print method of the
#' output of [model_set()].
#'
#' @return
#' `x` is returned invisibly.
#' Called for its side effect.
#'
#' @param x A `model_set`-class object.
#'
#' @param bic_digits The number of
#' decimal places to be displayed
#' for BIC.
#' Default is 3.
#'
#' @param bpp_digits The number of
#' decimal places to be displayed
#' for BIC posterior probability
#' and prior probabilities.
#' Default is 3.
#'
#' @param sort_models Whether the models
#' will be sorted by BIC posterior
#' probability.
#' Default is `TRUE`.
#'
#' @param max_models The maximum number
#' of models to be printed. Default is
#' 20.
#'
#' @param bpp_target The desired
#' BIC probability. Used to compute
#' and print
#' the minimum prior probability
#' of the target model required to
#' achieve `bpp_target`. Default
#' is `NULL`.
#'
#' @param target_name The name of the
#' target model as appeared in the
#' model list. Default is `"original"`.
#' Used if `bpp_target` is not `NULL`.
#'
#' @param more_fit_measures Character
#' vector. To be passed to
#' [lavaan::fitMeasures()]. Default
#' is `c("cfi", "rmsea", "srmr")`. Set it to
#' `NULL` to disable printing additional
#' fit measures.
#'
#' @param fit_measures_digits The number of
#' decimal places to be displayed
#' for additional fit measures, if
#' requested. Default is 3.
#'
#' @param short_names If `TRUE`,
#' then simple short names will be
#' printed
#' along with full model names.
#' Default is `FALSE`. Short names
#' can be used when interpreting
#' the graph from `model_graph()`
#' if short names are used in the graph.
#'
#' @param cumulative_bpp If `TRUE` and
#' the models are sorted by BPPs,
#' cumulative BPPs will be printed.
#' Default is `FALSE`.
#'
#' @param ...  Optional arguments.
#' Ignored.
#'
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>
#'
#' @seealso A `model_set`-class object
#' is generated by [model_set()].
#'
#' @examples
#'
#' library(lavaan)
#'
#' dat <- dat_path_model
#'
#' mod <-
#' "
#' x3 ~ a*x1 + b*x2
#' x4 ~ a*x1
#' ab := a*b
#' "
#'
#' fit <- sem(mod, dat_path_model, fixed.x = TRUE)
#'
#' out <- model_set(fit)
#' out
#'
#' @export

print.model_set <- function(x,
                            bic_digits = 3,
                            bpp_digits = 3,
                            sort_models = TRUE,
                            max_models = 20,
                            bpp_target = NULL,
                            target_name = "original",
                            more_fit_measures = c("cfi", "rmsea", "srmr"),
                            fit_measures_digits = 3,
                            short_names = FALSE,
                            cumulative_bpp = FALSE,
                            ...) {
    fit_n <- length(x$models)
    fit_names <- names(x$models)
    models_fitted <- !is.null(x$fit)
    model_set_call <- x$model_set_call
    if (!models_fitted) {
        fit_many_call <- NULL
        k_converged <- NA
        k_post_check <- NA
        all_converged <- NA
        all_post_checked <- NA
      } else {
        fit_many_call <- x$call
        k_converged <- sum(sapply(x$converged, isTRUE))
        k_post_check <- sum(sapply(x$post_check, isTRUE))
        all_converged <- all(x$converged)
        all_post_checked <- all(x$post_check)
      }
    if (!models_fitted) {
        change_tmp <- rep(NA, fit_n)
        model_df_tmp <- rep(NA, fit_n)
        prior_tmp <- rep(NA, fit_n)
        bic_tmp <- rep(NA, fit_n)
        postprob_tmp <- rep(NA, fit_n)
      } else {
        change_tmp <- x$change
        if (!is.null(x$model_df)) {
            model_df_tmp <- x$model_df
          } else {
            model_df_tmp <- sapply(x$fit, lavaan::fitMeasures,
                                  fit.measures = "df")
          }
        model_df_tmp <- unname(model_df_tmp)
        prior_tmp <- x$prior
        bic_tmp <- x$bic
        postprob_tmp <- x$bpp
      }
    out_table <- data.frame(modification = fit_names,
                            model_df = model_df_tmp,
                            df_diff = change_tmp)
    if (models_fitted && !all_converged) {
        out_table$Converged <- ifelse(x$converged,
                                      "Yes",
                                      "No")
      }
    if (models_fitted && !all_post_checked) {
        out_table$Check <- ifelse(x$post_check,
                                      "Passed",
                                      "Failed")
      }
    out_table$Prior <- prior_tmp
    out_table$BIC <- bic_tmp
    out_table$BPP <- postprob_tmp
    if (!is.null(more_fit_measures)) {
        fit_fm <- sapply(x$fit,
                         function(xx) {
                            out <- tryCatch(lavaan::fitMeasures(xx,
                                            fit.measures = more_fit_measures,
                                            output = "vector"),
                                            error = function(e) e)
                            if (inherits(out, "error")) {
                                out <- rep(NA, length(more_fit_measures))
                              }
                            out
                          })
        fit_fm <- t(fit_fm)
        out_table <- cbind(out_table, fit_fm)
      }
    if (sort_models && models_fitted && all_converged) {
        i <- order(out_table$BPP,
                   decreasing = TRUE)
        out_table <- out_table[i, ]
        if (cumulative_bpp) {
            tmp <- round(cumsum(out_table$BPP), bpp_digits)
            tmp <- formatC(tmp,
                          digits = bpp_digits,
                          format = "f")
            out_table["Cumulative"] <- tmp
          }
      }
    out_table_print <- out_table
    out_table_print$Prior <- round(out_table_print$Prior,
                                 digits = bpp_digits)
    out_table_print$Prior <- formatC(out_table_print$Prior,
                                   digits = bpp_digits,
                                   format = "f")
    out_table_print$BIC <- round(out_table_print$BIC,
                                 digits = bic_digits)
    out_table_print$BIC <- formatC(out_table_print$BIC,
                                   digits = bic_digits,
                                   format = "f")
    out_table_print$BPP <- round(out_table_print$BPP,
                                 digits = bpp_digits)
    out_table_print$BPP <- formatC(out_table_print$BPP,
                                   digits = bpp_digits,
                                   format = "f")
    if (!is.null(more_fit_measures)) {
        fm_names <- colnames(fit_fm)
        for (xx in fm_names) {
            out_table_print[xx] <- formatC(out_table_print[, xx, drop = TRUE],
                                           digits = fit_measures_digits,
                                           format = "f")
          }
      }
    if (short_names) {
        if (is.character(x$short_names)) {
            tmp1 <- x$short_names
            tmp2 <- out_table_print$modification
            if (isTRUE(setequal(names(tmp1),
                                tmp2))) {
                xx <- unname(tmp1[tmp2])
                out_table_print <- cbind(Short = xx,
                                         out_table_print)
              }
          }
      }
    if (isTRUE(fit_n > max_models)) {
        gt_max_models <- TRUE
        x_tmp <- out_table_print[seq_len(max_models), ]
      } else {
        gt_max_models <- FALSE
        x_tmp <- out_table_print
      }
    cat("\n")
    cat("Call:\n")
    # Handle elements that may not be names
    if (!is.name(model_set_call[[1]])) {
      model_set_call[[1]] <- quote(modelbpp::model_set)
    }
    if (!is.name(model_set_call$sem_out) &&
        !is.null(model_set_call$sem_out)) {
      model_set_call$sem_out <- "<hidden>"
    }
    print(model_set_call)
    cat("\n")
    if (models_fitted) {
        tmp1 <- c("Number of model(s) fitted",
                  "Number of model(s) converged",
                  "Number of model(s) passed post.check")
        tmp2 <- c(fit_n,
                  k_converged,
                  k_post_check)
        tmp3 <- auto_tab(tmp1,
                         tmp2,
                         between = ": ")
        cat(tmp3, sep = "\n")
        if (!is.null(bpp_target)) {
            bpp_min <- min_prior(x$bic,
                                 bpp_target = bpp_target,
                                 target_name = target_name)
            tmp <- data.frame(x = c(
                      formatC(bpp_target, digits = bpp_digits, format = "f"),
                      formatC(bpp_min, digits = bpp_digits, format = "f"),
                      formatC(x$bpp[target_name], digits = bpp_digits, format = "f")))
            colnames(tmp) <- paste0("Target Model: ",
                                    target_name)
            rownames(tmp) <- c("Desired minimum BPP:",
                               "Required minimum prior probability:",
                               "Current BPP:")
            print(tmp)
          }
      } else {
        cat("Models are not fitted.")
        cat("\n")
      }
    cat("\n")
    tmp1 <- ifelse(sort_models && all(!is.na(out_table$BPP)),
                   " (sorted by BPP)",
                   "")
    if (!models_fitted) tmp1 <- ""
    if (gt_max_models) {
        cat("The first ",
            max_models,
            " model(s)",
            tmp1,
            ":\n", sep = "")
      } else {
        cat("The models",
            tmp1,
            ":\n", sep = "")
      }
    rownames(x_tmp) <- x_tmp$modification
    x_tmp$modification <- NULL
    print(x_tmp)

    if (models_fitted && !all_converged) {
        x_tmp2 <- out_table_print[out_table_print$Converged != "Yes", ]
        rownames(x_tmp2) <- x_tmp2$modification
        x_tmp2$modification <- NULL
        cat("\nModel(s) not converged:\n")
        print(x_tmp2)
      }

    if (models_fitted && !all_post_checked) {
        x_tmp2 <- out_table_print[out_table_print$Check != "Passed", ]
        rownames(x_tmp2) <- x_tmp2$modification
        x_tmp2$modification <- NULL
        cat("\nModel(s) failed lavaan's post.check:\n")
        print(x_tmp2)
      }

    if (!is.null(x$equivalent_clusters)) {
        tmp <- as.vector(sapply(x$equivalent_clusters, paste, collapse = ", "))
        tmp <- data.frame(Cluster = tmp)
        tmp2 <- names(x$models_equivalent)
        cat("\nModels that are equivalent:\n")
        print(tmp, right = FALSE)
        cat("\nEquivalent model(s) excluded from the analysis:\n")
        catwrap(paste(tmp2, collapse = ", "))
      }

    cat("\nNote:\n")
    cat("- BIC: Bayesian Information Criterion.\n")
    cat("- BPP: BIC posterior probability.\n")
    cat("- model_df: Model degrees of freedom.\n")
    cat("- df_diff: Difference in df compared to the original/target model.\n")
    if (sort_models) {
        if (("Cumulative" %in% colnames(x_tmp))) {
            cat("- Cumulative: Cumulative BIC posterior probability.\n")
          } else {
            cat("- To show cumulative BPPs, call print() with 'cumulative_bpp = TRUE'.\n")
          }
      }
    if (gt_max_models) {
        tmp <- paste(fit_n,
                   "models were fitted but",
                   max_models,
                   "were printed. Call print() and",
                   "set 'max_models' to a larger number",
                   "to print more models, or set it to",
                   "NA to print all models.")
        catwrap(tmp, initial = "- ", exdent = 2)
      }
    if (models_fitted &&
        (k_converged != fit_n) &&
        any(is.na(postprob_tmp))) {
        tmp <- "BPP and/or prior not computed because one or more models not converged."
        catwrap(tmp, initial = "- ", exdent = 2)
      }
    if (models_fitted &&
        (k_post_check != fit_n)) {
        tmp <- "Interpret with caution. One or more models failed lavaan's post.check."
        catwrap(tmp, initial = "- ", exdent = 2)
      }
    if (x$drop_equivalent_models && x$fixed_x) {
        tmp <- "At least one model has fixed.x = TRUE. The models are not checked for equivalence."
        catwrap(tmp, initial = "- ", exdent = 2)
      }
    tmp <- paste0("Since Version 0.1.6.3, the default ways to ",
                  "handle factor loadings have changed. ",
                  "Check the NEWS by news(package = 'modelbpp') to ",
                  "see how to reproduce results from previous versions.")
    catwrap(tmp, initial = "- ", exdent = 2)

    invisible(x)
  }

#' @title BIC Posterior Probabilities
#' of Neighboring Models
#'
#' @description Identify neighboring
#' models, fit them, and return the
#' BIC posterior probabilities.
#'
#' @details It computes the BIC
#' posterior probabilities of a set
#' of models by the method presented
#' in Wu, Cheung, and Leung (2020).
#'
#' First, a list of model is identified
#' based on user-specified criteria.
#' By default, models differ from a fitted
#' model by one degree of freedom,
#' the 1-df-away *neighboring* models,
#' will be found using [get_add()]
#' and [get_drop].
#'
#' Second, these models will be fitted
#' to the sample dataset, and their
#' BICs will be computed.
#'
#' Third, their BIC posterior
#' probabilities will be computed
#' using their BICs. By default,
#' equal prior
#' probabilities for all the models
#' being fitted will be assumed in
#' the current version. This can be
#' changed by `prior_sem_out`.
#'
#' The results can then be printed,
#' with the models sorted by descending
#' order of BIC posterior
#' probabilities. The results can
#' also be visualized using
#' [model_graph()].
#'
#' @inheritParams get_add
#' @inheritParams get_drop
#'
#' @param sem_out It can be the output
#' from an
#' SEM function. Currently it supports
#' [lavaan::lavaan-class] objects only.
#' If it is a named list of
#' [lavaan::lavaan-class] objects, then
#' all arguments for model generation
#' will be ignored, and models will not
#' be refitted. Users need to ensure
#' that the models can be meaningfully
#' compared because they will not be
#' checked.
#'
#' @param partables A `partables`-class
#' object, usually generated by
#' [get_add()] or [get_drop()]. A named
#' list of parameter tables to be fitted
#' along with the original model in
#' `sem_out`.
#' If supplied, all arguments related
#' to identifying models will be ignored.
#' Default is `NULL`.
#'
#' @param model_set_out If set to
#' the output of a previous call
#' to [model_set()] (a `model_set`-class
#' object), the list of stored models
#' will be used. All arguments related
#' to generate neighboring models will
#' be ignored.
#' If supplied, `sem_out` will also be
#' ignored and will be retrieved from
#' `model_set_out`, and `partables`
#' will also be ignored.
#' Default is `NULL`.
#'
#' @param prior_sem_out The prior of the
#' model fitted in `sem_out`. Default
#' is `NULL`, and all models will have
#' equal prior probabilities.
#'
# @param must_add
#'
# @param must_not_add
#'
# @param remove_constraints
#'
# @param exclude_error_cov
#'
#' @param exclude_feedback Exclude
#' paths that will result in a feedback
#' loop. For example, if there is
#' path from `x` through `m` to `y`,
#' then the path `x ~ y` will create
#' a feedback loop.  Default has been
#' changed to
#' `TRUE` since Version 0.1.3.5 because
#' feedback loops are usually
#' not included except when theoretically
#' justified. To reproduce results
#' based on previous version, set this
#' argument to `FALSE`.
#'
#' @param exclude_xy_cov Exclude
#' covariance between two variables,
#' in which one has a path to another.
#' For example, if there is
#' path from `x` through `m` to `y`,
#' then the covariance `x ~~ y`,
#' which denotes the covariance between
#' `x` and the error term of `y`, will
#' be excluded if this argument is
#' `TRUE`. Default has been changed to
#' `TRUE` since Version 0.1.3.5 because
#' these covariances rarely are
#' interpretable. To reproduce results
#' based on previous version, set this
#' argument to `FALSE`.
#'
# @param must_drop
#'
# @param must_not_drop
#'
#' @param loadings_to_exclude_from_drop
#' Whether factor loadings
#' will be excluded when deciding
#' parameters to drop. If `"single"`,
#' then only "single" loadings (an indicator
#' loads on only one latent factor)
#' will be excluded (they will not be
#' dropped). If `"all"`, then
#' all factor loadings will be excluded
#' (no factor loadings will be dropped).
#' If `"none"`, then no loadings will
#' be excluded, and all free loadings
#' will be considered to be dropped. Be careful when using
#' `"none"` because the models may not
#' make sense. The settings in `must_drop`
#' and `must_not_drop` will override this
#' argument.
#'
#' @param df_change_add How many degrees
#' of freedom (*df*) away in the list.
#' All models with *df* change less than
#' or equal to this number will be
#' included, taking into account
#' requirements set by other arguments.
#' Default is 1.
#'
#' @param df_change_drop How many degrees
#' of freedom away in the list. All
#' models with *df* change less than
#' or equal to this number will be
#' included, taking into account
#' requirements set by other arguments.
#' Default is 1.
#'
# @param remove_duplicated
#'
#' @param fit_models If `TRUE`,
#' the
#' models will be fitted to the data,
#' usually stored in `sem_out`.
#' If `FALSE`, the models will be returned
#' as is, in the element `models`
#' of the output. If `model_set_out`
#' is set and models have been fitted,
#' then default is `FALSE`. Otherwise,
#' default is `TRUE`.
#'
#' @param compute_bpp If `TRUE`, then BIC
#' posterior probabilities will be
#' computed. Default
#' is `TRUE`.
#'
#' @param original String. The name of the
#' original (traget) model. Default is
#' `"original"`. Used if `prior_sem_out`
#' is unnamed and only has one value.
#'
#' @param parallel If `TRUE`, parallel
#' processing will be used to fit the
#' models. Default is `FALSE`.
#'
#' @param ncores Numeric. The number of
#' CPU cores to be used if `parallel`
#' is `TRUE`.
#'
#' @param make_cluster_args A list of
#' named arguments to be passed to
#' `parallel::makeCluster()`. Used by
#' advanced users to configure the
#' cluster if `parallel` is `TRUE`.
#' Default is `list()`.
#'
#' @param progress Whether a progress
#' bar will be displayed, implemented
#' by the `pbapply` package or by
#' `utils::txtProgressBar`. Default
#' is `TRUE`.
#'
#' @param verbose Whether additional
#' messages will be displayed, such
#' as the expected processing time.
#' Default is `TRUE`.
#'
#' @param output If `"model_set"`,
#' then the output is a `model_set`-class
#' object. If `"partables"`, the output
#' is a `partables`-class object.
#' Default is `partables`.
#'
#' @param skip_check_sem_out If `TRUE`
#' and `sem_out` is set, check
#' whether `sem_out` is of a supported
#' type (estimator is `"ML"` and
#' the model has only one group). If
#' not, an error will be raised.
#' Can be set to `FALSE` for experimenting
#' the functions on models not officially
#' supported.
#'
#' @param drop_equivalent_models If
#' `TRUE`, the default, equivalent
#' models will be dropped in the final
#' output. This check can only be
#' conducted when no models are fitted
#' in [lavaan::lavaan()] with
#' `fixed.x = TRUE` (which is the
#' default of [lavaan::sem()]).
#'
#' @return The function [model_set()]
#' returns an object of the class
#' `model_set`, a list with the following
#' major elements:
#'
#' * `models`: A named list of parameter
#'    tables. Each represent the models
#'    identified.
#'
#' * `bic`: A numeric vector, of the
#'    same length as `model`. The BIC of
#'    each model.
#'
#' * `postprob`: A numeric vector, of
#'    the same length as `model`. The
#'    BIC posterior probability of each
#'    model.
#'
#' * `fit`: A named list of
#'    [lavaan::lavaan()] output objects
#'    or [update()] for fitting a model
#'    with the added parameters, of the
#'    same length as `model`.
#'
#' * `change`: A numeric vector, of the
#'    same length as `model`. The change
#'    in model df for each fit. A
#'    positive number denotes one less
#'    free parameter. A negative number
#'    denotes one more free parameter or
#'    one less constraint.
#'
#' * `converged`: A named vector of
#'    boolean values, of the same length
#'    as `model`. Indicates whether each
#'    fit converged or not.
#'
#' * `post_check`: A named vector of
#'    boolean values, of the same length
#'    as `model`. Indicates whether the
#'    solution of each fit is admissible
#'    or not. Checked by
#'    [lavaan::lavInspect()].
#'
#' The object returned by [gen_models()]
#' depends on the argument `output`.
#' See the argument `output` for the
#' details
#'
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>
#'
#' @references
#' Wu, H., Cheung, S. F., & Leung, S. O.
#' (2020). Simple use of BIC to assess
#' model selection uncertainty: An
#' illustration using mediation and
#' moderation models.
#' *Multivariate Behavioral Research*,
#' *55*(1), 1--16.
#' \doi{10.1080/00273171.2019.1574546}
#'
#' @seealso [print.model_set()]
#'
#' @examples
#'
#' library(lavaan)
#'
#' dat <- dat_path_model
#'
#' mod <-
#' "
#' x3 ~ a*x1 + b*x2
#' x4 ~ a*x1
#' ab := a*b
#' "
#'
#' fit <- sem(mod, dat_path_model, fixed.x = TRUE)
#'
#' out <- model_set(fit)
#' out
#'
#' @describeIn model_set Compute the BPPs of a list of models.
#'  Can generate the models and/or fit the models. Can also
#'  accept pregenerated models, or just update BPPs.
#' @order 1
#'
#' @export

model_set <- function(sem_out,
                      partables = NULL,
                      model_set_out = NULL,
                      prior_sem_out = NULL,
                      must_add = NULL,
                      must_not_add = NULL,
                      must_drop = NULL,
                      must_not_drop = NULL,
                      loadings_to_exclude_from_drop = c("single", "none", "all"),
                      remove_constraints = TRUE,
                      exclude_error_cov = TRUE,
                      exclude_feedback = TRUE,
                      exclude_xy_cov = TRUE,
                      exclude_x_changed_to_y = TRUE,
                      cross_add = c("pure_x", "pure_y"),
                      cross_sets = NULL,
                      df_change_add = 1,
                      df_change_drop = 1,
                      remove_duplicated = TRUE,
                      fit_models = ifelse(!is.null(model_set_out$fit),
                                          FALSE,
                                          TRUE),
                      compute_bpp = TRUE,
                      original = "original",
                      parallel = FALSE,
                      ncores = max(parallel::detectCores(logical = FALSE) - 1, 1),
                      make_cluster_args = list(),
                      progress = TRUE,
                      verbose = TRUE,
                      skip_check_sem_out = FALSE,
                      drop_equivalent_models = TRUE) {
  # if (missing(sem_out)) stop("sem_out is not supplied.")
  user_fits <- FALSE
  if (!missing(sem_out)) {
      if (is.list(sem_out)) {
          all_lavaan <- sapply(sem_out,
                               inherits,
                               what = "lavaan")
          if (!all(all_lavaan)) {
              stop("Not all objects in sem_out are lavaan-class object.")
            }
          if (is.null(names(sem_out))) {
              stop("The list for sem_out needs to be a named list.")
            }
          tmp <- sapply(sem_out, lavaan::lavInspect, what = "converged")
          if (any(!tmp)) {
              stop("One or more supplied models have not converged.")
            }
          user_fits <- TRUE
          model_set_out <- NULL
          remove_duplicated <- FALSE
          fit_models <- FALSE
        } else {
          if (!inherits(sem_out, "lavaan")) {
              stop("sem_out is not a lavaan-class object.")
            }
          if (!skip_check_sem_out) {
              check_sem_out(sem_out)
            }
        }
    }
  if (!is.null(model_set_out)) {
      if (!inherits(model_set_out, "model_set")) {
          stop("model_set_out is not a model_set-class object.")
        }
      sem_out <- model_set_out$sem_out
      mod_to_fit <- model_set_out$models
    } else if (user_fits) {
      mod_to_fit <- do.call(to_partables, sem_out)
    } else {
      if (!is.null(partables)) {
          if (!inherits(partables, "partables")) {
              stop("partables is not a partables-class object.")
            }
          mod_all <- partables
        } else {
          mod_to_add <- get_add(sem_out,
                                must_add = must_add,
                                must_not_add = must_not_add,
                                cross_add = cross_add,
                                cross_sets = cross_sets,
                                remove_constraints = remove_constraints,
                                exclude_error_cov = exclude_error_cov,
                                exclude_feedback = exclude_feedback,
                                exclude_xy_cov = exclude_xy_cov,
                                exclude_x_changed_to_y = exclude_x_changed_to_y,
                                df_change = df_change_add,
                                remove_duplicated = FALSE,
                                progress = progress)
          mod_to_drop <- get_drop(sem_out,
                                  must_drop = must_drop,
                                  must_not_drop = must_not_drop,
                                  loadings_to_exclude = loadings_to_exclude_from_drop,
                                  df_change = df_change_drop,
                                  remove_duplicated = FALSE,
                                  progress = progress)
          mod_all <- c(mod_to_add, mod_to_drop)
        }
      # NOTE: No need to call unique_models() here.
      # There is a final check below
      # if (remove_duplicated) {
      #     mod_all <- unique_models(mod_all)
      #   }
      class(mod_all) <- setdiff(class(mod_all), "partables")
      pt0 <- lavaan::parameterTable(sem_out)
      mod_to_fit <- c(mod_all, list(`original` = pt0))
    }
  if (remove_duplicated) {
      mod_to_fit <- unique_models(mod_to_fit,
                                  progress = progress)
    }
  if (!inherits(mod_to_fit, "partables")) {
      class(mod_to_fit) <- c("partables", class(mod_to_fit))
    }
  if (fit_models) {
      out <- fit_many(model_list = mod_to_fit,
                      sem_out = sem_out,
                      parallel = parallel,
                      ncores = ncores,
                      make_cluster_args = make_cluster_args,
                      progress = progress,
                      verbose = verbose)
    } else if (user_fits) {
      out <- lavaan_to_sem_outs(sem_out,
                                original = original)
    } else {
      if (!is.null(model_set_out)) {
          out <- model_set_out
          tmp1 <- class(out)
          tmp2 <- which(class(out) == "model_set")
          if (length(tmp2) == 1) {
              tmp1 <- tmp1[-tmp2]
            }
          class(out) <- tmp1
        } else if (user_fits) {
          # Placeholder
        } else {
          # Create an object of the class "sem_outs"
          # Not a good solution but do not have a better one for now
          out <- list(fit = NULL,
                      change = NULL,
                      converged = NULL,
                      post_check = NULL,
                      model_df = NULL,
                      call = NULL)
          class(out) <- c("sem_outs", class(out))
        }
    }
  # Check equivalent clusters
  any_fixedx <- any(sapply(out$fit, lavaan::lavInspect, what = "fixed.x"),
                    na.rm = TRUE)
  # if (drop_equivalent_models && isTRUE(any_fixedx)) {
  #     warning("Cannot check for equivalent models if fixed.x = TRUE.")
  #   }
  if (!is.null(out$fit) && drop_equivalent_models && isFALSE(any_fixedx)) {
      mod_eq <- tryCatch(equivalent_clusters(out$fit,
                                             progress = progress,
                                             name_cluster = TRUE),
                         error = function(e) e)
      if (inherits(mod_eq, "error")) {
          if (grepl("fixed.x = TRUE", mod_eq$message, fixed = TRUE)) {
              warning("Cannot check for equivalent models if fixed.x = TRUE.")
            }
          mod_eq <- NULL
        }
      if (!is.null(mod_eq)) {
          to_drop <- numeric(0)
          m_names <- names(out$fit)
          for (xx in mod_eq) {
              i <- xx[-1]
              to_drop <- c(to_drop, match(i, m_names))
            }
          out$fit <- out$fit[-to_drop]
          if (!is.null(out$change)) {
              out$change <- out$change[-to_drop]
            }
          out$post_check <- out$post_check[-to_drop]
          out$converged <- out$converged[-to_drop]
          if (!is.null(out$model_df)) {
              out$model_df <- out$model_df[-to_drop]
            }
          mod_to_fit_equivalent <- mod_to_fit[to_drop]
          mod_to_fit <- mod_to_fit[-to_drop]
        } else {
          mod_to_fit_equivalent <- NULL
          mod_eq <- NULL
        }
    } else {
      mod_to_fit_equivalent <- NULL
      mod_eq <- NULL
    }
  out$models <- mod_to_fit
  out$models_equivalent <- mod_to_fit_equivalent
  out$equivalent_clusters <- mod_eq
  out$drop_equivalent_models <- drop_equivalent_models
  out$fixed_x <- any_fixedx
  if (compute_bpp && !is.null(out$fit)) {
      bic_list <- sapply(out$fit,
            function(x) {
                out <- tryCatch(as.numeric(lavaan::fitMeasures(x, "bic")),
                                error = function(e) NA)
                out
              })
      out$bic <- bic_list
      if (!is.null(prior_sem_out)) {
          if (sum(prior_sem_out) >= 1) {
              stop("The sum of user prior probabilities must be less than 1.")
            }
          if ((length(prior_sem_out) == 1) && is.null(names(prior_sem_out))) {
              p <- length(out$bic)
              i_original <- which(names(out$models) == original)
              prior_tmp <- rep((1 - prior_sem_out) / (p - 1), p)
              prior_tmp[i_original] <- prior_sem_out
              out$prior <- prior_tmp
            } else {
              p <- length(out$bic)
              q <- length(prior_sem_out)
              i_original <- match(names(prior_sem_out), names(out$models))
              prior_sem_out_tmp <- prior_sem_out[!is.na(i_original)]
              i_original <- i_original[!is.na(i_original)]
              prior_tmp <- rep((1 - sum(prior_sem_out)) / (p - q), p)
              prior_tmp[i_original] <- prior_sem_out_tmp
              out$prior <- prior_tmp
            }
        } else {
          # Assume unbiased priors for all models
          out$prior <- rep(1 / length(out$bic), length(out$bic))
        }
      out$bpp <- bpp(bic = out$bic,
                          prior = out$prior)
    } else {
      out$bic <- NULL
      out$bpp <- NULL
    }
  # TODO: Finalize what to do with NA BIC
  if (any(is.na(out$bic)) && !is.null(out$bic)) {
      out$prior <- rep(NA, length(bic_list))
      out$bpp <- rep(NA, length(bic_list))
    }
  out$model_set_call <- match.call()
  out$sem_out <- sem_out
  # Create Short Names
  out <- short_model_names(out)
  class(out) <- c("model_set", class(out))
  out
}

#' @describeIn model_set Generate a list of models (parameter tables).
#' @order 2
#' @export

gen_models <- function(sem_out,
                       must_add = NULL,
                       must_not_add = NULL,
                       must_drop = NULL,
                       must_not_drop = NULL,
                       loadings_to_exclude_from_drop = c("single", "none", "all"),
                       remove_constraints = TRUE,
                       exclude_error_cov = TRUE,
                       exclude_feedback = TRUE,
                       exclude_xy_cov = TRUE,
                       exclude_x_changed_to_y = TRUE,
                       cross_add = c("pure_x", "pure_y"),
                       cross_sets = NULL,
                       df_change_add = 1,
                       df_change_drop = 1,
                       remove_duplicated = TRUE,
                       drop_equivalent_models = TRUE,
                       progress = TRUE,
                       output = c("partables", "model_set")) {
    output <- match.arg(output)
    loadings_to_exclude_from_drop <- match.arg(loadings_to_exclude_from_drop)
    if (!is.null(cross_add)) {
      cross_add <- match.arg(cross_add)
    }
    out <- model_set(sem_out = sem_out,
                     must_add = must_add,
                     must_not_add = must_not_add,
                     must_drop = must_drop,
                     must_not_drop = must_not_drop,
                     loadings_to_exclude_from_drop = loadings_to_exclude_from_drop,
                     remove_constraints = remove_duplicated,
                     exclude_error_cov = exclude_error_cov,
                     exclude_feedback = exclude_feedback,
                     exclude_xy_cov = exclude_xy_cov,
                     exclude_x_changed_to_y = exclude_x_changed_to_y,
                     cross_add = cross_add,
                     cross_sets = cross_sets,
                     df_change_add = df_change_add,
                     df_change_drop = df_change_drop,
                     remove_duplicated = remove_duplicated,
                     drop_equivalent_models = drop_equivalent_models,
                     progress = progress,
                     fit_models = FALSE,
                     compute_bpp = FALSE)
    out2 <- switch(output,
                   model_set = out,
                   partables = out$models)
    out2
  }

#' @title BIC Posterior Probabilities
#' @noRd

bpp <- function(bic,
                prior = NULL) {
    k <- length(bic)
    if (is.null(prior)) {
        prior <- rep(1 / k, k)
      }
    di <- exp(-.5 * (bic - bic[1])) * prior
    out <- di / sum(di)
    out
  }

#' @noRd

short_model_names <- function(object) {
    if (!is.character(names(object$models))) {
        return(object)
      }
    p <- length(object$models)
    short_names <- paste0("M", seq_len(p))
    names(short_names) <- names(object$models)
    object$short_names <- short_names
    object
  }

suppressMessages(library(lavaan))

# Update bpp

dat <- dat_path_model

mod <-
"
x2 ~ x3 + 0*x4
x1 ~ x3
"

fit <- sem(mod,
           dat_path_model,
           fixed.x = TRUE)

out <- model_set(fit,
                 progress = FALSE,
                 parallel = FALSE)

out2 <- model_set(fit,
                  prior_sem_out = .40,
                  parallel = FALSE,
                  progress = FALSE)

out3 <- model_set(model_set_out = out,
                  fit_models = FALSE,
                  prior_sem_out = .40,
                  progress = FALSE,
                  parallel = FALSE)

expect_equal(out2$bpp,
              out3$bpp,
              info = "BPP update")

suppressMessages(library(lavaan))

# Basic Tests

dat <- dat_path_model

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)

out <- model_set(fit,
                 progress = FALSE,
                 parallel = FALSE)

expect_equivalent(
    out$bpp,
    c(9.999998e-01, 2.382290e-08, 8.302384e-13, 1.289951e-09, 1.712318e-07),
    tolerance = 1e-5,
    info = "Posterior probability as expected"
  )

expect_equivalent(
    out$bic,
    c(400.2915, 435.3967, 455.9256, 441.2288, 431.4519),
    tolerance = 1e-5,
    info = "BIC as expected"
  )

# Check more_fit_measures

expect_stdout(print(out,
                    more_fit_measures = c("CFI", "RMSEA"),
                    fit_measures_digits = 2),
              "rmsea")

out <- model_set(fit,
                 df_change_add = 2,
                 df_change_drop = 2,
                 progress = FALSE,
                 parallel = FALSE)

expect_equivalent(
    out$bpp,
    c(8.567850e-01, 2.041111e-08, 1.432148e-01, 7.113360e-13, 1.105211e-09,
      2.891960e-21, 1.467088e-07),
    tolerance = 1e-5,
    info = "Posterior probability as expected"
  )

expect_equivalent(
    out$bic,
    c(400.2915, 435.3967, 403.8691, 455.9256, 441.2288,
      494.5670, 431.4519),
    tolerance = 1e-5,
    info = "BIC as expected"
  )

expect_stdout(print(out),
              "The models (sorted by BPP)",
              fixed = TRUE)

# Test previous output

out0 <- model_set(fit,
                  progress = FALSE,
                  parallel = FALSE)
out1 <- model_set(model_set_out = out0,
                  prior = .50,
                  progress = FALSE,
                  parallel = FALSE)
out2 <- model_set(fit,
                  prior_sem_out = .50,
                  progress = FALSE,
                  parallel = FALSE)

expect_identical(out1$bpp,
                 out2$bpp,
                 info = "User supplied model_set")

# Test user partables

out0 <- model_set(fit,
                  progress = FALSE,
                  parallel = FALSE)
tmp <- out0$models[-5]
class(tmp) <- "partables"
out1 <- model_set(fit,
                  partables = tmp,
                  progress = FALSE,
                  parallel = FALSE)

expect_identical(out0$bic,
                  out1$bic,
                  info = "User supplied partables")
expect_identical(out0$bpp,
                  out1$bpp,
                  info = "User supplied partables")

# Test no fit

mod <-
"
x2 ~ x3 + 0*x4
x1 ~ x3
"

fit <- sem(mod,
           dat_path_model,
           fixed.x = TRUE)
out <- model_set(fit,
                 fit_models = FALSE,
                 progress = FALSE,
                 parallel = FALSE)
expect_stdout(print(out),
              "not fitted",
              fixed = TRUE,
              info = "fit_models = FALSE")

# Gen models

out <- model_set(fit,
                 progress = FALSE,
                 parallel = FALSE)
out2 <- gen_models(fit,
                   output = "model_set",
                   progress = FALSE)
out3 <- model_set(model_set_out = out2,
                  parallel = FALSE,
                  progress = FALSE)
out4 <- gen_models(fit,
                   progress = FALSE)
out5 <- model_set(sem_out = fit,
                  partables = out4,
                  progress = FALSE,
                  parallel = FALSE)

expect_identical(out$models,
                  out2$models,
                  info = "Test generated models")
expect_identical(out$bpp,
                  out3$bpp,
                  info = "Test generated models")
expect_identical(out$bpp,
                  out5$bpp,
                  info = "Test generated models")

suppressMessages(library(lavaan))

# Test c()

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)

moda <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fita <- sem(moda, dat_path_model, fixed.x = TRUE)

outa <- model_set(fita,
                  progress = FALSE,
                  parallel = FALSE)

modb <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x2
ab := a*b
"

fitb <- sem(modb, dat_path_model, fixed.x = TRUE)

outb <- model_set(fitb,
                  progress = FALSE,
                  parallel = FALSE)

tmp <- c(outa$models, outb$models)

expect_true(inherits(tmp, "partables"))
expect_identical(setdiff(names(tmp),
                         unique(c(names(outa$models),
                                  names(outb$models)))),
                 character(0))

tmp <- c(outb, outa$models)

expect_true(inherits(tmp, "partables"))
expect_identical(setdiff(names(tmp),
                         unique(c(names(outa$models),
                                  names(outb$models)))),
                 character(0))

mod2 <-
"
x2 ~ x1
x3 ~ x2
x4 ~ x3
"

mod3 <-
"
x2 ~ x4
x3 ~ x4
x1 ~ x2 + x3
"

fit2 <- sem(mod2, dat_path_model)
fit3 <- sem(mod3, dat_path_model)

pt2 <- parameterTable(fit2)

tmp <- c(outa, pt2, fit3, outb$models)

expect_true(inherits(tmp, "partables"))
expect_true(all(c("pt2", "fit3") %in% names(tmp)))

tmp <- c(outb, user2 = pt2, user3 = fit3, outa$models)

expect_true(inherits(tmp, "partables"))
expect_true(all(c("user2", "user3") %in% names(tmp)))

out_all <- model_set(sem_out = fit,
                     partables = tmp,
                     progress = FALSE,
                     parallel = FALSE)
expect_equal(length(out_all$fit),
             9)
out_all2 <- model_set(sem_out = fit,
                      partables = partables_drop(tmp,
                                                 c("user3",
                                                   "add: x4~x1")),
                      progress = FALSE,
                      parallel = FALSE)

expect_false(any(c("user3", "add: x4~x1") %in% out_all2))

suppressMessages(library(lavaan))

mod1 <-
"
x2 ~ x1 + x4
x3 ~ x2
"

mod2 <-
"
x3 ~ x1
x3 ~ x2
x4 ~ x3
"

fit1 <- sem(mod1, HolzingerSwineford1939, fixed.x = FALSE)
fit2 <- sem(mod2, HolzingerSwineford1939, fixed.x = FALSE)

out1 <- model_set(fit1,
                  progress = FALSE,
                  parallel = FALSE)
out2 <- model_set(fit2,
                  progress = FALSE,
                  parallel = FALSE)

# Need to rename the fits due to name conflicts.
out1b <- out1
out2b <- out2
names(out1b$fit) <- paste0("fit1_", names(out1b$fit))
names(out2b$fit) <- paste0("fit2_", names(out2b$fit))

fit_all <- c(out1b$fit,
             out2b$fit)
names(fit_all)

outa <- model_set(fit_all,
                  progress = FALSE,
                  parallel = FALSE)

outa

outb <- model_set_combined(
            list(fit1 = out1,
                 fit2 = out2),
            progress = FALSE,
            parallel = FALSE)

outb

names(outa)
names(outb)

expect_equal(outa$bpp,
             outb$bpp)

time0 <- proc.time()

suppressMessages(library(lavaan))

dat <- dat_serial_4_weak

mod0 <-
"
m1 ~ x
m2 ~ m1 + x
y ~ m1 + m2 + x
"
fit0 <- sem(mod0, dat_serial_4_weak)

out0 <- model_set(fit0,
                  progress = FALSE,
                  parallel = FALSE)

mod1 <-
"
m1 ~ x
m2 ~ m1
"
fit1 <- sem(mod1, dat_serial_4_weak)

out1 <- model_set(fit1,
                  must_not_drop = c("m2 ~ m1", "m1 ~ x"),
                  progress = FALSE,
                  parallel = FALSE)

out2 <- model_set(fit1,
                  must_not_drop = c("m2 ~ m1", "m1 ~ x"),
                  must_not_add = c("m2 ~ x"),
                  progress = FALSE,
                  parallel = FALSE)

expect_false(
    any(grepl("add", names(out0$fit))),
    info = "No add"
  )
expect_false(
    any(grepl("drop", names(out1$fit))),
    info = "No drop"
  )
expect_true(
    identical(names(out2$fit), "original"),
    info = "original only"
  )

time1 <- proc.time()
# print(as.vector((time1 - time0)))

suppressMessages(library(lavaan))

mod <-
"
fx =~ x1 + x2 + x3
fm1 =~ x4 + x5 + x6
fm2 =~ x10 + x11 + x12
fy =~ x7 + x8 + x9
fm1 ~ fx
fm2 ~ fm1
fy ~ fm2
"

fit <- sem(mod,
           dat_sem,
           warn = FALSE)

pt <- parameterTable(fit)
lvnames <- lavNames(fit, "lv")
indnames <- lavNames(fit, "ov.ind")
all_loadings <- expand.grid(lvnames, indnames)
all_loadings <- apply(all_loadings,
                      MARGIN = 1,
                      paste0,
                      collapse = "=~")
mod_loadings <- paste0(pt[pt$op == "=~", "lhs"],
                       "=~",
                       pt[pt$op == "=~", "rhs"])
all_loadings <- setdiff(all_loadings,
                        mod_loadings)
all_must_not_add <- c(all_loadings,
                      c("fy1~fy2", "fy2~fy1",
                        "fx1~~fy2", "fx1~~fy3",
                        "fy2~~fx1", "fy3~~fx1"))
all_must_not_drop <- c(mod_loadings)

pts <- suppressWarnings(model_set(fit,
                 must_not_add = all_must_not_add,
                 must_not_drop = all_must_not_drop,
                 fit_models = FALSE,
                 compute_bpp = FALSE,
                 progress = FALSE))
pts <- pts$models

expect_true(!any(grepl("~~", names(pts))))

suppressMessages(library(lavaan))

# User Prior

dat <- dat_path_model

mod <-
"
x2 ~ x3 + 0*x4
x1 ~ x3
"

fit <- sem(mod,
           dat_path_model,
           fixed.x = TRUE)

prior0 <- .50
out <- model_set(fit,
                 prior_sem_out = prior0,
                 parallel = FALSE,
                 progress = FALSE)
bic <- out$bic
i <- which(names(bic) == "original")
p <- length(bic)
prior <- rep((1 - prior0) / (p - 1), p)
prior[i] <- prior0
d <- exp(-.5 * (bic - bic[1])) * prior
chk_bpp <- d / sum(d)

expect_equivalent(
    out$bpp,
    chk_bpp,
    info = "Posterior probability as expected: User prior"
  )

expect_error(model_set(fit,
                       prior_sem_out = c(original = .80,
                                         `add: x1~x4` = .30),
                       parallel = FALSE,
                       progress = FALSE),
             info = "Invalid priors")

suppressMessages(library(lavaan))

data(dat_path_model)
set.seed(8970431)
dat_path_model$group <- sample(c("Gp1", "Gp2"),
                               nrow(dat_path_model),
                               replace = TRUE)
dat_path_model2 <- dat_path_model
dat_path_model2[1, 2] <- NA

mod <-
"
x3 ~ x1 + x2
x4 ~ x1
"

fit1 <- sem(mod, dat_path_model, fixed.x = TRUE)
fit2 <- sem(mod, dat_path_model, group = "group")
fit3 <- sem(mod, dat_path_model, estimator = "GLS")
fit4 <- sem(mod, dat_path_model, missing = "fiml")

expect_true(modelbpp:::check_sem_out(fit1))
expect_error(modelbpp:::check_sem_out(fit2), "groups")
expect_error(modelbpp:::check_sem_out(fit3), "estimator")
expect_error(model_set(fit2), "groups")
expect_error(model_set(fit3), "estimator")
expect_true(modelbpp:::check_sem_out(fit4))

suppressMessages(library(lavaan))

mod1 <-
"
x2 ~ x1 + x4
x3 ~ x2
"

mod2 <-
"
x3 ~ x1
x3 ~ x2
x4 ~ x3
"

fit1 <- sem(mod1, HolzingerSwineford1939, fixed.x = FALSE)
fit2 <- sem(mod2, HolzingerSwineford1939, fixed.x = FALSE)

out1 <- model_set(fit1,
                  progress = FALSE,
                  parallel = FALSE)
out2 <- model_set(fit2,
                  progress = FALSE,
                  parallel = FALSE)

expect_error(
    model_set_combined(
            list(out1,
                 out2),
            progress = FALSE,
            parallel = FALSE)
  )

expect_error(
    model_set_combined(
            list(a = out1,
                 out2),
            progress = FALSE,
            parallel = FALSE)
  )

if (TRUE &&
    interactive() &&
    length(unclass(packageVersion("modelbpp"))[[1]]) == 4) {

# Long tests

suppressMessages(library(lavaan))

data(dat_path_model)

mod1 <-
"
x2 ~ x1
x3 ~ x2
"
fit1_nofix <- sem(mod1, dat_path_model, fixed.x = FALSE)

mod2 <-
"
x2 ~ x3
x1 ~ x2
"
fit2_nofix <- sem(mod2, dat_path_model, fixed.x = FALSE)

mod3 <-
"
x2 ~~ x3
x1 ~ x2
"
fit3_nofix <- sem(mod3, dat_path_model, fixed.x = FALSE)

mod4 <-
"
x2 ~ x3
x2 ~ x1
x1 ~~ 0*x3
"
fit4_nofix <- sem(mod4, dat_path_model, fixed.x = FALSE)

mod5 <-
"
x2 ~ x3
x2 ~~ x1
x1 ~~ 0*x3
"
fit5_nofix <- sem(mod5, dat_path_model, fixed.x = FALSE)

mod6 <-
"
x3 ~ x1
x2 ~ x3
"
fit6_nofix <- sem(mod6, dat_path_model, fixed.x = FALSE)

out <- model_set(list(a = fit1_nofix,

                      b = fit2_nofix,
                      c = fit3_nofix,
                      d = fit4_nofix,
                      e = fit5_nofix,
                      f = fit6_nofix),
                 progress = FALSE)
expect_true(setequal(names(out$equivalent_clusters),
                     c("d", "a")))
expect_true(setequal(out$equivalent_clusters$d,
                     c("d", "e")))
expect_true(setequal(out$equivalent_clusters$a,
                     c("a", "b", "c")))
}

suppressMessages(library(lavaan))

# Target BPP

dat <- dat_path_model

mod <-
"
x2 ~ x3 + 0*x4
x1 ~ x3
"

fit <- sem(mod,
           dat_path_model,
           fixed.x = TRUE)

out <- model_set(fit,
                 progress = FALSE,
                 parallel = FALSE)

postprob0 <- .75

e <- exp(-.5 * (out$bic - out$bic[1]))
e1 <- e[5]
estar <- sum(e[-5])
k <- length(e)
p1 <- (1 / (k - 1)) * postprob0 * estar / (e1 * (1 - postprob0) + postprob0 * estar / (k - 1))

p1_out <- min_prior(out$bic,
                    bpp_target = .75)

out <- model_set(fit,
                 prior_sem_out = unname(p1),
                 progress = FALSE,
                 parallel = FALSE)

expect_equal(p1_out,
              unname(p1),
              info = "Target BPP")
expect_equal(unname(out$bpp["original"]),
              postprob0,
              info = "Target BPP")

#' @noRd

get_diff <- function(x, pt_no_user) {
  x_lor <- x[, c("lhs", "op", "rhs")]
  pt_lor <- pt_no_user[, c("lhs", "op", "rhs")]
  pt_lor$old <- TRUE
  # New parameters
  x_pt <- merge(x_lor, pt_lor, all.x = TRUE)
  x_new <- x_pt[is.na(x_pt$old), c("lhs", "op", "rhs")]
  if (nrow(x_new) > 0) {
      x_new_str <- apply(x_new, 1, paste, collapse = "")
    } else {
      x_new_str <- NULL
    }
  # Constraints released
  x_lor$mod <- FALSE
  x_pt_c <- merge(pt_lor, x_lor, all.x = TRUE)
  x_released <- x_pt_c[is.na(x_pt_c$mod), c("lhs", "op", "rhs")]
  if (nrow(x_released) > 0) {
      x_constr <- lapply(seq_len(nrow(x_released)),
                      function(z) c(x_released[z, 1],
                                    x_released[z, 2],
                                    x_released[z, 3]))
      x_constr_names <- sapply(x_constr, function(z, pt) {
          id_x <- which((pt$lhs == z[1]) & (pt$op == z[2]) & (pt$rhs == z[3]))
          i_lhs <- which(pt$plabel == pt[id_x, "lhs"])
          i_rhs <- which(pt$plabel == pt[id_x, "rhs"])
          p_lhs <- paste0("(", paste(pt[i_lhs, c("lhs", "op", "rhs")], collapse = ""), ")")
          p_rhs <- paste0("(", paste(pt[i_rhs, c("lhs", "op", "rhs")], collapse = ""), ")")
          paste0(p_lhs, ",", p_rhs)
        }, pt = pt_no_user)
    } else {
      x_constr_names <- NULL
    }
  out_str <- paste0(c(x_new_str, x_constr_names), collapse = ";")
  out_ops <- paste("add:", out_str)
  out_ops
}

#' @noRd

get_diff_drop <- function(x, pt_no_user) {
  tmp <- c("lhs", "op", "rhs", "block", "group", "label", "free")
  x0 <- x[, tmp]
  pt_no_user0 <- pt_no_user[, tmp]
  x0$free <- x0$free > 0
  pt_no_user0$free <- pt_no_user0$free > 0
  out <- x0[which(apply(x0 != pt_no_user0, 1, any, na.rm = TRUE)), ]
  out_ops <- sapply(seq_len(nrow(out)), function(x) {
      paste0(out[x, "lhs"], out[x, "op"], out[x, "rhs"], collapse = "")
    })
  out_ops <- paste(out_ops, collapse = ";")
  out_ops <- paste("drop:", out_ops)
}

#' @title Manipulate Parameter Tables
#'
#' @description Functions to manipulate
#' a `partables`-class object
#'
#' @name manipulate_partables
NULL

#'
#' @details
#' The `partables`-class objects have
#' a `c()` method that can be used to
#' combine parameter tables.
#' Each object must be
#'
#' a. a `partables`-class object,
#' b. a `model_set`-class object,
#' c. a `lavaan`-class object, or
#' d. a parameter table of
#' the class `lavaan.data.frame()`,
#' usually generated by
#' [lavaan::parameterTable()].
#' Other objects will be discarded.
#'
#' Names will be used when combining
#' objects. If two objects have the same
#' names, then only the first one will be
#' retained. No warning message will be
#' issued. Users are encouraged to
#' explicitly name all objects carefully.
#'
#' Note that, to invoke this method,
#' the first object must be a
#' `partables` object.
#'
#' @return
#' A `partables`-class objects with
#' all the objects supplied combined
#' together. If an object is
#' a `lavaan`-class object, its
#' parameter table will be retrieved
#' by [lavaan::parameterTable()].
#' If an object is a `model_set`-class
#' object, the stored `partables`-class
#' object will be retrieved.
#'
#' @param ... An arbitrary number of
#' objects. All invalid objects (see
#' details) will be discarded. If an
#' object is named and is not
#' `partables` object, its name will be
#' used.
#'
#' @param x A `partables`-class object.
#'
#' @param model_names A character vector
#' of the names of models in a
#' `partables`-class object.
#'
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>
#'
#' @examples
#'
#' library(lavaan)
#'
#' moda <-
#' "
#' x3 ~ a*x1 + b*x2
#' x4 ~ a*x1
#' ab := a*b
#' "
#'
#' fita <- sem(moda, dat_path_model, fixed.x = TRUE)
#'
#' outa <- model_set(fita,
#'                   progress = FALSE,
#'                   parallel = FALSE)
#'
#' modb <-
#' "
#' x3 ~ a*x1 + b*x2
#' x4 ~ a*x2
#' ab := a*b
#' "
#'
#' fitb <- sem(modb, dat_path_model, fixed.x = TRUE)
#'
#' outb <- model_set(fitb,
#'                   progress = FALSE,
#'                   parallel = FALSE)
#'
#' mod2 <-
#' "
#' x2 ~ 0*x3 + 0*x4
#' x1 ~ 0*x3
#' "
#' fit2 <- sem(mod2, dat_path_model)
#'
#' mod3 <-
#' "
#' x2 ~ x3 + 0*x4
#' x1 ~ x3
#' "
#' fit3 <- sem(mod3, dat_path_model)
#'
#' out <- c(outa$models, user2 = fit2, outb$models, user3 = fit3)
#' out
#'
#' out2 <- c(outa, user2 = fit2, outb$models, user3 = fit3)
#' out2
#'
#' out3 <- c(outa, user2 = fit2, outb, user3 = fit3)
#' out3
#'
#' @rdname manipulate_partables
#' @order 1
#' @export

c.partables <- function(...) {
    ddd <- list(...)
    ddd_names <- sapply(as.list(match.call()[-1]), deparse)
    if (!is.null(names(ddd_names))) {
        i <- which(sapply(names(ddd_names), function(x) {nchar(x) > 0}))
        if (length(i) > 0) {
            ddd_names[i] <- names(ddd_names)[i]
            ddd_names <- unname(ddd_names)
          }
      }
    is_partables <- sapply(ddd,
                           inherits,
                           what = "partables")
    is_lavaan <- sapply(ddd,
                        inherits,
                        what = "lavaan")
    is_ptable <- sapply(ddd,
                        inherits,
                        what = "lavaan.data.frame")
    is_model_set <- sapply(ddd,
                           inherits,
                           what = "model_set")
    if (any(is_lavaan)) {
        p <- which(is_lavaan)
        for (i in p) {
            tmp <- list(lavaan::parameterTable(ddd[[i]]))
            names(tmp) <- ddd_names[i]
            ddd[[i]] <- tmp
          }
      }
    if (any(is_ptable)) {
        p <- which(is_ptable)
        for (i in p) {
            tmp <- list(ddd[[i]])
            names(tmp) <- ddd_names[i]
            class(tmp) <- c("partables", class(tmp))
            ddd[[i]] <- tmp
          }
      }
    if (any(is_model_set)) {
        p <- which(is_model_set)
        for (i in p) {
            ddd[[i]] <- ddd[[i]]$models
          }
      }
    is_valid <- is_ptable |
                is_lavaan |
                is_partables |
                is_model_set
    ddd <- ddd[is_valid]
    out <- c_partables(ddd)
  }

#' @details
#' The `model_set` class also has a
#' `c`-method. It
#' will replace the first object by the stored
#' partables and then call the `c`-method of `partables` objects.
#'
#' @rdname manipulate_partables
#' @order 2
#' @export

c.model_set <- function(...) {
    ddd <- list(...)
    ddd[[1]] <- ddd[[1]]$models
    ddd_names <- sapply(as.list(match.call()[-1]), deparse)
    if (!is.null(names(ddd_names))) {
        i <- which(sapply(names(ddd_names), function(x) {nchar(x) > 0}))
        if (length(i) > 0) {
            ddd_names[i] <- names(ddd_names)[i]
            ddd_names <- unname(ddd_names)
          }
      }
    names(ddd) <- ddd_names
    do.call(c.partables, ddd)
  }

#' @details
#' The function [partables_drop()] is
#' for dropping models from a
#' `partables`-class object.
#'
#' @rdname manipulate_partables
#' @order 2
#'
#' @export

partables_drop <- function(x,
                           model_names = NULL) {
    if (!inherits(x, "partables")) {
        stop("The input is not a partables-class object.")
      }
    model_names <- intersect(model_names, names(x))
    if (length(model_names) == 0) {
        return(x)
      }
    x[model_names] <- NULL
    return(x)
  }


#' @noRd

c_partables <- function(ddd) {
    if (length(ddd) == 1) {
        return(ddd[[1]])
      }
    out <- ddd[[1]]
    for (i in seq_along(ddd)[-1]) {
        ddd_i <- ddd[[i]]
        i_names <- names(ddd_i)
        i_names <- setdiff(i_names, names(out))
        if (length(i_names) == 0) next
        for (j in i_names) {
            out[[j]] <- ddd_i[[j]]
          }
      }
    out
  }

#' @noRd

syntax_to_id <- function(x,
                         ptable) {
    x0 <- paste0(unlist(x), collapse = "\n")
    x_df0 <- lavaan::lavParseModelString(
                        model.syntax = x0,
                        as.data.frame. = TRUE,
                        warn = FALSE)
    x_df1 <- x_df0[, c("lhs", "op", "rhs")]
    x_df3 <- merge(x_df1, ptable)
    x_df3$id
  }

#' @noRd

syntax_to_add_list <- function(x) {
    x0 <- lavaan::lavParseModelString(
                        model.syntax = x,
                        as.data.frame. = TRUE,
                        warn = FALSE)
    x1 <- mapply(c,
                 lhs = x0$lhs,
                 op = x0$op,
                 rhs = x0$rhs,
                 USE.NAMES = FALSE,
                 SIMPLIFY = FALSE)
    x1
  }

#' @noRd

add_list_duplicate_cov <- function(x) {
    if (length(x) == 0) {
        return(x)
      }
    x0 <- lapply(x, function(y) {
                        if (y[2] == "~~") {
                            out <- y[3:1]
                            names(out) <- names(y)[1:3]
                            return(out)
                          }
                        NULL
                      })
    x0 <- x0[!sapply(x0, is.null)]
    x1 <- c(x, x0)
    x1
  }

#' @noRd

add_list_clean_duplicated_cov <- function(x) {
    if (length(x) == 0) {
        return(x)
      }
    x0 <- sapply(seq_len(length(x)),
                 function(y) {
                    y0 <- x[[y]]
                    if (y0[2] != "~~") return(TRUE)
                    for (j in seq_len(y)) {
                        if (identical(x[[j]], y0[3:1])) {
                            return(FALSE)
                          }
                      }
                    TRUE
                   })
    x[x0]
  }

#' @noRd

constr_to_lor <- function(x,
                          pt) {
    rowid1 <- (pt$plabel == x[1])
    rowid2 <- (pt$plabel == x[3])
    lor1 <- c(lhs = pt$lhs[rowid1],
              op = pt$op[rowid1],
              rhs = pt$rhs[rowid1])
    lor2 <- c(lhs = pt$lhs[rowid2],
              op = pt$op[rowid2],
              rhs = pt$rhs[rowid2])
    list(lor1, lor2)
  }

#' @noRd

constr_pars <- function(constr, pt) {
    out <- sapply(constr, function(z, pt) {
        id_x <- which((pt$lhs == z["lhs"]) &
                      (pt$op == z["op"]) &
                      (pt$rhs == z["rhs"]))
        i_lhs <- which(pt$plabel == pt[id_x, "lhs"])
        i_rhs <- which(pt$plabel == pt[id_x, "rhs"])
        p_lhs <- paste0("(",
                        paste(pt[i_lhs, c("lhs", "op", "rhs")],
                              collapse = ""),
                        ")")
        p_rhs <- paste0("(",
                        paste(pt[i_rhs, c("lhs", "op", "rhs")],
                              collapse = ""),
                        ")")
        paste0(p_lhs, ",", p_rhs)
      }, pt = pt)
    return(out)
  }

#' @noRd

par_names <- function(pars_list) {
    out <- sapply(pars_list, paste, collapse = "")
    out <- paste0(out, collapse = ";")
    return(out)
  }

#' @noRd

release_constr <- function(constr, pt) {
    for (j in constr) {
        i <- which((pt$lhs == j[1]) & (pt$op == j[2]) & (pt$rhs == j[3]))
        dot_p <- c(pt[i, "lhs"], pt[i, "rhs"])
        p_others <- unique(c(pt[-i, "lhs"], pt[-i, "rhs"]))
        dot_p_out <- dot_p[!(dot_p %in% p_others)]
        pt <- pt[-i, ]
        pt[pt$plabel %in% dot_p_out, "label"] <- ""
      }
    return(pt)
  }

#' @noRd

mt_exclude_existing_pars <- function(mt, pt) {
    if (nrow(mt) == 0) {
        return(mt)
      }
    # Remove those already in the parameter tables
    mt_in_pt1 <- mapply(function(x, y) {
                            any((pt$lhs == x) & (pt$rhs == y))
                          },
                        x = mt$lhs,
                        y = mt$rhs,
                        USE.NAMES = FALSE)
    mt_in_pt2 <- mapply(function(x, y) {
                            any((pt$lhs == x) & (pt$rhs == y))
                          },
                        x = mt$rhs,
                        y = mt$lhs,
                        USE.NAMES = FALSE)
    mt_exclude_in_pt2 <- mt_in_pt1 | mt_in_pt2
    out <- mt[!mt_exclude_in_pt2, ]
    return(out)
  }

#' @noRd

mt_exclude_reversed <- function(mt, pt) {
    if (nrow(mt) == 0) {
        return(mt)
      }
    user_v <- unique(c(pt$lhs[pt$user %in% c(1, 0)],
                       pt$rhs[pt$user %in% c(1, 0)]))
    i_iv <- user_v[!user_v %in% pt$lhs[pt$op == "~"]]
    mt_exclude_iv_to_dv <- (mt$op == "~") & (mt$lhs %in% i_iv)
    out <- mt[!mt_exclude_iv_to_dv, ]
    return(out)
  }

#' @noRd

mt_remove_error_cov <- function(mt_list, sem_out) {
    if (length(mt_list) == 0) {
        return(mt_list)
      }
    ind <- lavaan::lavNames(sem_out, "ov.ind")
    ind_cov <- sapply(mt_list, function(x, indnames) {
                        if ((x[1] %in% indnames) &&
                            (x[2] == "~~") &&
                            (x[3] %in% indnames)) {
                              return(FALSE)
                            } else {
                              return(TRUE)
                            }
                      }, indnames = ind)
    out <- mt_list[ind_cov]
    return(out)
  }

#' @noRd

lor_to_list <- function(x) {
    out <- mapply(c,
                  lhs = x[, "lhs"],
                  op = x[, "op"],
                  rhs = x[, "rhs"],
                  SIMPLIFY = FALSE,
                  USE.NAMES = FALSE)
    return(out)
  }

#' @noRd

sets_remove_inadmissible <- function(sets) {
    out <- lapply(sets, function(x) {
        i <- length(x)
        x_to_remove <- rep(FALSE, i)
        x_lr <- lapply(x, function(y) y[c("lhs", "rhs")])
        for (j in seq_len(i)) {
            if (j > 1) {
                for (k in seq_len(j - 1)) {
                  if (all(x_lr[[j]] == x_lr[[k]]) |
                      all(x_lr[[j]] == x_lr[[k]][c("rhs", "lhs")])) {
                      x_to_remove[j] <- TRUE
                    }
                }
              }
          }
        x[!x_to_remove]
      })
    return(out)
  }

#' @noRd

pt_remove_user_defined <- function(pt,
                                   remove_constrained = TRUE,
                                   return_id = FALSE) {
    i1 <- pt$op == ":="
    i1_labels <- pt[i1, "lhs"]
    i2 <- (pt$lhs %in% i1_labels) & (pt$op == "==")
    if (remove_constrained) {
        i0 <- i1 | i2
      } else {
        i0 <- i1 & !i2
      }
    if (return_id) {
        return(!i0)
      } else {
        return(pt[!i0, ])
      }
  }

#' @noRd

pt_remove_constrained_equal <- function(pt,
                                        return_id = FALSE) {
    i1 <- pt$op == "=="
    id_eq <- c(pt$lhs[i1], pt$rhs[i1])
    id_eq <- unique(id_eq)
    i0 <- pt$plabel %in% id_eq
    if (return_id) {
        return(!i0)
      } else {
        return(pt[!i0, ])
      }
  }

#' @noRd

pt_remove_exo_var <- function(pt,
                              return_id = FALSE) {
    i1 <- (pt$exo > 0) & (pt$op == "~~")
    i2 <- (pt$lhs == pt$rhs)
    i0 <- i1 & i2
    if (return_id) {
        return(!i0)
      } else {
        return(pt[!i0, ])
      }
  }

#' @noRd

pt_remove_end_var <- function(pt,
                              return_id = FALSE) {
    i1 <- (pt$exo == 0) & (pt$op == "~~")
    i2 <- (pt$lhs == pt$rhs)
    i0 <- i1 & i2
    if (return_id) {
        return(!i0)
      } else {
        return(pt[!i0, ])
      }
  }

#' @noRd

feedback_and_xy_cov <- function(sem_out) {
    mod_all_paths_org <- suppressWarnings(manymome::all_indirect_paths(sem_out))
    if (length(mod_all_paths_org) == 0) {
        return(list())
      }
    mod_all_paths <- manymome::all_paths_to_df(mod_all_paths_org)
    all_feedback <- data.frame(lhs = mod_all_paths$x,
                               op = "~",
                               rhs = mod_all_paths$y)
    all_feedback <- unique(all_feedback)
    all_xy_cov <- data.frame(lhs = mod_all_paths$x,
                             op = "~~",
                             rhs = mod_all_paths$y)
    all_xy_cov <- unique(all_xy_cov)
    out <- list(all_feedback = all_feedback,
                all_xy_cov = all_xy_cov)
    out
  }

#' @noRd

df_to_lor <- function(object) {
    if (nrow(object) == 0) {
        return(character(0))
      }
    out <- mapply(function(a1, a2, a3) {
                      c(lhs = a1,
                        op = a2,
                        rhs = a3)
                    },
                    a1 = object$lhs,
                    a2 = object$op,
                    a3 = object$rhs,
                    SIMPLIFY = FALSE,
                    USE.NAMES = FALSE)
    out
  }

#' @noRd
# Identify cross loadings
cross_loadings <- function(pt,
                           free_only = TRUE,
                           return_id = FALSE) {
  # TODO:
  # - Handle multigroup models
  i1 <- pt$op == "=~"
  i2 <- duplicated(pt$rhs)
  i3 <- i1 & i2
  i4 <- pt$rhs[i3]
  i5 <- (pt$op == "=~") & (pt$rhs %in% i4)
  if (free_only) {
    i6 <- pt$free > 0
  } else {
    i6 <- rep(TRUE, nrow(pt))
  }
  i5 <- i5 & i6
  if (return_id) {
    return(i5)
  } else {
    return(pt[i5, ])
  }
}

#' @noRd
# Identify all loadings
all_loadings <- function(pt,
                         free_only = TRUE,
                         return_id = FALSE) {
  # TODO:
  # - Handle multigroup models
  i1 <- pt$op == "=~"
  if (free_only) {
    i6 <- pt$free > 0
  } else {
    i6 <- rep(TRUE, nrow(pt))
  }
  i1 <- i1 & i6
  if (return_id) {
    return(i1)
  } else {
    return(pt[i1, ])
  }
}

#' @noRd
# Identify single-factor loadings
pt_remove_all_single_loadings <- function(
  pt,
  return_id = FALSE
) {
  # TODO:
  # - Handle multigroup models
  # No need to handle fixed parameters.
  # get_drop() will do this.
  i1 <- pt$op == "=~"
  i2 <- duplicated(pt$rhs)
  i3 <- i1 & i2
  i4 <- pt$rhs[i3]
  i5 <- (pt$op == "=~") & (!(pt$rhs %in% i4))
  if (return_id) {
    return(!i5)
  } else {
    return(pt[!i5, ])
  }
}

#' @noRd
# Identify all factor loadings
pt_remove_all_loadings <- function(
  pt,
  return_id = FALSE
) {
  # TODO:
  # - Handle multigroup models
  # No need to handle fixed parameters.
  # get_drop() will do this.
  i1 <- pt$op == "=~"
  if (return_id) {
    return(!i1)
  } else {
    return(pt[!i1, ])
  }
}

suppressMessages(library(lavaan))

out1 <- modelbpp:::syntax_to_add_list(c("m3 ~ m4", "x4 ~ x3", "f0 =~ x1"))
out2 <- modelbpp:::syntax_to_add_list(list("m3 ~ m4", "x4 ~ x3", "f0 =~ x1"))
out3 <- modelbpp:::syntax_to_add_list("m3 ~ m4\nx4 ~ x3\nf0 =~ x1")

out1_check <- list(c("m3", "~", "m4"),
                   c("x4", "~", "x3"),
                   c("f0", "=~", "x1"))

out1_check <- lapply(out1_check,
                      function(x) {
                          names(x) <- c("lhs", "op", "rhs")
                          x
                        })

expect_equal(out1,
              out1_check,
              info = "modelbpp:::syntax_to_add_list")
expect_equal(out2,
              out1_check,
              info = "modelbpp:::syntax_to_add_list")
expect_equal(out3,
              out1_check,
              info = "modelbpp:::syntax_to_add_list")

suppressMessages(library(lavaan))

dat <- dat_path_model

mod <-
"
x1 ~ x2
x3 ~ x1
x4 ~ x3 + x2
"

fit <- sem(mod,
           dat_path_model,
           meanstructure = TRUE,
           fixed.x = TRUE)

mod2 <-
"
f1 =~ x4 + x2 + x3
"

fit2 <- cfa(mod2,
            dat_path_model,
            meanstructure = TRUE,
            fixed.x = TRUE)

out <- modelbpp:::feedback_and_xy_cov(fit)
out2 <- modelbpp:::feedback_and_xy_cov(fit2)

feedback_check <- structure(list(lhs = c("x1", "x2", "x2"), op = c("~", "~", "~"
), rhs = c("x4", "x3", "x4")), row.names = c(NA, 3L), class = "data.frame")

xy_cov_check <- structure(list(lhs = c("x1", "x2", "x2"), op = c("~~", "~~",
"~~"), rhs = c("x4", "x3", "x4")), row.names = c(NA, 3L), class = "data.frame")

expect_true(setequal(out$all_feedback,
                     feedback_check),
            info = "modelbpp:::feedback_and_xy_cov")
expect_true(setequal(out$all_xy_cov,
                     xy_cov_check),
            info = "modelbpp:::feedback_and_xy_cov")

lor_feedback_check <- list(c(lhs = "x1", op = "~", rhs = "x4"), c(lhs = "x2", op = "~",
rhs = "x3"), c(lhs = "x2", op = "~", rhs = "x4"))

lor_xy_cov_check <- list(c(lhs = "x1", op = "~~", rhs = "x4"), c(lhs = "x2", op = "~~",
rhs = "x3"), c(lhs = "x2", op = "~~", rhs = "x4"))

expect_true(setequal(modelbpp:::df_to_lor(out$all_feedback),
                     lor_feedback_check),
            info = "modelbpp:::feedback_and_xy_cov")
expect_true(setequal(modelbpp:::df_to_lor(out$all_xy_cov),
                     lor_xy_cov_check),
            info = "modelbpp:::feedback_and_xy_cov")

out_add_fb <- get_add(fit, exclude_feedback = TRUE)
names(out_add_fb)
out2_add_fb <- get_add(fit2, exclude_feedback = TRUE)
names(out2_add_fb)

out_add_xy_cov <- get_add(fit, exclude_xy_cov = TRUE)
names(out_add_xy_cov)
out2_add_xy_cov <- get_add(fit2, exclude_xy_cov = TRUE)
names(out2_add_xy_cov)

out_add_both <- get_add(fit, exclude_xy_cov = TRUE,
                             exclude_feedback = TRUE)
names(out_add_both)
out2_add_both <- get_add(fit2, exclude_xy_cov = TRUE,
                             exclude_feedback = TRUE)
names(out2_add_both)

expect_false("add: x1~x4" %in% names(out_add_fb),
            info = "modelbpp:::feedback_and_xy_cov")
expect_false("add: x1~~x4" %in% names(out_add_xy_cov),
            info = "modelbpp:::feedback_and_xy_cov")
expect_true("add: x1~~x4" %in% names(out_add_fb),
            info = "modelbpp:::feedback_and_xy_cov")
expect_true("add: x1~x4" %in% names(out_add_xy_cov),
            info = "modelbpp:::feedback_and_xy_cov")
expect_false("add: x1~x4" %in% names(out_add_both),
            info = "modelbpp:::feedback_and_xy_cov")
expect_false("add: x1~~x4" %in% names(out_add_both),
            info = "modelbpp:::feedback_and_xy_cov")

suppressMessages(library(lavaan))

dat <- dat_path_model

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod,
           dat_path_model,
           meanstructure = TRUE,
           fixed.x = TRUE)

ptable <- parameterTable(fit)

out1 <- modelbpp:::syntax_to_id(list("x3 ~ x1", "x3 ~~ x4", "x3 ~1",
                          "m1 ~~ m2"),
                     ptable = ptable)

out2 <- modelbpp:::syntax_to_id(list("m3 ~ m71",
                          "m1 ~~ m2"),
                     ptable = ptable)

expect_equal(sort(out1),
              c(1, 6, 10),
              info = "modelbpp:::syntax_to_id")
expect_true(length(out2) == 0,
            info = "modelbpp:::syntax_to_id")

#' @title Print an `sem_outs`-Class Object
#'
#' @description Print the content of
#' an `sem_outs`-class object.
#'
#' @details
#' The print method for the output
#' of `fit_many()`.
#'
#' @return
#' `x` is returned invisibly.
#' Called for its side effect.
#'
#' @param x An `sem_outs`-class object.
#'
#' @param max_models The maximum number
#' of models to be printed. Default
#' is 20.
#'
#' @param ...  Optional arguments.
#' Ignored.
#'
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>
#'
#' @seealso An `sem_outs`-class object
#' is generated by `fit_many()`.
#'
#' @examples
#'
#' library(lavaan)
#' dat <- dat_path_model
#' mod <-
#' "
#' x3 ~ a*x1 + b*x2
#' x4 ~ a*x1
#' ab := a*b
#' "
#' fit <- sem(mod, dat_path_model, fixed.x = TRUE)
#' mod_to_add <- get_add(fit)
#' out <- fit_many(mod_to_add, fit)
#' out
#' print(out, max_models = 1)
#'
#' @export

print.sem_outs <- function(x,
                           max_models = 20,
                           ...) {
    fit_n <- length(x$fit)
    fit_names <- names(x$fit)
    x_call <- x$call
    k_converged <- sum(sapply(x$converged, isTRUE))
    k_post_check <- sum(sapply(x$post_check, isTRUE))
    cat("\n")
    cat("Call:\n")
    print(x_call)
    cat("\n")
    cat("Number of model(s) fitted: ", fit_n, "\n", sep = "")
    cat("Number of model(s) converged: ", k_converged, "\n", sep = "")
    cat("Number of model(s) passed post.check: ", k_post_check, "\n", sep = "")
    cat("\n")
    if (fit_n > max_models) {
        x_tmp <- fit_names[seq_len(max_models)]
        cat("The first", max_models, "modifications(s)/model(s):\n")
      } else {
        x_tmp <- fit_names
        cat("The modification(s)/model(s):\n")
      }
    cat(paste(x_tmp, collapse = "\n"), "\n")
    invisible(x)
  }

#' @title Models That Are Less Restricted
#'
#' @description Generate a list of
#' models with one or more fixed
#' parameter freed.
#'
#' @details It generates a list of models
#' with one or more fixed parameter
#' freed (and the degrees of freedom, *df*,
#' increases by one or more). If a
#' model has equality constraints, models
#' with one or more of the constraints
#' between two free parameters
#' released will also be included.
#'
#' Graphically, paths or covariances
#' are "added" to form the list of models.
#'
#' The models to be included are
#' identified by
#' [lavaan::modificationIndices()].
#'
#' The models will be checked by `lavaan`
#' to make sure that the decrease in
#' model degrees of freedom is of the
#' expected value.
#'
#' This function is called by
#' [model_set()] and usually users do
#' not need to call it. It is exported
#' for advanced users.
#'
#' @param sem_out The original model,
#' which is the output from an
#' structural equation modeling
#' function. Currently support
#' [lavaan::lavaan-class] objects only.
#'
#' @param must_add A character vector
#' of parameters, named in
#' [lavaan::lavaan()] style (e.g.,
#' `"y ~ x"`), that must be added.
#' Default is `NULL`.
#'
#' @param must_not_add A character
#' vector of parameters, named in
#' [lavaan::lavaan()] style (e.g.,
#' `"x1 ~~ x1"`), that must not be
#' added. Default is `NULL`.
#'
#' @param remove_constraints Whether
#' equality constraints will be
#' removed. Default is `TRUE`.
#'
#' @param exclude_error_cov Exclude
#' error covariances of indicators.
#' Default is `TRUE`.
#'
#' @param exclude_feedback Exclude
#' paths that will result in a feedback
#' loop. For example, if there is
#' path from `x` through `m` to `y`,
#' then the path `x ~ y` will create
#' a feedback loop. Default is `FALSE`
#' for now, to maintain backward
#' compatibility.
#' Do not rely on the default value
#' because it will be changed to `TRUE`
#' in a future major version.
#'
#' @param exclude_xy_cov Exclude
#' covariance between two variables,
#' in which one has a path to another.
#' For example, if there is
#' path from `x` through `m` to `y`,
#' then the covariance `x ~~ y`,
#' which denotes the covariance between
#' `x` and the error term of `y`, will
#' be excluded if this argument is
#' `TRUE`. Default is `FALSE` for now,
#' to maintain backward compatibility.
#' Do not rely on the default value
#' because it will be changed to `TRUE`
#' in a future major version.
#'
#' @param exclude_x_changed_to_y Exclude
#' changes that make a "pure" `x`
#' variable to a `y` variable. For example,
#' adding the path `x2 ~ x1` to
#' the model `y ~ x1 + x2`.
#'
#' @param cross_add A character vector
#' of whether and how cross-loadings
#' (an indicator loads on two or more
#' latent factors) will be added. If
#' `NULL`, no cross-loadings will be
#' added. If `"lav_x"` is in the vector,
#' then cross-loadings among "pure-x"
#' latent factors, latent
#' factors that do not regress on
#' any other variables, will be considered.
#' If `"lav_y"` is in the vector, then
#' cross-loadings among "pure-y" latent
#' factors, latent factors that regress
#' on at least one variables but they
#' themselves do not predict any other
#' variables, will be considered.
#' If `"user"` is in the vector, then
#' cross-loadings among latent factors listed
#' in `cross_sets` will be considered.
#'
#' @param cross_sets A character vector
#' of latent variables for which cross-loadings
#' will be considered. Ignored if
#' `"user"` is not in `cross_add`.
#' Note that cross-loadings involving latent
#' variables which are mediators will
#' not be considered. To add them,
#' they must be specified explicitly
#' in `must_add`.
#'
#' @param df_change How many degrees
#' of freedom (*df*) away in the list.
#' All models with *df* change less than
#' or equal to this number will be
#' included, taking into account
#' requirements set by other arguments.
#' Default is 1.
#'
#' @param remove_duplicated If `TRUE`,
#' the default, duplicated models are
#' removed.
#'
#' @param model_id The identification
#' number of the starting model.
#' Default is `NA`, no identification
#' number.
#'
#' @param keep_correct_df_change Keep
#' only models with actual *df* change
#' equal to expected *df* change.
#'
#' @param progress Whether a progress
#' bar will be displayed, implemented
#' by the `pbapply` package. Default
#' is `FALSE`.
#'
#' @return An object of the class
#' `partables`, a named list of parameter
#' tables, each of them to be used by
#' [lavaan::lavaan()] or [update()]
#' for fitting a model with the added
#' parameters.
#'
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>
#'
#' @seealso [print.partables()]
#'
#' @examples
#'
#' library(lavaan)
#' dat <- dat_path_model
#' mod <-
#' "
#' x3 ~ a*x1 + b*x2
#' x4 ~ a*x1
#' ab := a*b
#' "
#' fit <- sem(mod, dat_path_model, fixed.x = TRUE)
#' mod_to_add <- get_add(fit)
#' mod_to_add
#'
#' @export

get_add <- function(sem_out,
                     must_add = NULL,
                     must_not_add = NULL,
                     remove_constraints = TRUE,
                     exclude_error_cov = TRUE,
                     exclude_feedback = FALSE,
                     exclude_xy_cov = FALSE,
                     exclude_x_changed_to_y = TRUE,
                     cross_add = c("pure_x", "pure_y"),
                     cross_sets = NULL,
                     df_change = 1,
                     model_id = NA,
                     keep_correct_df_change = TRUE,
                     remove_duplicated = TRUE,
                     progress = FALSE
                    ) {
    if (!is.null(cross_add)) {
      cross_add <- match.arg(cross_add,
                             c("pure_x", "pure_y", "user"),
                             several.ok = TRUE)
    }
    if (missing(sem_out)) stop("sem_out is not supplied.")
    if (!inherits(sem_out, "lavaan")) {
        stop("sem_out is not a lavaan-class object.")
      }
    pt <- lavaan::parameterTable(sem_out)

    # ==== Remove all user-defined parameters ====
    pt <- pt[pt$op != ":=", ]

    # ==== Get the MI table ====
    # Suppress the warning about equality constraints.
    mt <- suppressWarnings(lavaan::modificationIndices(sem_out,
                    standardized = FALSE,
                    power = FALSE))

    # ==== Remove those already in the parameter tables ====
    mt1 <- mt_exclude_existing_pars(mt = mt, pt = pt)

    # ==== Remove those convert an IV to a DV ====
    if (exclude_x_changed_to_y) {
      mt1 <- mt_exclude_reversed(mt = mt1, pt = pt)
    }

    # ==== Identify parameters to be added ====
    mt1_op2 <- lor_to_list(mt1)

    # ==== Cross-loadings ====

    # Default to remove cross-loadings,
    # and add them back only if explicitly specified
    loadings <- sapply(
                    mt1_op2,
                    function(x) x["op"] == "=~"
                  )
    if (any(loadings)) {
      mt1_op2_no_loadings <- mt1_op2[!loadings]
      mt2_op2_loadings <- mt1_op2[loadings]
      ix <- list()
      iy <- list()
      iuser <- list()
      if (!is.null(cross_add) || isTRUE(length(cross_add) > 0)) {
        if ("pure_x" %in% cross_add) {
          # ==== Extract pure_x cross loadings ====
          tmp <- lav_cross_set(
                  sem_out,
                  type = "pure_x"
                )
          tmp <- lor_to_list(tmp)
          ix <- intersect(
                    tmp,
                    mt2_op2_loadings
                  )
        }
        if ("pure_y" %in% cross_add) {
          # ==== Extract pure_y cross loadings ====
          tmp <- lav_cross_set(
                  sem_out,
                  type = "pure_y"
                )
          tmp <- lor_to_list(tmp)
          iy <- intersect(
                    tmp,
                    mt2_op2_loadings
                  )
        }
        if ("user" %in% cross_add) {
          if (is.null(cross_sets)) {
            stop("cross_sets must be specified if 'user' is in cross_add")
          }
          # ==== Extract user cross loadings ====
          tmp <- lav_cross_set(
                  sem_out,
                  type = "user",
                  user_lav = cross_sets
                )
          tmp <- lor_to_list(tmp)
          iuser <- intersect(
                    tmp,
                    mt2_op2_loadings
                  )
        }
      }
      # Add cross-loadings, if any
      mt1_op2 <- c(
                    mt1_op2_no_loadings,
                    ix,
                    iy,
                    iuser
                  )
    }

    # ==== Add must_add ====
    if (!is.null(must_add)) {
        mt1_must_add <- syntax_to_add_list(must_add)
        mt1_op2 <- union(mt1_op2, mt1_must_add)
      }

    # ==== Remove feedback and xy_cov ====
    if (exclude_feedback || exclude_xy_cov) {
        tmp <- feedback_and_xy_cov(sem_out)
        if (length(tmp) != 0) {
          tmp1 <- df_to_lor(tmp$all_feedback)
          tmp1 <- sapply(tmp1, paste, collapse = " ")
          tmp2 <- df_to_lor(tmp$all_xy_cov)
          tmp2 <- sapply(tmp2, paste, collapse = " ")
          if (is.null(must_not_add)) {
              must_not_add <- character(0)
            }
          if (exclude_feedback) {
              must_not_add <- c(must_not_add,
                                tmp1)
            }
          if (exclude_xy_cov) {
              must_not_add <- c(must_not_add,
                                tmp2)
            }
        }
      }

    # ==== Remove must_not_add ====
    if (!is.null(must_not_add)) {
        mt1_must_not_add <- syntax_to_add_list(must_not_add)
        mt1_must_not_add1 <- add_list_duplicate_cov(mt1_must_not_add)
        mt1_op2 <- setdiff(mt1_op2, mt1_must_not_add1)
      }
    # ==== Remove duplicated covariances ====
    mt1_op2 <- add_list_clean_duplicated_cov(mt1_op2)

    # ==== Remove error covariances between indicators ====
    if (exclude_error_cov) {
        mt1_op2 <- mt_remove_error_cov(mt_list = mt1_op2,
                                       sem_out = sem_out)
      }

    # ==== Identify parameters constrained to be equal by labels ====
    i_eq <- pt$op == "=="
    row_eq <- which(i_eq)
    row_eq_op2 <- lor_to_list(pt[row_eq, ])

    # ==== Fix the modifications  ====
    if (length(c(mt1_op2, row_eq_op2)) != 0) {
        # ==== Determine the sets of changes ====
        sets_to_gen2 <- lapply(seq_len(df_change),
                    function(x) {
                        utils::combn(c(mt1_op2, row_eq_op2), x, simplify = FALSE)
                      }
                  )
        sets_to_gen2 <- unlist(sets_to_gen2, recursive = FALSE)

        # ==== Clean up inadmissible modification ====
        sets_to_gen2_ok <- sets_remove_inadmissible(sets_to_gen2)

        # ==== Generate parameter tables ====
        if (progress) {
            cat("\nGenerate", length(sets_to_gen2_ok), "less restrictive model(s):\n")
            op_old <- pbapply::pboptions(type = "timer")
            out <- suppressWarnings(
                      pbapply::pblapply(
                          sets_to_gen2_ok,
                          gen_pt_add,
                          pt = pt,
                          sem_out = sem_out,
                          from = model_id
                        )
                      )
            pbapply::pboptions(op_old)
          } else {
            out <- suppressWarnings(
                      lapply(
                        sets_to_gen2_ok,
                        gen_pt_add,
                        pt = pt,
                        sem_out = sem_out,
                        from = model_id
                      )
                    )
          }
      } else {
        out <- list()
      }

    # ==== Keep tables with expected df only? ====
    if (keep_correct_df_change) {
        chk1 <- sapply(out, attr, which = "df_actual")
        chk2 <- sapply(out, attr, which = "df_expected")
        out <- out[chk1 == chk2]
      }

    out_names <- sapply(out, function(x) {
        out <- paste0(c(attr(x, "parameters_added_str"),
              attr(x, "constraints_released_str")), collapse = ";")
        paste("add:", out)
      })
    names(out) <- out_names
    attr(out, "call") <- match.call()
    attr(out, "sem_out") <- sem_out
    class(out) <- c("partables", class(out))
    if (remove_duplicated) {
        out <- unique_models(out, progress = progress)
      }
    out
  }

#' @noRd

gen_pt_add <- function(x, pt, sem_out, from = NA) {
    # Generate pt
    # Collects free parameters to add
    x_free <- x[sapply(x,
                       function(y) {y["op"] != "=="})]
    # Collects constraints to release
    x_constr <- x[sapply(x,
                         function(y) {y["op"] == "=="})]
    # Form constraint in parameter names (lhs-op-rhs)
    x_constr_pars <- constr_pars(constr = x_constr,
                                 pt = pt)
    # Loop and release constraints
    if (length(x_constr) > 0) {
        pt <- release_constr(x_constr, pt = pt)
        x_constr_labels <- paste0(sapply(x_constr,
                                         paste,
                                         collapse = ""),
                                  collapse = ";")
        x_constr_str <- paste(x_constr_pars, collapse = ";")
        x_constr_out <- lapply(x_constr,
                               constr_to_lor,
                               pt = pt)
      } else {
        x_constr_labels <- NULL
        x_constr_str <- NULL
        x_constr_out <- NULL
      }
    # Add free parameters
    if (length(x_free) > 0) {
        x_free_str <- par_names(pars_list = x_free)
        p_to_add <- sapply(x_free, paste0, collapse = "")
        sem_out_update <- lavaan::update(sem_out,
                                         pt,
                                         add = x_free_str,
                                         do.fit = TRUE,
                                         optim.force.converged = TRUE,
                                         control = list(max.iter = 1))
        pt_update <- lavaan::parameterTable(sem_out_update)
        pt_update$se <- NA
        pt_update_df <- unname(lavaan::fitMeasures(sem_out_update,
                                                   fit.measures = "df"))
      } else {
        x_free_str <- NULL
        p_to_add <- NULL
        sem_out_update <- lavaan::update(sem_out,
                                         pt,
                                         do.fit = TRUE,
                                         optim.force.converged = TRUE,
                                         control = list(max.iter = 1))
        pt_update <- lavaan::parameterTable(sem_out_update)
        pt_update$se <- NA
        pt_update_df <- unname(lavaan::fitMeasures(sem_out_update,
                                                   fit.measures = "df"))
      }
    attr(pt_update, "parameters_added") <- p_to_add
    attr(pt_update, "parameters_added_str") <- x_free_str
    attr(pt_update, "parameters_added_list") <- x_free
    attr(pt_update, "constraints_released_labels") <- x_constr_labels
    attr(pt_update, "constraints_released_str") <- x_constr_str
    attr(pt_update, "constraints_released") <- x_constr_pars
    attr(pt_update, "constraints_released_list") <- x_constr_out
    attr(pt_update, "from") <- from
    attr(pt_update, "df_expected") <- unname(lavaan::fitMeasures(sem_out, "df")) -
                                      length(x)
    attr(pt_update, "df_actual") <- pt_update_df
    pt_update
  }

suppressMessages(library(lavaan))
mod <-
"
fx1 =~ x1 + x2 + x3
fx2 =~ x4 + x5 + x6
fm1 =~ x7 + x8 + x9
fy2 =~ x10 + x11 + x12
fy3 =~ x13 + x14 + x15
fm1 ~ fx1 + fx2
fy2 ~ fm1
fy3 ~ fm1
"

# set.seed(2134)
# n <- nrow(dat_sem)
# fy4 <- rnorm(n)
# dat_sem$x16 <- .6 * fy4 + rnorm(n)
# dat_sem$x17 <- .6 * fy4 + rnorm(n)
# dat_sem$x18 <- .6 * fy4 + rnorm(n)

fit <- sem(mod,
           dat_sem,
           warn = FALSE)

chk_pure_x <- "add: fx1=~x4"
chk_pure_y <- "add: fy2=~x14"

mod_to_add_a <- get_add(fit, cross_add = c("pure_y", "pure_x"))
expect_true(chk_pure_x %in% names(mod_to_add_a))
expect_true(chk_pure_y %in% names(mod_to_add_a))

mod_to_add_b <- get_add(fit, cross_add = NULL)
expect_false(chk_pure_x %in% names(mod_to_add_b))
expect_false(chk_pure_y %in% names(mod_to_add_b))

mod_to_add_c <- get_add(fit, cross_add = "pure_x")
expect_true(chk_pure_x %in% names(mod_to_add_c))
expect_false(chk_pure_y %in% names(mod_to_add_c))

mod_to_add_d <- get_add(fit, cross_add = "pure_y")
expect_false(chk_pure_x %in% names(mod_to_add_d))
expect_true(chk_pure_y %in% names(mod_to_add_d))

mod_to_add_e <- get_add(fit, cross_add = "user", cross_sets = c("fx1", "fy2"))
expect_true("add: fx1=~x11" %in% names(mod_to_add_e))
expect_true("add: fy2=~x1" %in% names(mod_to_add_e))

mod_to_add_f <- get_add(fit, cross_add = NULL,
                        must_add = "fm1 =~ x1")
expect_true("add: fm1=~x1" %in% names(mod_to_add_f))

if (interactive() &&
    length(unclass(packageVersion("modelbpp"))[[1]]) == 4) {

    suppressMessages(library(lavaan))

    dat <- dat_path_model_p06

    mod <- attr(dat, "mod")

    fit <- sem(mod,
               dat_path_model_p06,
               fixed.x = TRUE)
    pt <- parameterTable(fit)
    pt_no_user <- pt[pt$op != ":=", ]
    pt0 <- parameterTable(fit)
    mod_to_add <- get_add(fit)
    fit_add <- lapply(mod_to_add,
                      function(x) update(fit, x))
    anova_add <- lapply(fit_add,
                        function(x) anova(x, fit))

    expect_true(
        all(names(mod_to_add) %in%
          c("add: y4~~y6", "add: y4~y6", "add: y5~x2", "add: y6~y4", "add: y6~x1",
            "add: y6~x2", "add: (y4~x1),(y4~x2)")),
        info = "Parameters to add as expected"
      )

    expect_true(
        all(sapply(anova_add,
                   function(x) x[2, "Df diff"]) == 1),
        info = "All df differences are one"
      )

    expect_true(
        all(sapply(mod_to_add,
                   modelbpp:::get_diff, pt_no_user) == names(mod_to_add)),
        info = "Generated difference matches the names"
      )
  }

suppressMessages(library(lavaan))

dat <- dat_path_model

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)
pt <- parameterTable(fit)
pt_no_user <- pt[pt$op != ":=", ]
mod_to_add <- get_add(fit)
fit_add <- lapply(mod_to_add,
                  function(x) update(fit, x))
anova_add <- lapply(fit_add,
                    function(x) anova(x, fit))

expect_true(
    all(names(mod_to_add) %in% c("add: x4~x2", "add: (x3~x1),(x4~x1)")),
    info = "Parameters to drop as expected"
  )

expect_true(
    all(sapply(anova_add,
               function(x) x[2, "Df diff"]) == 1),
    info = "All df differences are one"
  )

expect_true(
    all(sapply(mod_to_add,
               modelbpp:::get_diff, pt_no_user) == names(mod_to_add)),
    info = "Generated difference matches the names"
  )

mod_to_add <- modelbpp:::get_add(fit,
                                 df_change = 2)
fit_add <- lapply(mod_to_add,
                  function(x) update(fit, x))
anova_add <- lapply(fit_add,
                    function(x) anova(fit, x))

expect_true(
    all(names(mod_to_add) %in%
      c("add: x4~x2", "add: (x3~x1),(x4~x1)", "add: x4~x2;(x3~x1),(x4~x1)")),
    info = "Parameters to drop as expected"
  )

expect_true(
    all(sapply(anova_add,
               function(x) x[2, "Df diff"]) ==
        c(1, 1, 2)),
    info = "All df differences are one"
  )

expect_true(
    all(sapply(mod_to_add,
               modelbpp:::get_diff, pt_no_user) == names(mod_to_add)),
    info = "Generated difference matches the names"
  )

# Test Print

expect_stdout(print(mod_to_add),
              "$`add: x4~x2;(x3~x1),(x4~x1)`",
              fixed = TRUE)

#' @title Set Layout Based on Model Degrees of Freedom
#'
#' @noRd

layer_by_df <- function(g,
                        model_set_out) {
  if (!is.null(model_set_out$model_df)) {
      model_dfs <- model_set_out$model_df
    } else {
      if (!all(model_set_out$converged)) {
          stop("Not all models converged.")
        } else {
          model_dfs <- sapply(model_set_out$fit,
                              lavaan::fitMeasures,
                              fit.measures = "df")
        }
    }
  names(model_dfs) <- names(model_set_out$fit)
  tmp <- sort(unique(model_dfs),
              decreasing = TRUE)
  model_layer <- model_dfs
  for (i in seq_along(tmp)) {
      model_layer[which(model_dfs == tmp[i])] <- i
    }
  model_layer <- model_layer - 1
  out <- igraph::add_layout_(g,
                             igraph::with_sugiyama(layers = model_layer))
  out
}

#' @title Identify All Simple Paths Between Two Models
#'
#' @noRd

all_simples <- function(i, j, graph) {
    igraph::all_simple_paths(graph = graph,
                             from = i,
                             to = j,
                             mode = "out")
  }

#' @title Identify All Simple Paths In A Network
#'
#' @return
#' A list of simple paths list, one for each
#' combination of models.
#'
#' @noRd

all_simples_list <- function(graph) {
    p <- length(igraph::V(graph))
    tmp <- utils::combn(p, 2)
    ij <- cbind(tmp, tmp[2:1, ])
    i0 <- ij[1, ]
    j0 <- ij[2, ]
    simples <- mapply(all_simples,
                      i = i0,
                      j = j0,
                      MoreArgs = list(graph = graph),
                      SIMPLIFY = FALSE)
    simples
  }

#' @title Delete A Redundant Direct Path
#'
#' @description
#' Check the path between two models. If
#' They can be connected through any
#' other models, then the direct path
#' is removed.
#'
#' @return
#' A `igraph`, with the redundant path,
#' if any, removed.
#'
#' @noRd

delete_redundant_direct <- function(x,
                                    graph) {
    if (length(x) <= 1) {
        return(graph)
      }
    x_length <- sapply(x, length)
    if (all(x_length == 2)) {
        return(graph)
      }
    i <- which(x_length == 2)
    if (length(i) == 0) {
        return(graph)
      }
    out <- graph
    for (xx in i) {
        out <- igraph::delete_edges(out,
                  paste(igraph::as_ids(x[[xx]]),
                        collapse = "|"))
      }
    out
  }

#' @title Delete All Redundant Direct Paths
#'
#' @description
#' Call `delete_redundant_direct()` to delete
#' the redundant direct path, if any.
#'
#' @return
#' A `igraph`, with redundant paths removed.
#'
#' @noRd

delete_all_redundant_direct <- function(graph) {
    graph_simples <- all_simples_list(graph)
    out <- graph
    for (xx in graph_simples) {
        out <- delete_redundant_direct(xx,
                                       graph = out)
      }
    out
  }

#' @title Store Basic Info to A Graph
#'
#' @noRd

doc_model_graph <- function(object,
                            graph) {
    p_V <- length(igraph::V(graph))
    V_names <- names(igraph::V(graph))
    V_order <- match(names(object$models), V_names)
    if (!is.null(object$change)) {
        igraph::V(graph)$change <- object$change[V_order]
      } else {
        igraph::V(graph)$change <- rep(NA, p_V)
      }
    if (!is.null(object$converged)) {
        igraph::V(graph)$converged <- object$converged[V_order]
      } else {
        igraph::V(graph)$converged <- rep(NA, p_V)
      }
    if (!is.null(object$post_check)) {
        igraph::V(graph)$post_check <- object$post_check[V_order]
      } else {
        igraph::V(graph)$post_check <- rep(NA, p_V)
      }
    igraph::V(graph)$model_name <- names(object$models)[V_order]
    if (!is.null(object$bic)) {
        igraph::V(graph)$bic <- object$bic[V_order]
      } else {
        igraph::V(graph)$bic <- rep(NA, p_V)
      }
    if (!is.null(object$prior)) {
        igraph::V(graph)$prior <- object$prior[V_order]
      } else {
        igraph::V(graph)$prior <- rep(NA, p_V)
      }
    if (!is.null(object$bpp)) {
        igraph::V(graph)$bpp <- object$bpp[V_order]
      } else {
        igraph::V(graph)$bpp <- rep(NA, p_V)
      }
    graph
  }

#' @title Label Arrows By 'df' Differences
#'
#' @noRd

label_by_df <- function(graph,
                              mode) {
    mode0 <- mode
    graph_df <- igraph::E(graph)$df
    if (is.null(mode)) {
        any_gt_1 <- (length(setdiff(graph_df, c(0, 1))) > 0)
        if (any_gt_1) {
            mode0 <- TRUE
          } else {
            mode0 <- FALSE
          }
      }
    if (mode0) {
        igraph::E(graph)$label <- graph_df
      } else {
        # Placehodler
      }
    graph
  }

#' @title Set Arrows Width By 'df' Differences
#'
#' @noRd

edge_weight <- function(graph,
                        mode = c("inverse", "normal", "none"),
                        min_width = .5,
                        max_width = 2) {
    if (is.null(igraph::E(graph)$df)) {
        igraph::E(graph)$width <- max_width
        return(graph)
      }
    tmp1 <- igraph::E(graph)$df
    tmp2 <- switch(mode,
                   inverse = min(tmp1) / tmp1,
                   normal = tmp1 / max(tmp1),
                   none = rep(max_width, length(tmp1)))
    tmp2 <- normalize_edge_width(tmp2,
                                 min_width = min_width,
                                 max_width = max_width)
    igraph::E(graph)$width <- tmp2
    graph
  }

#' @noRd

normalize_edge_width <- function(x,
                                 min_width = .5,
                                 max_width = 2) {
    x_min <- min(x)
    x_max <- max(x)
    x_range <- x_max - x_min
    if (x_range == 0) {
        x[] <- max_width
        return(x)
      } else {
        x_out <- max_width * (x - x_min)/x_range +
                min_width
        x_out
      }
  }

#' @title Minimum Prior
#'
#' @description Find the minimum prior
#' probability required to achieve
#' the desired BIC posterior
#' probability.
#'
#' @details
#' It assumes that all models other than
#' the original model have the same prior
#' probabilities.
#'
#' This function is called by
#' [model_set()] or [print.model_set()]
#' and usually users do
#' not need to call it. It is exported
#' for advanced users.
#'
#' @param bic A named vector of BIC values
#' for a set of models.
#'
#' @param bpp_target A value from zero
#' to 1. The desired BIC posterior
#' probability.
#'
#' @param target_name The name of the
#' original model, as appeared in the
#' names of `bic`.
#'
#' @return A scalar. The required prior
#' probability.
#'
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>
#'
#' @references
#' Wu, H., Cheung, S. F., & Leung, S. O.
#' (2020). Simple use of BIC to assess
#' model selection uncertainty: An
#' illustration using mediation and
#' moderation models.
#' *Multivariate Behavioral Research*,
#' *55*(1), 1--16.
#' \doi{10.1080/00273171.2019.1574546}
#'
#' @seealso [model_set()] and
#'  [print.model_set()]
#'
#' @examples
#'
#' library(lavaan)
#'
#' dat <- dat_path_model
#'
#' mod <-
#' "
#' x3 ~ a*x1 + b*x2
#' x4 ~ a*x1
#' ab := a*b
#' "
#'
#' fit <- sem(mod, dat_path_model, fixed.x = TRUE)
#'
#' out <- model_set(fit)
#' min_prior(out$bic, bpp_target = .8)
#'
#' @export

min_prior <- function(bic,
                      bpp_target,
                      target_name = "original") {
    e <- exp(-.5 * (bic - bic[target_name]))
    i <- which(names(bic) == target_name)
    e1 <- e[i]
    estar <- sum(e[-i])
    k <- length(e)
    p1 <- (1 / (k - 1)) * bpp_target * estar /
            (e1 * (1 - bpp_target) + bpp_target * estar / (k - 1))
    unname(p1)
  }

#' @title Convert a 'model_set'-Object
#' to an Adjacency Matrix
#'
#' @description Convert a
#' 'model_set'-Object to an adjacency
#' matrix for 'igraph'.
#'
#' @param object A `model_set`-class
#' object or a `partables`-class
#' object.
#'
#' @return
#' A numeric matrix of zeros and ones,
#' to indicate the relations between
#' models.
#'
#' @noRd

models_network <- function(object) {
    if (inherits(object, "model_set")) {
        models <- object$models
      } else {
        if (!inherits(object, "partables")) {
            stop("The object must be either a",
                 "model_set object or a",
                 "partables object.")
          }
      }
    m_added <- added(models)
    m_dropped <- dropped(models)
    chk <- mapply(c, m_added, m_dropped)

    p <- length(models)
    i <- which(names(models) == "original")
    net_out <- matrix(0, p, p)
    colnames(net_out) <- rownames(net_out) <- names(models)

    tmp <- sapply(m_added, length) == 1
    net_out[i, tmp] <- 1

    tmp <- sapply(m_dropped, length) == 1
    net_out[tmp, i] <- 1

    net_out <- mark_added(net_out,
                          m_added = m_added)
    net_out <- mark_dropped(net_out,
                            m_dropped = m_dropped)
    net_out
  }

#' @title Convert a 'model_set'-Object
#' to an Adjacency Matrix Using 'semTools::net()' Method
#'
#' @description Convert a
#' 'model_set'-Object to an adjacency
#' matrix for 'igraph'.
#'
#' @details
#' The function [semTools::net()] cannot
#' be used here because the models are
#' not fitted by function calls, but by
#' setting slots directly.
#'
#' @param object A `model_set`-class
#' object.
#'
#' @references
#' Bentler, P. M., & Satorra, A. (2010).
#' Testing model nesting and equivalence.
#' *Psychological Methods, 15*(2),
#' 111--123. \doi{10.1037/a0019625}
#' Asparouhov, T., & Muthén, B. (2019).
#' Nesting and Equivalence Testing for
#' Structural Equation Models.
#' *Structural Equation Modeling: A Multidisciplinary Journal, 26*(2),
#' 302--309. \doi{10.1080/10705511.2018.1513795}
#'
#' @return
#' A numeric matrix of zeros and ones,
#' to indicate the relations between
#' models.
#'
#' @noRd

models_network2 <- function(object,
                            one_df_only = TRUE,
                            progress = FALSE,
                            mark_equivalent = FALSE) {
    if (inherits(object, "model_set")) {
        models <- object$fit
      } else {
        models <- object
      }
    if (!all(sapply(models, inherits, what = "lavaan"))) {
        stop("The object must be either a",
              "model_set object or a",
              "list of lavaan objects")
      }
    converged <- sapply(models,
                        lavaan::lavInspect,
                        what = "converged")
    if (!all(converged)) {
        stop("One or more models did not converge.")
      }
    if (inherits(object, "model_set")) {
        if (!is.null(object$model_df)) {
            model_dfs <- object$model_df
          } else {
            model_dfs <- sapply(models, function(x) {
                  unname(lavaan::fitMeasures(x,
                                             fit.measures = "df"))
                })
          }
      } else {
        model_dfs <- sapply(models, function(x) {
              unname(lavaan::fitMeasures(x,
                                         fit.measures = "df"))
            })
      }
    p <- length(models)
    i <- which(names(models) == "original")
    net_out <- matrix(0, p, p)
    colnames(net_out) <- rownames(net_out) <- names(models)
    if (p == 1) {
        return(net_out)
      }
    if (progress) {
        pstar <- p * (p - 1) / 2
        cat("\nCheck for nested models (",
             pstar,
             " pair[s] of models to check):\n",
             sep = "")
        pb <- utils::txtProgressBar(min = 0,
                                    max = pstar,
                                    width = 50,
                                    char = "+",
                                    style = 3)
      }
    k <- 0
    for (i in seq_len(p - 1)) {
        for (j in seq(from = i + 1,
                      to = p)) {
            k <- k + 1
            if (progress) {
                utils::setTxtProgressBar(pb,
                                         value = k)
              }
            df_i <- model_dfs[i]
            df_j <- model_dfs[j]
            net_chk <- x_net_y(x = models[[i]],
                               y = models[[j]],
                               x_df = df_i,
                               y_df = df_j)
            if (is.na(net_chk)) next
            if (net_chk == "x_within_y") {
                net_out[i, j] <- df_i - df_j
              } else if (net_chk == "y_within_x") {
                net_out[j, i] <- df_j - df_i
              }
            # TODO: Decide what to do with equivalent models
            if (net_chk == "equivalent" && mark_equivalent) {
                net_out[i, j] <- net_out[j, i] <- NA
              }
          }
      }
    if (progress) {
        close(pb)
      }
    if (one_df_only) {
        net_out[net_out > 1] <- 0
      }
    net_out
  }

#' @title Net Without Call
#'
#' @description Based on semTools:::x.within.y().
#'
#' @noRd

x_net_y <- function(x,
                    y,
                    x_df = NULL,
                    y_df = NULL,
                    crit = 1e-4,
                    check_x_y = TRUE) {
    # Based on semTools:::x.within.y().
    if (check_x_y) {
        chk <- check_x_net_y(x, y,
                             ignore_fixed_x = FALSE)
      }
    if (is.null(x_df)) {
        x_df <- lavaan::fitMeasures(x, fit.measures = "df")
      }
    if (is.null(y_df)) {
        y_df <- lavaan::fitMeasures(y, fit.measures = "df")
      }
    if ((x_df == y_df) &&
        (abs(lavaan::fitMeasures(x, fit.measures = "chisq") -
             lavaan::fitMeasures(y, fit.measures = "chisq")) < crit)) {
        return("equivalent")
      }
    # Reorder the models f1 >= f2 on df
    if (x_df < y_df) {
        f1 <- y
        f2 <- x
        out <- "y_within_x"
      } else {
        f1 <- x
        f2 <- y
        out <- "x_within_y"
      }

    f1_fixed <- (lavaan::lavInspect(f1, "options")$fixed.x &&
                 length(lavaan::lavNames(f1, "ov.x")) > 0)
    if (f1_fixed) {
        stop("The net-method does not support models with fixed.x = TRUE. ",
             "Please fit the models with fixed.x = FALSE.")
        # f1 <- lavaan_full_update(f1,
        #                          new_options = list(se = "none",
        #                                             baseline = FALSE,
        #                                             test = "standard",
        #                                             fixed.x = FALSE))
      }

    slot_opt <- f2@Options
    slot_pat <- f2@ParTable
    slot_mod <- f2@Model

    slot_opt2 <- slot_opt
    slot_opt2$se <- "none"
    slot_opt2$fixed.x <- FALSE
    slot_opt2$baseline <- FALSE
    slot_opt2$test <- "standard"

    f1_nobs <- lavaan::lavInspect(f1, "nobs")

    implied_sigma <- lavaan::lavInspect(f1, "cov.ov")
    implied_mean <- lavaan::lavInspect(f1, "mean.ov")
    if (is.list(implied_mean)) {
        if (all(sapply(implied_mean, length) == 0)) {
            implied_mean <- NULL
          }
      } else {
        if (length(implied_mean) == 0) {
            implied_mean <- NULL
          }
      }
    implied_threshold <- lavaan::lavInspect(f1, "thresholds")
    if (is.list(implied_threshold)) {
        if (all(sapply(implied_threshold, length) == 0)) {
            implied_threshold <- NULL
          }
      } else {
        if (length(implied_threshold) == 0) {
            implied_threshold <- NULL
          }
      }

    f1_estimator <- lavaan::lavInspect(f1, "options")$estimator
    if (f1_estimator == "DWLS") {
        f1_WLS.V <- lavaan::lavInspect(f1, "WLS.V")
        f1_NACOV <- lavaan::lavInspect(f1, "gamma")
      } else {
        f1_WLS.V <- NULL
        f1_NACOV <- NULL
      }

    f2_1 <- lavaan::lavaan(slotOptions = slot_opt2,
                           slotParTable = slot_pat,
                           slotModel = slot_mod,
                           data = NULL,
                           sample.cov = implied_sigma,
                           sample.mean = implied_mean,
                           sample.nobs = f1_nobs,
                           sample.th = implied_threshold,
                           WLS.V = f1_WLS.V,
                           NACOV = f1_NACOV,
                           warn = FALSE)
    if (!lavaan::lavInspect(f2_1, "converged")) {
        # If estimation failed to converged,
        # then do not assume a nested relation.
        # Should not be a major problem because this only
        # affect the graph.
        return(NA)
      }
    f2_1_chisq <- unname(lavaan::fitMeasures(f2_1, fit.measures = "chisq"))
    chisq_eq <- f2_1_chisq < crit
    if (chisq_eq) {
        # Equivalence should not be determined this way
      } else {
        out <- "not_nested"
      }
    return(out)
  }

#' @noRd

check_x_net_y <- function(x,
                          y,
                          ignore_fixed_x = FALSE) {
    # Based on semTools::x.within.y()
    # lavaan objects?
    if (!inherits(x, what = "lavaan") ||
        !inherits(y, what = "lavaan")) {
        stop("One or both objects are not lavaan objects.")
      }
    # Same variables?
    vnamesx <- lavaan::lavNames(x, type = "ov")
    vnamesy <- lavaan::lavNames(y, type = "ov")
    if (!setequal(vnamesx, vnamesy)) {
        stop("The two models do not have the same observed variables.")
      }
    # Clustered?
    if (any(lavaan::lavInspect(x, "nclusters") != 1) ||
        any(lavaan::lavInspect(y, "nclusters") != 1)) {
        stop("Models with clustered data not supported.")
      }
    # Check sample sizes
    if (!identical(lavaan::lavInspect(x, "nobs"),
                   lavaan::lavInspect(y, "nobs"))) {
        stop("Sample sizes not identical.")
      }
    # fixed.x?
    if (!ignore_fixed_x) {
        x_fixed <- (lavaan::lavInspect(x, "options")$fixed.x &&
                    length(lavaan::lavNames(x, "ov.x")) > 0)
        y_fixed <- (lavaan::lavInspect(y, "options")$fixed.x &&
                    length(lavaan::lavNames(y, "ov.x")) > 0)
        if (x_fixed || y_fixed) {
            stop("Does not support models with fixed.x = TRUE. ",
                 "Please refit models with fixed.x = FALSE.")
          }
      }
    # Meanstructure
    if (!identical(lavaan::lavInspect(x, "meanstructure"),
                   lavaan::lavInspect(y, "meanstructure"))) {
        stop("One model has a mean structure but the other does not.")
      }
    # Same dataset?
    x_sp <- lavaan::lavTech(x, "sampstat",
                            drop.list.single.group = FALSE,
                            add.labels = TRUE,
                            list.by.group = TRUE)
    y_sp <- lavaan::lavTech(y, "sampstat",
                            drop.list.single.group = FALSE,
                            add.labels = TRUE,
                            list.by.group = TRUE)
    x_g <- length(x_sp)
    y_g <- length(y_sp)
    x_names <- colnames(x_sp[[1]]$cov)
    for (j in seq_len(x_g)) {
        if (!identical(x_sp[[1]]$cov,
                       y_sp[[1]]$cov[x_names, x_names])) {
            stop("The sample statistics are not identical.")
          }
        # Identical if both mean vectors are NULL
        if (!identical(x_sp[[1]]$mean,
                       y_sp[[1]]$mean[x_names])) {
            stop("The sample statistics are not identical.")
          }
      }
    return(TRUE)
  }

#' @title Internal Fast Update
#'
#' @noRd

lavaan_fast_update <- function(x,
                               new_options = list()) {
    stop("This function should not be called. It is a work-in-progress.")
    # slot_opt <- x@Options
    # slot_pat <- x@ParTable
    # slot_mod <- x@Model
    # slot_smp <- x@SampleStats
    # slot_dat <- x@Data

    # slot_opt2 <- slot_opt
    # slot_opt2$se <- "none"
    # slot_opt2$fixed.x <- FALSE
    # slot_opt2$baseline <- FALSE
    # slot_opt2$test <- "none"
    # slot_opt2 <- utils::modifyList(slot_opt2,
    #                                new_options,
    #                                keep.null = TRUE)

    # fit2 <- lavaan::lavaan(slotOptions = slot_opt2,
    #                        slotParTable = slot_pat,
    #                        slotModel = slot_mod,
    #                        slotSampleStats = slot_smp,
    #                        slotData = slot_dat)
    # fit2
  }

#' @title Internal Full Update
#'
#' @noRd

lavaan_full_update <- function(x,
                               new_options = list()) {
    stop("This function should not be called. It is a work-in-progress.")
#     x_model <- lavaan::parameterTable(x)
#     slot_opt <- x@Options
#     slot_opt2 <- utils::modifyList(slot_opt,
#                                   new_options,
#                                   keep.null = TRUE)

#     raw_data <- tryCatch(lavaan::lavInspect(x, "data"),
#                         error = function(e) e)
#     if (inherits(raw_data, "error")) {
#         raw_data <- NULL
#         has_data <- FALSE
#       } else {
#         colnames(raw_data) <- lavaan::lavNames(x)
#         rownames(raw_data) <- lavaan::lavInspect(x, "case.idx")
#         has_data <- TRUE
#       }

#     if (has_data) {
#         # Placeholder
#       } else {
#         x_nobs <- lavaan::lavInspect(x, "nobs")
#         x_sp <- lavaan::lavInspect(x, "sampstat")
#         ng <- lavaan::lavInspect(x, "ngroups")
#         if (ng > 1) {
#             x_cov <- lapply(x_sp, function(x) x$cov)
#           } else {
#             x_cov <- x_sp$cov
#           }
#         if (ng > 1) {
#             x_mean <- lapply(x_sp, function(x) x$mean)
#           } else {
#             x_mean <- x_sp$mean
#           }
#         x_thresholds <- lavaan::lavInspect(x, "thresholds")
#         if (is.list(x_thresholds)) {
#             if (all(sapply(x_thresholds, length) == 0)) {
#                 x_thresholds <- NULL
#               }
#           } else {
#             if (length(x_thresholds) == 0) {
#                 x_thresholds <- NULL
#               }
#           }
#         x_estimator <- lavaan::lavInspect(x, "options")$estimator
#         if (x_estimator == "DWLS") {
#             x_WLS.V <- lavaan::lavInspect(x, "WLS.V")
#             x_NACOV <- lavaan::lavInspect(x, "gamma")
#           } else {
#             x_WLS.V <- NULL
#             x_NACOV <- NULL
#           }
#       }

#     if (has_data) {
#         out <- lavaan::lavaan(model = x_model,
#                               slotOptions = slot_opt2,
#                               data = raw_data)
#       } else {
#         out <- lavaan::lavaan(model = x,
#                               slotOptions = slot_opt2,
#                               sample.cov = x_cov,
#                               sample.mean = x_mean,
#                               sample.nobs = x_nobs,
#                               semple.th = x_thresholds,
#                               WLS.V = x_WLS.V,
#                               NACOV = x_NACOV)
#       }
#     return(out)
  }

#' @title Mark Models Formed by Adding Free Parameters
#'
#' @noRd

mark_added <- function(net_out,
                       m_added) {
    for (i in seq_len(length(m_added))) {
        for (j in seq_len(length(m_added))) {
            x <- m_added[[i]]
            y <- m_added[[j]]
            if (identical(x, y)) next
            if (is.null(x)) next
            if (is.null(y)) next
            tmp <- union(x, y)
            tmp1 <- setequal(tmp, x)
            tmp2 <- setequal(tmp, y)
            if (tmp1 && !tmp2) {
                z <- length(setdiff(x, y))
                if (z == 1) {
                    net_out[j, i] <- z
                  }
              }
          }
      }
    net_out
  }

#' @title Mark Models Formed by Dropping Free Parameters
#'
#' @noRd

mark_dropped <- function(net_out,
                       m_dropped) {
    for (i in seq_len(length(m_dropped))) {
        for (j in seq_len(length(m_dropped))) {
            x <- m_dropped[[i]]
            y <- m_dropped[[j]]
            if (identical(x, y)) next
            if (is.null(x)) next
            if (is.null(y)) next
            tmp <- union(x, y)
            tmp1 <- setequal(tmp, x)
            tmp2 <- setequal(tmp, y)
            if (tmp1 && !tmp2) {
                z <- length(setdiff(x, y))
                if (z == 1) {
                    net_out[i, j] <- z
                  }
              }
          }
      }
    net_out
  }

#' @title Add Weights to an Adjacency Matrix
#'
#' @noRd

add_weight <- function(net_out,
                       bpp) {
    for (i in seq_len(nrow(net_out))) {
        for (j in seq_len(ncol(net_out))) {
            if (net_out[i, j] != 0) {
                net_out[i, j] <- bpp[i] - bpp[j]
              }
          }
      }
    net_out
  }

#' @title Normalize Node Size
#'
#' @noRd

normalize_size <- function(bpp,
                           min_size = 1,
                           max_size = 30) {
    bpp_min <- min(bpp)
    bpp_max <- max(bpp)
    bpp_range <- bpp_max - bpp_min
    tmp_size <- max_size * (bpp - bpp_min)/bpp_range +
                min_size
    tmp_size
  }

#' @title Labels for Vertices
#'
#' @noRd

v_labels <- function(x) {
    out0 <- gsub(": ", ":\n", x, fixed = TRUE)
    out0 <- gsub(";", "\n", out0, fixed = TRUE)
    out0
  }

#' @title Identify Clusters of Equivalent Models
#'
#' @noRd

equivalent_clusters <- function(fits,
                                name_cluster = TRUE,
                                progress = FALSE) {
    net_eq <- models_network2(fits,
                              mark_equivalent = TRUE,
                              progress = progress,
                              one_df_only = TRUE)
    if (!any(is.na(net_eq))) {
        return(NULL)
      }
    net_eq[!is.na(net_eq)] <- 0
    net_eq[is.na(net_eq)] <- 1
    p_eq <- igraph::graph_from_adjacency_matrix(net_eq,
                                                mode = "undirected")
    p_eq_group <- igraph::max_cliques(p_eq)
    i <- sapply(p_eq_group, length)
    p_eq_group <- p_eq_group[i > 1]
    p_eq_group_names <- lapply(p_eq_group, function(x) x$name)
    if (name_cluster) {
        tmp <- sapply(p_eq_group_names, function(x) x[1])
        names(p_eq_group_names) <- tmp
      }
    p_eq_group_names
  }

if (TRUE &&
    interactive() &&
    length(unclass(packageVersion("modelbpp"))[[1]]) == 4) {

# Long tests

suppressMessages(library(lavaan))

data(dat_path_model)

mod1 <-
"
x3 ~ x1 + x2
x4 ~ x1
"
fit1_fixed <- sem(mod1, dat_path_model, fixed.x = TRUE)
fit1_nofix <- sem(mod1, dat_path_model, fixed.x = FALSE)
fit1_mean <- sem(mod1, dat_path_model, fixed.x = FALSE, meanstructure = TRUE)

mod2 <-
"
x3 ~ x2 + 0*x1
x4 ~ x1
"
fit2_fixed <- sem(mod2, dat_path_model, fixed.x = TRUE)
fit2_nofix <- sem(mod2, dat_path_model, fixed.x = FALSE)
fit2_mean <- sem(mod2, dat_path_model, fixed.x = FALSE, meanstructure = TRUE)

mod3 <-
"
x3 ~ x1 + x2
x4 ~ x1 + x2
"
fit3_fixed <- sem(mod3, dat_path_model, fixed.x = TRUE)
fit3_nofix <- sem(mod3, dat_path_model, fixed.x = FALSE)
fit3_mean <- sem(mod3, dat_path_model, fixed.x = FALSE, meanstructure = TRUE)

mod4 <-
"
x3 ~ x1
x4 ~ x1
"
fit4_fixed <- sem(mod4, dat_path_model, fixed.x = TRUE)
fit4_nofix <- sem(mod4, dat_path_model, fixed.x = FALSE)
fit4_mean <- sem(mod4, dat_path_model, fixed.x = FALSE, meanstructure = TRUE)

# suppressMessages(library(semTools))
# net(fit1_nofix, fit2_nofix, fit3_nofix)@test
# net(fit1_nofix, fit2_nofix)@test

# x_net_y(fit1_nofix, fit2_nofix)
# x_net_y(fit2_nofix, fit1_nofix)
# x_net_y(fit1_nofix, fit1_nofix)
# x_net_y(fit1_nofix, fit3_nofix)
# x_net_y(fit3_nofix, fit1_nofix)
# x_net_y(fit2_nofix, fit3_nofix)
# x_net_y(fit3_nofix, fit2_nofix)

out <- model_set(fit1_nofix,
                 df_change_add = 1,
                 df_change_drop = 2,
                 progress = FALSE)
expect_identical(models_network(out),
                 models_network2(out,
                                 progress = FALSE))

# User models

suppressMessages(library(igraph))

mod1 <-
"
x2 ~ x1
x3 ~ x2
x4 ~ x3
"
fit1 <- sem(mod1, dat_path_model, fixed.x = FALSE)

mod2 <-
"
x3 ~ x4
x2 ~ x3
x1 ~ x2
"
fit2 <- sem(mod2, dat_path_model, fixed.x = FALSE)

mod3 <-
"
x2 ~ x1
x3 ~ x2 + x1
x4 ~ x3
"
fit3 <- sem(mod3, dat_path_model, fixed.x = FALSE)

mod4 <-
"
x2 ~ x1
x3 ~ x2 + x1
x4 ~ x3 + x2
"
fit4 <- sem(mod4, dat_path_model, fixed.x = FALSE)

mod5 <-
"
x2 ~ x4
x1 ~ x4
x3 ~ x2 + x1
"
fit5 <- sem(mod5, dat_path_model, fixed.x = FALSE)

fits <- list(fit2 = fit2,
             fit3 = fit3,
             fit4 = fit4,
             fit5 = fit5)
ptables <- lapply(fits, parameterTable)
class(ptables) <- c("partables", class(ptables))
out <- model_set(fit1,
                 partables = ptables,
                 progress = FALSE)
g <- model_graph(out,
                 progress = FALSE)
plot(g)
tmp <- lapply(E(g), attributes)
expect_equal(length(E(g)), 3)
expect_equal(length(V(g)), 5)
}

#' @title A Sample Dataset Based on a
#'  Path Model (For Testing)
#'
#' @description Generated from the
#'  a path model (n = 100).
#'
#' @details  The model used to generate
#' this dataset:
#'
#' ```
#' x1 ~~ x2
#' x3 ~ x1 + x2
#' x4 ~ x3 + x1 + x2
#' ```
#'
#' @format A data frame with four variables:
#' \describe{
#'  \item{x1}{Predictor}
#'  \item{x2}{Predictor}
#'  \item{x3}{Mediator}
#'  \item{x4}{Outcome}
#' }
"dat_path_model"


#' @title A Sample Dataset Based On a
#' Confirmatory Factor Analysis Model
#' (For Testing)
#'
#' @description Generated from
#' a confirmatory factor analysis
#' model (n = 200).
#'
#' @details The model used to generate
#' this dataset:
#'
#' ```
#' f1 =~ x1 + x2 + x3 + x5
#' f2 =~ x3 + x4 + x5 + x6
#' f1 ~~ f2
#' ```
#'
#' @format A data frame with six variables:
#' \describe{
#'  \item{x1}{Indicator}
#'  \item{x2}{Indicator}
#'  \item{x3}{Indicator}
#'  \item{x4}{Indicator}
#'  \item{x5}{Indicator}
#'  \item{x6}{Indicator}
#' }
"dat_cfa"

#' @title A Sample Dataset Based On a
#' Complex Path Model (For Testing)
#'
#' @description Generated from
#' a complex path model (n = 200).
#'
#' @details The model used to generate
#' this dataset:
#'
#' ```
#' y4 ~  x1 + x2 + x3
#' y5 ~  y4 + x1 + x2
#' y6 ~  y4 + y5 + x1 + x2 + x3
#' x1 ~~ x2 + x3
#' x2 ~~ x3
#' ```
#'
#' @format A data frame with six variables:
#' \describe{
#'  \item{x1}{Predictor}
#'  \item{x2}{Predictor}
#'  \item{x3}{Predictor}
#'  \item{y4}{Mediator}
#'  \item{y5}{Mediator}
#'  \item{y6}{Outcome}
#' }
"dat_path_model_p06"

#' @title A Sample Dataset Based On a
#' Serial Mediation Model (For Testing)
#'
#' @description Generated from
#' a serial mediation model (n = 100).
#'
#' @details The model used to generate
#' this dataset:
#'
#' ```
#' m1 ~ x
#' m2 ~ m1
#' y  ~ m2
#' ```
#'
#' @format A data frame with four variables:
#' \describe{
#'  \item{x}{Predictor}
#'  \item{m1}{Mediator}
#'  \item{m2}{Mediator}
#'  \item{y}{Outcome}
#' }
"dat_serial_4"

#' @title A Sample Dataset Based On a
#' Structural Model (For Testing)
#'
#' @description Generated from
#' a structural model with
#' latent variables (n = 250).
#'
#' @details The model to be fitted:
#'
#' ```
#' f1 =~ x1 + x2 + x3 + x4
#' f2 =~ x5 + x6 + x7 + x8
#' f3 =~ x9 + x10 + x11 + x12
#' f4 =~ x13 + x14 + x15 + x16
#' f3 ~ f1 + f2
#' f4 ~ f3
#' ```
#'
"dat_sem"

#' @title A Sample Dataset Based On a
#' Serial Mediation Model
#' With Weak Paths (For Testing)
#'
#' @description Generated from
#' a serial mediation model (n = 100).
#'
#' @details The model to be fitted:
#'
#' ```
#' m1 ~ x
#' m2 ~ m1 + x
#' y  ~ m2 + m1 + x
#' ```
#'
#' @format A data frame with four variables:
#' \describe{
#'  \item{x}{Predictor}
#'  \item{m1}{Mediator}
#'  \item{m2}{Mediator}
#'  \item{y}{Outcome}
#' }
"dat_serial_4_weak"

#' @noRd

check_sem_out <- function(sem_out) {
    # Estimator: Must be ML
    # No. of groups: 1
    # No. of levels: 1
    if (!("ML" %in% sem_out@Options$estimator)) {
        stop("The estimator of 'sem_out' is not 'ML'.")
      }
    if (lavaan::lavInspect(sem_out, "ngroups") != 1) {
        stop("The number of groups in 'sem_out' is not 1.")
      }
    if (lavaan::lavInspect(sem_out, "nlevels") != 1) {
        stop("The number of levels in 'sem_out' is not 1.")
      }
    return(TRUE)
  }

# WIP

if (FALSE) {

    suppressMessages(library(lavaan))

    dat <- dat_sem

    mod <-
    "
    f1 =~ x1 + x2 + x3 + x4
    f2 =~ x5 + x6 + x7 + x8
    f3 =~ x9 + x10 + x11 + x12
    f4 =~ x13 + x14 + x15 + x16
    f2 ~ a * f1
    f3 ~ a * f2
    f4 ~ a * f3
    "

    fit <- sem(mod, dat_sem)
    pt <- parameterTable(fit)

    all_loadings <- sapply(paste0("f", 1:4),
                           function(x) {
                              paste0(x,
                                     "=~",
                                     paste0("x", 1:16))
                            })
    mod_to_add <- get_add(fit,
                          must_not_add = all_loadings,
                          df_change = 2)
    length(mod_to_add)
    head(mod_to_add)
    names(mod_to_add)
    attributes(mod_to_add[[10]])
    attributes(mod_to_add[[35]])
    attributes(mod_to_add[[36]])
    attributes(mod_to_add[[45]])

    tmp1 <- sapply(mod_to_add, attr, "df_expected")
    tmp2 <- sapply(mod_to_add, attr, "df_actual")
    all.equal(tmp1, tmp2)

    mod_to_add <- get_add(fit,
                          must_not_add = all_loadings,
                          df_change = 4)
    length(mod_to_add)
    head(mod_to_add)
    names(mod_to_add)
    attributes(mod_to_add[[128]])
    attributes(mod_to_add[[78]])

    # Test Print

    mod_to_add
    print(mod_to_add,
          max_tables = 25)
  }

#' @title Fit a List of Models
#'
#' @description Fit a list of models
#' to a dataset.
#'
#' @details It receives a list of
#' models, defined by `lavaan` parameter
#' tables (usually generated by
#' [model_set()], [get_add()] or
#' [get_drop()]),
#' and fit them to a dataset stored
#' in a `lavaan`-class object.
#'
#' This function is called by
#' [model_set()] and usually users do
#' not need to call it. It is exported
#' for advanced users.
#'
#' @param model_list A list of parameter
#' tables to be used by
#' [lavaan::lavaan()] or [update()].
#' Usually generated by [get_add()] or
#' [get_drop()].
#'
#' @param sem_out The output from an
#' structural equation modeling
#' function. It currently supports
#' [lavaan::lavaan-class] objects
#' only. Usually
#' the one used in [model_set()],
#' [get_add()] or
#' [get_drop()] to generate the list of
#' models.
#'
#' @param original String. If provided,
#' it should be a name of a model
#' in `model_list`, with which
#' differences in model degrees of
#' freedom will be computed for other
#' models. If `NULL`, the default,
#' then the model in `sem_out` will
#' be used to computed the differences
#' in model degrees of freedom. If `NA`,
#' then differences in model *df* will
#' not be computed.
#'
#' @param parallel If `TRUE`, parallel
#' processing will be used to fit the
#' models. Default is `FALSE`.
#'
#' @param ncores Numeric. The number of
#' CPU cores to be used if `parallel`
#' is `TRUE`.
#'
#' @param make_cluster_args A list of
#' named arguments to be passed to
#' `parallel::makeCluster()`. Used by
#' advanced users to configure the
#' cluster if `parallel` is `TRUE`.
#' Default is `list()`.
#'
#' @param progress Whether a progress
#' bar will be displayed, implemented
#' by the `pbapply` package. Default
#' is `TRUE`.
#'
#' @param verbose Whether additional
#' messages will be displayed, such
#' as the expected processing time.
#' Default is `TRUE`.
#'
#' @return An object of the class
#' `sem_outs`, a list with the
#' following major elements:
#'
#' * `fit`: A named list of
#'    [lavaan::lavaan()] output objects or
#'    [update()] for fitting a model with
#'    the added parameters.
#'
#' * `change`: A numeric vector, of the
#'    same length as `fit`. The change
#'    in model *df* for each fit compared
#'    to the original model. A
#'    positive number denotes one less
#'    free parameter. A negative number
#'    denotes one more free parameter or
#'    one less constraint.
#'
#' * `converged`: A named vector of
#'    boolean values, of the same length
#'    as `fit`. Indicates whether each
#'    fit converged or not.
#'
#' * `post_check`: A named vector of
#'    boolean values, of the same length
#'    as `fit`. Indicates whether the
#'    solution of each fit is
#'    admissible or not. Checked by
#'    [lavaan::lavInspect()] with
#'    the `what` argument set to
#'    `"post.check"`.
#'
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>
#'
#' @examples
#'
#' library(lavaan)
#' dat <- dat_path_model
#' mod <-
#' "
#' x3 ~ a*x1 + b*x2
#' x4 ~ a*x1
#' ab := a*b
#' "
#' fit <- sem(mod, dat_path_model, fixed.x = TRUE)
#' mod_to_add <- get_add(fit)
#' fit_many(mod_to_add, fit)
#'
#' @export

fit_many <- function(model_list,
                     sem_out,
                     original = NULL,
                     parallel = FALSE,
                     ncores = max(parallel::detectCores(logical = FALSE) - 1, 1),
                     make_cluster_args = list(),
                     progress = TRUE,
                     verbose = TRUE) {
  if (missing(model_list)) stop("model_list is not supplied.")
  if (missing(sem_out)) stop("sem_out is not supplied.")
  if (!inherits(sem_out, "lavaan")) {
      stop("sem_out is not a lavaan-class object.")
    }

  p_models <- length(model_list)

  slot_opt <- sem_out@Options
  # slot_smp <- sem_out@SampleStats
  # slot_dat <- sem_out@Data
  slot_opt$se <- "none"
  slot_opt$baseline <- FALSE
  slot_opt$verbose <- FALSE
  has_groups <- lavaan::lavTech(sem_out, "ngroups") > 1
  if (has_groups) {
      group_var <- lavaan::lavTech(sem_out, "group")
    }
  raw_data <- tryCatch(lavaan::lavInspect(sem_out, "data",
                                          drop.list.single.group = FALSE),
                       error = function(e) e)
  if (inherits(raw_data, "error")) {
      raw_data <- NULL
      has_data <- FALSE
    } else {
      for (i in seq_along(raw_data)) {
          idx <- lavaan::lavInspect(sem_out, "case.idx",
                                    drop.list.single.group = FALSE)
          colnames(raw_data[[i]]) <- lavaan::lavNames(sem_out)
          rownames(raw_data[[i]]) <- idx[[i]]
        }
      has_data <- TRUE
    }

  if (has_data) {
      # Placeholder
    } else {
      # This block works with ngroups > 1
      sem_out_nobs <- lavaan::lavInspect(sem_out, "nobs")
      sem_out_sp <- lavaan::lavInspect(sem_out, "sampstat")
      ng <- lavaan::lavInspect(sem_out, "ngroups")
      if (ng > 1) {
          sem_out_cov <- lapply(sem_out_sp, function(x) x$cov)
        } else {
          sem_out_cov <- sem_out_sp$cov
        }
      if (ng > 1) {
          sem_out_mean <- lapply(sem_out_sp, function(x) x$mean)
        } else {
          sem_out_mean <- sem_out_sp$mean
        }
      sem_thresholds <- lavaan::lavInspect(sem_out, "thresholds")
      if (is.list(sem_thresholds)) {
          if (all(sapply(sem_thresholds, length) == 0)) {
              sem_thresholds <- NULL
            }
        } else {
          if (length(sem_thresholds) == 0) {
              sem_thresholds <- NULL
            }
        }
      sem_out_estimator <- lavaan::lavInspect(sem_out, "options")$estimator
      if (sem_out_estimator == "DWLS") {
          sem_out_WLS.V <- lavaan::lavInspect(sem_out, "WLS.V")
          sem_out_NACOV <- lavaan::lavInspect(sem_out, "gamma")
        } else {
          sem_out_WLS.V <- NULL
          sem_out_NACOV <- NULL
        }
    }

  # fit_i <- function(x) {
  #     lavaan::lavaan(model = x,
  #                    slotOptions = slot_opt,
  #                    slotSampleStats = slot_smp,
  #                    slotData = slot_dat)
  #   }

  if (has_data) {
      # Merge into one data frame
      if (length(raw_data) == 1) {
          raw_data <- as.data.frame(raw_data[[1]])
        } else {
          for (i in seq_along(raw_data)) {
              raw_data[[i]] <- as.data.frame(raw_data[[i]])
              raw_data[[i]][, group_var] <- names(raw_data)[i]
            }
          raw_data <- do.call(rbind, raw_data)
        }
    }

  if (has_data) {
      if (has_groups) {
          fit_i <- function(x,
                            opt_args = list()) {
              slot_opt1 <- utils::modifyList(slot_opt,
                                             opt_args)
              # We need the raw data because the order
              # of variables may change
              lavaan::lavaan(model = x,
                             slotOptions = slot_opt1,
                             group = group_var,
                             data = raw_data)
            }
        } else {
          fit_i <- function(x,
                            opt_args = list()) {
              slot_opt1 <- utils::modifyList(slot_opt,
                                             opt_args)
              # We need the raw data because the order
              # of variables may change
              lavaan::lavaan(model = x,
                            slotOptions = slot_opt1,
                            data = raw_data)
            }
        }
    } else {
      fit_i <- function(x,
                        opt_args = list()) {
          slot_opt1 <- utils::modifyList(slot_opt,
                                         opt_args)
          lavaan::lavaan(model = x,
                         slotOptions = slot_opt1,
                         sample.cov = sem_out_cov,
                         sample.mean = sem_out_mean,
                         sample.nobs = sem_out_nobs,
                         semple.th = sem_thresholds,
                         WLS.V = sem_out_WLS.V,
                         NACOV = sem_out_NACOV)
        }
    }

  # Check ncores
  # Adapted from manymome::fit2boot_out_do_boot()
  ft <- as.numeric(lavaan::lavInspect(sem_out, "timing")$total)
  if (parallel) {
        if (is.numeric(ncores)) {
            ncores0 <- parallel::detectCores()
            if (ncores == ncores0) {
                warning(paste0("'ncores' >= The number of detected cores (",
                                ncores0,"). The computer may not be responsive",
                                " when models are estimated."),
                        immediate. = TRUE)
                utils::flush.console()
              }
            if (ncores > ncores0) {
                ncores <- max(ncores0 - 1, 1L)
              }
          } else {
            ncores <- 1L
          }
    } else {
      ncores <- 1L
    }
  # Set has_cl
  # Adapted from manymome::fit2boot_out_do_boot()
  if (ncores > 1L) {
      make_cluster_args <- utils::modifyList(make_cluster_args,
                                      list(spec = ncores))
      tmp <- tryCatch({cl <- do.call(parallel::makeCluster,
                                    make_cluster_args)},
                      error = function(e) e)
      has_cl <- !inherits(tmp, "error")
    } else {
      has_cl <- FALSE
    }
  # Do the analysis.
  # Adapted from manymome::fit2boot_out_do_boot()
  if (has_cl) {
      texp <-  1.2 * p_models * ft / length(cl)
      if (verbose) {
          message(paste0(length(cl), " processes started to run model fitting."))
          message(paste0("The expected CPU time is about ",
                          round(texp, 2),
                          " second(s)."))
          utils::flush.console()
        }
      pkgs <- .packages()
      pkgs <- rev(pkgs)
      parallel::clusterExport(cl, "pkgs", envir = environment())
      parallel::clusterEvalQ(cl, {
                      sapply(pkgs,
                      function(x) {
                          # Packages temporarily loaded are excluded
                          try(library(x, character.only = TRUE),
                                      silent = TRUE)
                        })
                    })
      if (progress) {
          op_old <- pbapply::pboptions(type = "timer")
          cat("\nFit the", length(model_list), "models:\n")
          tmp <- tryCatch({rt <- system.time(fit_list <- suppressWarnings(
                            pbapply::pblapply(model_list,
                                              fit_i,
                                              cl = cl)))},
                            error = function(e) e)
          pbapply::pboptions(op_old)
        } else {
          tmp <- tryCatch({rt <- system.time(fit_list <- suppressWarnings(
                            parallel::parLapplyLB(cl,
                                                  model_list,
                                                  fit_i)))},
                            error = function(e) e)
        }
      if (inherits(tmp, "error")) {
          try(parallel::stopCluster(cl), silent = TRUE)
          stop("Running in parallel failed. Please set 'parallel' to FALSE.")
        }
      parallel::stopCluster(cl)
    } else {
      if (progress) {
          cat("\nFit the", length(model_list), "model(s) (duplicated models removed):\n")
          rt <- system.time(fit_list <- suppressWarnings(pbapply::pblapply(model_list, fit_i)))
        } else {
          rt <- system.time(fit_list <- suppressWarnings(lapply(model_list, fit_i)))
        }
    }
  df_list <- mapply(fit_many_get_df,
                    fit = fit_list,
                    model = model_list,
                    MoreArgs = list(fit_i = fit_i),
                    SIMPLIFY = TRUE)
  if (is.null(original)) {
      sem_out_df <- as.numeric(lavaan::fitMeasures(sem_out, "df"))
      # change_list <- sapply(fit_list,
      #     function(x) sem_out_df - as.numeric(lavaan::fitMeasures(x, fit.measures = "df")))
      change_list <- sem_out_df - df_list
    } else {
      if (original %in% names(model_list)) {
          i_original <- match(original, names(model_list))
          # change_list <- sapply(fit_list,
          #     function(x) as.numeric(lavaan::fitMeasures(x, fit.measures = "df")))
          df_original <- df_list[i_original]
          change_list <- df_original - df_list
        } else {
          change_list <- rep(NA, p_models)
        }
    }
  converged_list <- sapply(fit_list,
      function(x) lavaan::lavInspect(x, "converged"))
  post_check_list <- sapply(fit_list,
      function(x) lavaan::lavInspect(x, "post.check"))
  out <- list(fit = fit_list,
              change = change_list,
              converged = converged_list,
              post_check = post_check_list,
              model_df = df_list,
              call = match.call())
  class(out) <- c("sem_outs", class(out))
  out
}

#' @noRd

lavaan_to_sem_outs <- function(x,
                               original = NULL) {
    p_models <- length(x)
    if (is.null(original)) {
        change_list <- rep(NA, x)
      } else {
        if (original %in% names(x)) {
            i_original <- match(original, names(x))
            change_list <- sapply(x,
                function(x) as.numeric(lavaan::fitMeasures(x, fit.measures = "df")))
            df_original <- change_list[i_original]
            change_list <- df_original - change_list
          } else {
            change_list <- rep(NA, p_models)
          }
      }
    converged_list <- sapply(x,
        function(x) lavaan::lavInspect(x, "converged"))
    post_check_list <- sapply(x,
        function(x) lavaan::lavInspect(x, "post.check"))
    out <- list(fit = x,
                change = change_list,
                converged = converged_list,
                post_check = post_check_list,
                call = match.call())
    class(out) <- c("sem_outs", class(out))
    out
  }

#' @noRd

fit_many_get_df <- function(fit,
                            model,
                            fit_i) {
    out <- tryCatch(lavaan::fitMeasures(fit, fit.measures = "df"),
                    error = function(e) e)
    if (!inherits(out, "error")) {
        return(as.numeric(out))
      }
    fit1 <- suppressWarnings(fit_i(model,
                                   opt_args = list(optim.force.converged = TRUE,
                                                   do.fit = FALSE,
                                                   warn = FALSE)))
    out <- tryCatch(lavaan::fitMeasures(fit1, fit.measures = "df"),
                    error = function(e) e)
    if (!inherits(out, "error")) {
        return(as.numeric(out))
      }
    return(NA)
  }

if (interactive() &&
    length(unclass(packageVersion("modelbpp"))[[1]]) == 4) {

suppressMessages(library(lavaan))

data(dat_path_model)
set.seed(1234)
dat_path_model$group <- sample(c("gp1", "gp2", "gp3"),
                               size = nrow(dat_path_model),
                               replace = TRUE)
mod1 <-
"
x3 ~ x1 + x2
x4 ~ x1
"
fit1 <- sem(mod1,
           dat_path_model,
           fixed.x = FALSE,
           group = "group")
mod2 <-
"
x3 ~ x1 + x2
x4 ~ x1 + x2
"
fit2 <- sem(mod2,
           dat_path_model,
           fixed.x = FALSE,
           group = "group")
mod3 <-
"
x4 ~ x1
x3 ~ x1 + x2
"
fit3 <- sem(mod3,
           dat_path_model,
           fixed.x = FALSE,
           group = "group")
out <- fit_many(list(parameterTable(fit1),
                     parameterTable(fit2),
                     parameterTable(fit3)),
                fit,
                parallel = 3,
                ncores = 2)
expect_equal(lavInspect(out$fit[[1]], "group.label"),
             lavInspect(out$fit[[3]], "group.label"))
}

lav_pure_x <- function(sem_out) {
  # Find the names of pure-x latent variables
  a <- lavaan::lavNames(
          sem_out,
          type = "lv.x"
        )
  a
}

lav_pure_y <- function(sem_out) {
  # Find the names of pure-y latent variables
  a <- lavaan::lavNames(
          sem_out,
          type = "lv.y"
        )
  a
}

lav_indicators <- function(
  sem_out,
  lav
) {
  # Find the indicators of lav,
  # a vector of names of latent variables
  pt <- lavaan::parameterTable(sem_out)
  i1 <- pt$lhs %in% lav
  i2 <- pt$op == "=~"
  pt$rhs[i1 & i2]
}

lav_pure_x_indicators <- function(sem_out) {
  # Find the indicators of
  # all pure-x latent variables
  a <- lav_pure_x(sem_out)
  lav_indicators(sem_out,
                 lav = a)
}

lav_pure_y_indicators <- function(sem_out) {
  # Find the indicators of
  # all pure-y latent variables
  a <- lav_pure_y(sem_out)
  lav_indicators(sem_out,
                 lav = a)
}

lav_loadings <- function(
  sem_out,
  lav
) {
  pt <- lavaan::parameterTable(sem_out)
  i1 <- (pt$lhs %in% lav) &
        (pt$op == "=~")
  pt_lav <- pt[i1, c("lhs", "op", "rhs", "block", "group")]
  pt_lav
}

lav_pure_x_loadings <- function(sem_out) {
  x <- lav_pure_x(sem_out)
  lav_loadings(sem_out,
               lav = x)
}

lav_pure_y_loadings <- function(sem_out) {
  y <- lav_pure_y(sem_out)
  lav_loadings(sem_out,
               lav = y)
}

lav_cross_set <- function(
  sem_out,
  type = c("pure_x", "pure_y", "user"),
  user_lav = character(0)
) {
  # Find loadings between
  #   pure x latent variables and current x indicators
  #   pure y latent variables and current y indicators
  # which are
  #   not yet in the model.
  # That is:
  # Find all possible new cross loadings among
  # pure x/y latent factors
  type <- match.arg(type)
  x_inds <- switch(
              type,
              pure_x = lav_pure_x_indicators(sem_out),
              pure_y = lav_pure_y_indicators(sem_out),
              user = lav_indicators(sem_out, user_lav)
            )
  x <- switch(
          type,
          pure_x = lav_pure_x(sem_out),
          pure_y = lav_pure_y(sem_out),
          user = user_lav
        )
  out0 <- all_possible_loadings(sem_out)
  a1 <- out0[(out0$lhs %in% x) &
             (out0$rhs %in% x_inds), ]
  # TODO:
  # - Support multigroup model
  a2 <- switch(
          type,
          pure_x = lav_pure_x_loadings(sem_out)[, c("lhs", "op", "rhs")],
          pure_y = lav_pure_y_loadings(sem_out)[, c("lhs", "op", "rhs")],
          user = lav_loadings(sem_out, lav = user_lav)
        )
  a1$id <- seq_len(nrow(a1))
  tmp <- merge(
        x = a2,
        y = a1,
        all.y = FALSE
      )
  # Loadings not yet in the model
  a3 <- a1[!(a1$id %in% tmp$id), ]
  a3$id <- NULL
  rownames(a3) <- NULL
  if (nrow(a3) > 0) {
    a3 <- a3[order(a3$lhs, a3$rhs), ]
  }
  a3
}

xy_lav <- function(sem_out) {
  # Find latent variables with an x-y relation,
  # either directly or indirectly
  tmp1 <- lavaan::lavInspect(
            sem_out,
            what = "free",
            drop.list.single.group = FALSE
          )
  tmp2 <- lavaan::lavInspect(
            sem_out,
            what = "est",
            drop.list.single.group = FALSE
          )
  tmp1 <- lapply(tmp1, function(x) x$beta > 0)
  tmp2 <- lapply(tmp2, function(x) x$beta != 0)
  f <- function(x1, x2) {
          a <- x1 | x2
          a[] <- as.numeric(a)
          a
        }
  beta <- mapply(
            f,
            x1 = tmp1,
            x2 = tmp2,
          SIMPLIFY = FALSE
        )
  adj <- lapply(
            beta,
            t
          )
  graph_adj <- lapply(
              adj,
              igraph::graph_from_adjacency_matrix,
              mode = "directed",
            )
  lav_x <- lavaan::lavNames(
              sem_out,
              type = "lv.x"
            )
  lav_nox <- lavaan::lavNames(
              sem_out,
              type = "lv.nox"
            )
  xy_pairs <- expand.grid(y = lav_nox,
                          x = c(lav_x, lav_nox),
                          stringsAsFactors = FALSE)
  # TODO:
  # - Support multigroup models
  out <- mapply(igraph::all_simple_paths,
                from = xy_pairs$x,
                to = xy_pairs$y,
                MoreArgs = list(graph = graph_adj[[1]]),
                SIMPLIFY = FALSE)
  out <- out[sapply(out, length) > 0]
  out <- unlist(out, recursive = FALSE)
  out2 <- unname(lapply(out, names))

  to_x_y <- function(x) {
    out <- data.frame(x = x[1],
                y = x[length(x)])
    out
  }

  out3 <- lapply(
            out2,
            to_x_y
          )
  out4 <- do.call(
            rbind,
            out3
          )
  out4 <- out4[!duplicated(out4), ]
  rownames(out4) <- NULL
  out4
}

lav_y_on_x <- function(
  x,
  y,
  yx_table) {
  out <- (yx_table$x %in% x) &
         (yx_table$y %in% y)
  any(out)
}

all_possible_loadings <- function(
  sem_out
) {
  all_ins <- lavaan::lavNames(
                sem_out,
                type = "ov.ind"
              )
  all_lv <- lavaan::lavNames(
                sem_out,
                type = "lv"
              )
  out <- expand.grid(
            lhs = all_lv,
            op = "=~",
            rhs = all_ins,
            stringsAsFactors = FALSE
          )
  out
}

all_nonxy_loadings <- function(
  sem_out
) {
  pt <- lavaan::parameterTable(sem_out)
  lambdas0 <- all_possible_loadings(sem_out)
  ind <- unique(lambdas0$rhs)
  # lambdas0 <- lapply(
  #               seq_len(nrow(lambdas0)),
  #               function(x) {
  #                 c(lhs = lambdas0[x, "lhs"],
  #                   rhs = lambdas0[x, "rhs"])
  #               })
  i1 <- pt$op == "=~"
  i2 <- (pt$free > 0) |
        ((pt$free == 0) & (pt$est != 0))
  lambdas1 <- pt[i1 & i2, c("lhs", "op", "rhs")]
  # lambdas1 <- lapply(
  #               seq_len(nrow(lambdas1)),
  #               function(x) {
  #                 c(lhs = lambdas1[x, "lhs"],
  #                   rhs = lambdas1[x, "rhs"])
  #               })
  fxy <- xy_lav(sem_out)
  # fxy <- lapply(
  #               seq_len(nrow(fxy)),
  #               function(x) {
  #                 c(x = fxy[x, "x"],
  #                   y = fxy[x, "y"])
  #               })
  out <- numeric()
  for (x in ind) {
    a2 <- lambdas1$lhs[lambdas1$rhs == x]
    b1 <- fxy$y[fxy$x == a2]
    b2 <- fxy$x[fxy$y == a2]
    b <- union(b1, b2)
    i1 <- which((lambdas0$lhs %in% b) &
                (lambdas0$rhs == x))
    i2 <- which((lambdas0$lhs %in% a2) &
                (lambdas0$rhs == x))
    out <- c(out, i1, i2)
  }
  out <- unique(out)
  lambdas2 <- lambdas0[-out, ]
  lambdas2 <- lambdas2[
                order(lambdas2$lhs,
                      lambdas2$rhs),
              ]
  rownames(lambdas2) <- NULL
  lambdas2
}

suppressMessages(library(lavaan))
mod <-
"
fx1 =~ x1 + x2 + x3
fx2 =~ x4 + x5 + x6
fy1 =~ x7 + x8 + x9
fy2 =~ x10 + x11 + x12
fy3 =~ x13 + x14 + x15
fy1 ~ fx1 + fx2
fy2 ~ fx2
fy3 ~ fy1 + fy2
"

fit <- sem(mod,
           dat_sem,
           warn = FALSE)
pt <- parameterTable(fit)
out1 <- modelbpp:::xy_lav(fit)

expect_true(nrow(out1) == 7)
expect_true(modelbpp:::lav_y_on_x("fy1", "fy3", out1))
expect_true(modelbpp:::lav_y_on_x("fx2", "fy3", out1))
expect_false(modelbpp:::lav_y_on_x("fx2", "fx1", out1))

# TOOD:
# - Do we need to exclude `fx1 =~ x10`.
out2 <- modelbpp:::all_nonxy_loadings(fit)
tmp <- unique(out1$x)[1]
fx1_inds <- pt[(pt$lhs == tmp) & (pt$op == "=~"), "rhs"]
fx1to <- out1[out1$x == tmp, "y"]
fx1to_inds <- pt[(pt$lhs %in% fx1to) & (pt$op == "=~"), "rhs"]
expect_true(length(intersect(out2[out2$lhs == tmp, "rhs"],
                             fx1to_inds)) == 0)

out3 <- modelbpp:::lav_cross_set(fit, type = "pure_x")
expect_true(!any(out3$lhs %in% out1$y))
expect_true(all(out3$lhs %in% out1$x))

out4 <- modelbpp:::lav_cross_set(fit, type = "pure_y")
expect_true(nrow(out4) == 0)

out5 <- modelbpp:::lav_cross_set(fit, type = "user", user_lav = c("fx1", "fy1", "fy3"))
fuser_inds <- pt[(pt$lhs %in% c("fx1", "fy1", "fy3")) & (pt$op == "=~"), "rhs"]
expect_true(setequal(fuser_inds, unique(out5$rhs)))

# modelbpp:::lav_loadings(fit, lav = c("fx1", "fy3", "fy1"))

suppressMessages(library(lavaan))

mod1 <-
"
m1 ~ x
m2 ~ m1
y ~ m2 + x
"
fit1 <- sem(mod1, dat_serial_4_weak, fixed.x = FALSE)

mod2 <-
"
x ~ m1 + m2
y ~ x
"
fit2 <- sem(mod2, dat_serial_4_weak, fixed.x = FALSE)

mod3 <-
"
x ~ m1 + m2
y ~ x + m2
"
fit3 <- sem(mod3, dat_serial_4_weak, fixed.x = FALSE)

mod4 <-
"
m1 ~ y
m2 ~ m1
x ~ m2 + y + m1
"
fit4 <- sem(mod4, dat_serial_4_weak, fixed.x = FALSE)

fits <- list(fit1 = fit1,
             fit2 = fit2,
             fit3 = fit3,
             fit4 = fit4)
pts <- lapply(fits, parameterTable)
class(pts) <- c("partables", class(pts))

out1_df2 <- model_set(fit1,
                      partables = pts,
                      progress = FALSE)
expect_false("fit1" %in% names(out1_df2$models))

#' @title Measurement Invariance Models
#'
#' @description Generate metric and
#' scalar invariance models and their
#' partial invariance versions.
#'
#' @details
#' This a helper function to generate,
#' based on a multigroup confirmatory
#' factor analysis (CFA) model with
#' no between-group equality constraints,
#' the following models:
#'
#' - A metric invariance model (loadings
#' constrained to be equal across
#' groups).
#'
#' - A scalar invariance model (intercepts
#' and loadings constrained to be equal
#' across groups).
#'
#' - Partial invariance versions of the
#' previous two models, such as a
#' model with the loadings of all items,
#' except for one, constrained to be
#' equal across groups.
#'
#' The models generated can then be used
#' in `model_set()` to compute BPPs.
#'
#' ## Requirements
#'
#' The model used as the input needs to
#' be fitted with no between group
#' constrains, that is, it is a
#' configural invariance model. Although
#' not a must, it is advised to use
#' the default way to identify each
#' factor (that is, fixing a loading
#' to one).
#'
#' ## Implementation
#'
#' This function simply use the
#' `group.partial` and `group.equal`
#' argument of [lavaan::cfa()] to
#' generate the models.
#'
#' @param cfa_out The output of
#' [lavaan::cfa()].
#'
#' @param max_free The maximum number of
#' constraints to be released when
#' generating the partial invariance
#' models. For example, if set to 1,
#' then only the partial metric invariance
#' model only has at most one item allowed
#' to have different loadings across
#' group. Default is 1. If set to zero,
#' then no partial invariance models
#' will be generated.
#'
#' @param metric Logical. If `TRUE`,
#' the default, then metric invariance
#' model and its partial invariance
#' versions are generated.
#'
#' @param scalar Logical. If `TRUE`,
#' the default, then scalar invariance
#' model and its partial invariance
#' versions are generated.
#'
#' @param progress Logical. If `TRUE`,
#' the default, progress bars will be
#' displayed when fitting partial
#' invariance models.
#'
#' @return A list of [lavaan::cfa()]
#' output. The names are automatically
#' generated to indicate whether a model
#' is configural, metric, or scalar
#' invariance, or the item(s) without
#' between-group constraints on the
#' loadings (for partial metric
#' invariance) or intercepts (for
#' partial scalar invariance).
#'
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>
#'
#' @seealso [model_set()]
#'
#' @examples
#'
#' library(lavaan)
#' # For illustration, only one factor is used,
#' # with one item from another factor added
#' # just to make the model not saturated.
#' HSmod <-
#' "
#' spatial =~ x1 + x2 + x3 + x4
#' "
#' fit_config <- cfa(model = HSmod,
#'                   data = HolzingerSwineford1939,
#'                   group = "school")
#' fit_mi <- measurement_invariance_models(fit_config)
#' names(fit_mi)
#' # Need to add 'skip_check_sem_out = TRUE' to use multigroup models.
#' out <- model_set(sem_out = fit_mi,
#'                  skip_check_sem_out = TRUE,
#'                  progress = FALSE,
#'                  parallel = FALSE)
#' print(out)
#'
#' @export
#'

measurement_invariance_models <- function(cfa_out,
                                          max_free = 1,
                                          metric = TRUE,
                                          scalar = TRUE,
                                          progress = TRUE) {
    pt <- lavaan::parameterTable(cfa_out)
    if (!metric && !scalar) {
        return(list(config = cfa_out))
      }
    if (lavaan::lavInspect(cfa_out, "ngroups") == 1) {
        stop("Cannot be used on single-group models.")
      }
    opt <- lavaan::lavInspect(cfa_out, "options")
    if (length(opt$group.equal) != 0) {
        stop("Cannot be used models with 'group.equal' set.")
      }
    if (length(opt$group.partial) != 0) {
        stop("Cannot be used models with 'group.partial' set.")
      }
    cfa_out@call$model <- eval(cfa_out@call$model,
                               envir = parent.frame())
    if (metric) {
        fit_metric <- lavaan::update(cfa_out,
                                     group.equal = "loadings")
        fit_pi_metric <- partial_invariance(fit_metric,
                                            pars = "loadings",
                                            max_free = max_free,
                                            keep_original = FALSE,
                                            progress = progress)
        fit_metric <- list(metric = fit_metric)
      } else {
        fit_metric <- NULL
        fit_pi_metric <- NULL
      }
    if (scalar) {
        fit_scalar <- lavaan::update(cfa_out,
                                     group.equal = c("loadings", "intercepts"))
        fit_pi_scalar <- partial_invariance(fit_scalar,
                                            pars = "intercepts",
                                            max_free = max_free,
                                            keep_original = FALSE,
                                            progress = progress)
        fit_scalar <- list(scalar = fit_scalar)
      } else {
        fit_scalar <- NULL
        fit_pi_scalar <- NULL
      }
    out <- c(list(config = cfa_out),
                  fit_metric,
                  fit_scalar,
                  fit_pi_metric,
                  fit_pi_scalar)
    out
  }

#' @noRd

partial_invariance <- function(cfa_out,
                                pars = c("loadings", "intercepts"),
                                max_free = 1,
                                keep_original = TRUE,
                                progress = TRUE) {
    if (max_free < 1) {
        return(NULL)
      }
    pars <- match.arg(pars)
    opt <- lavaan::lavInspect(cfa_out, "options")
    if ((pars == "loadings") &&
        opt$group.equal != "loadings") {
        stop("To use pars = 'loadings', group.equal must be 'loadings'.")
      }
    chk <- opt$group.equal
    if ((pars == "intercepts") &&
        !setequal(opt$group.equal, c("loadings", "intercepts"))) {
        stop("To use pars = 'intercepts', group.equal must be c('loadings', 'intecepts').")
      }
    free_pars <- switch(pars,
                        loadings = get_free_loadings(cfa_out),
                        intercepts = get_free_intercepts(cfa_out))
    to_release <- lapply(seq_len(max_free), function(xx) {
          utils::combn(free_pars, m = xx, simplify = FALSE)
        })
    to_release <- unlist(to_release, recursive = FALSE)
    gp_eq <- switch(pars,
                    loadings = c("loadings"),
                    intercepts = c("loadings", "intercepts"))
    pt <- lavaan::parameterTable(cfa_out)
    fit_i <- function(x) {
            sem_out_update <- lavaan::update(cfa_out,
                                             group.equal = gp_eq,
                                             group.partial = x)
            sem_out_update
          }
    if (progress) {
        cat("\nFitting", length(to_release),
            "partial",
            switch(pars, loadings = "metric",
                         intercepts = "scalar"),
            "invariance models:\n")
        pt_out <- pbapply::pblapply(to_release, fit_i)
      } else {
        pt_out <- lapply(to_release, fit_i)
      }
    pt_out_names <- sapply(to_release, paste, collapse = ";")
    names(pt_out) <- pt_out_names
    if (keep_original) {
        if (pars == "loadings") {
            pt_out$metric <- cfa_out
          }
        if (pars == "intercepts") {
            pt_out$scalar <- cfa_out
          }
      }
    pt_out
  }

#' @noRd

get_free_loadings <- function(cfa_out) {
    pt <- lavaan::parameterTable(cfa_out)
    loadings <- pt[(pt$op == "=~") & (pt$free != 0) & (pt$group == 1), ]
    out <- sapply(split(loadings, seq_len(nrow(loadings))), function(xx) {
                paste0(xx$lhs, xx$op, xx$rhs)
              })
    out
  }

#' @noRd

get_free_intercepts <- function(cfa_out) {
    pt <- lavaan::parameterTable(cfa_out)
    intercepts <- pt[(pt$op == "~1") & (pt$free != 0) & (pt$group == 1), ]
    out <- sapply(split(intercepts, seq_len(nrow(intercepts))), function(xx) {
                paste0(xx$lhs, xx$op)
              })
    out
  }

if (TRUE &&
    interactive() &&
    length(unclass(packageVersion("modelbpp"))[[1]]) == 4) {

suppressMessages(library(lavaan))

dat <- dat_sem
set.seed(1234)
dat_sem$group <- sample(c("gp1", "gp2", "gp3"), size = nrow(dat_sem), replace = TRUE)

mod <-
"
f1 =~ x1 + x2 + x3 + x4
f2 =~ x5 + x6 + x7 + x8
"
fit_config <- cfa(mod, dat_sem, group = "group")
fit_metric <- cfa(mod, dat_sem, group = "group",
                  group.equal = "loadings")
fit_scalar <- cfa(mod, dat_sem, group = "group",
                  group.equal = c("intercepts", "loadings"))

# fit_partial_metric <- partial_invariance(fit_metric, max_free = 2, pars = "loadings")
# fit_partial_scalar <- partial_invariance(fit_scalar, max_free = 2, pars = "intercepts")
# names(fit_partial_metric)
# names(fit_partial_scalar)
fit_mi <- measurement_invariance_models(fit_config,
                                        progress = FALSE)
expect_equal(length(fit_mi), 17)
expect_true(all(c("config", "metric", "scalar") %in% names(fit_mi)))
expect_true(setequal(c(57, 69, 81, 67, 79), unique(sapply(fit_mi, fitMeasures, "df"))))

fit_mi2 <- measurement_invariance_models(fit_config,
                                         max_free = 0,
                                         progress = FALSE)
expect_equal(length(fit_mi2), 3)
expect_true(setequal(c("config", "metric", "scalar"), names(fit_mi2)))
expect_true(setequal(c(57, 69, 81), unique(sapply(fit_mi2, fitMeasures, "df"))))

fit_mi3 <- measurement_invariance_models(fit_config,
                                         scalar = FALSE,
                                         max_free = 2,
                                         progress = FALSE)
expect_equal(length(fit_mi3), 23)
expect_false("scalar" %in% names(fit_mi3))
expect_true(setequal(c(57, 69, 67, 65), unique(sapply(fit_mi3, fitMeasures, "df"))))

fit_mi4 <- measurement_invariance_models(fit_config,
                                         metric = FALSE,
                                         max_free = 1,
                                         progress = FALSE)
expect_equal(length(fit_mi4), 10)
expect_false("metric" %in% names(fit_mi4))
expect_true(setequal(c(57, 81, 79), unique(sapply(fit_mi4, fitMeasures, "df"))))

out <- model_set(sem_out = fit_mi4,
                 skip_check_sem_out = TRUE,
                 progress = FALSE,
                 parallel = FALSE)
expect_true(is.null(out$equivalent_clusters) ||
            length(out$equivalent_clusters) == 0)
}

#' @title Two or More Hypothesized Models
#'
#' @description Combine the 'model_set()'
#' results of two or more hypothesis
#' models.
#'
#' @details
#' There are cases in which users
#' have more than one hypothesized model,
#' each with its own set of neighboring
#' models.
#'
#' The function [model_set_combined()]
#' let users combine the [model_set()]
#' results two or more hypothesized model.
#' Users can then compare the BPPs
#' of these hypothesized models, as well
#' as their neighboring models. Equivalent
#' models will be removed in the process.
#'
#' @return
#' A `model_set`-class object, which is
#' simply an output of [model_set()].
#' All methods and functions for the
#' output of [model_set()] will also
#' work on this object.
#'
#' @param model_set_outputs This must be
#' a named list of the outputs of
#' [model_set()]. The names will be used
#' as prefixes to name the models.
#'
#' @param ... Additional arguments to be
#' passed to [model_set()].
#'
#' @seealso [model_set()]
#'
#' @examples
#'
#' library(lavaan)
#'
#' mod1 <-
#' "
#' x4 ~ x1
#' x7 ~ x4
#' "
#'
#' mod2 <-
#' "
#' x1 ~ x4
#' x7 ~ x4
#' "
#'
#' fit1 <- sem(mod1,
#'             HolzingerSwineford1939,
#'             fixed.x = FALSE)
#' fit2 <- sem(mod2,
#'             HolzingerSwineford1939,
#'             fixed.x = FALSE)
#'
#' out1 <- model_set(fit1)
#' out2 <- model_set(fit2)
#'
#' out1
#' out2
#'
#' outb <- model_set_combined(
#'             list(fit1 = out1,
#'                  fit2 = out2))
#'
#' outb
#'
#' @export
model_set_combined <- function(model_set_outputs,
                               ...) {

  # ---- Sanity Checks ----
  chk_class <- sapply(
                  model_set_outputs,
                  inherits,
                  what = "model_set")
  if (!all(chk_class)) {
    i <- names(chk_class)[!chk_class]
    tmp <- paste0(i,
                  collapse = ", ")
    stop("Some outputs are not from 'model_set()': ",
         tmp)
  }

  fit_names <- names(model_set_outputs)

  if (isTRUE(is.null(fit_names))) {
    stop("model_set_outputs must be a named list")
  }

  if (!isTRUE(all(sapply(fit_names, nchar) > 0))) {
    stop("Some elements have no valid names.")
  }

  # ---- Combine the Models ----

  f <- function(out, fit_name) {
          fits <- out$fit
          names(fits) <- paste0(fit_name, "_", names(fits))
          fits
        }
  all_fits <- mapply(
                f,
                out = model_set_outputs,
                fit_name = fit_names,
                SIMPLIFY = FALSE,
                USE.NAMES = TRUE
              )

  all_fits <- unlist(
                unname(all_fits),
                recursive = FALSE)

  # ---- Do the Analysis ----

  out <- model_set(all_fits,
                   ...)
  out
}

#' @noRd
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>

catwrap <- function(x,
                    width = 0.9 * getOption("width"),
                    indent = 0,
                    exdent = 0,
                    prefix = "",
                    simplify = TRUE,
                    initial = prefix,
                    sep = "\n",
                    fill = FALSE,
                    labels = NULL,
                    append = FALSE) {
    out <- strwrap(x,
                   width = width,
                   indent = indent,
                   exdent = exdent,
                   prefix = prefix,
                   simplify = simplify,
                   initial = initial)
    cat(out,
        sep = sep,
        fill = fill,
        labels = labels,
        append = append)
  }

#' @noRd
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>

auto_tab <- function(x,
                     y,
                     between = ": ") {
    x_max_length <- max(nchar(x))
    y_max_length <- max(nchar(y))
    y2 <- pad_whitespace(y,
                         align = "right",
                         common_length = y_max_length)
    x2 <- pad_whitespace(x,
                         align = "left",
                         common_length = x_max_length)
    y2 <- unname(y2)
    x2 <- unname(x2)
    out <- paste0(x2, between, y2)
    out
  }

#' @noRd
#' @author Shu Fai Cheung <https://orcid.org/0000-0002-9871-9448>

pad_whitespace <- function(object,
                           align = c("left", "right"),
                           common_length = NA) {
    align <- match.arg(align)
    out <- sapply(object, function(yy) {
            tmp <- common_length - nchar(yy)
            if (tmp > 0) {
                tmp2 <- paste0(rep(" ", tmp),
                              collapse = "")
                tmp3 <- switch(align,
                          right = paste0(tmp2, yy),
                          left = paste0(yy, tmp2))
                return(tmp3)
              } else {
                return(yy)
              }
          })
    out
  }

---
title: "Get Started"
author: "Shu Fai Cheung"
date: "2026-05-12"
output:
  rmarkdown::html_vignette:
    number_sections: true
vignette: >
  %\VignetteIndexEntry{Get Started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
csl: apa.csl
---



# Introduction

This article illustrates how to
use `model_set()` and other functions
from the package
[`modelbpp`](https://sfcheung.github.io/modelbpp/)
to:

- Fit a set of neighboring models,
  each has one more or one less degree
  of freedom than the original fitted
  model.

- Compute the BIC posterior probability
  (BPP), for each model [@wu_simple_2020].

- Use BPP to assess to what extent
  each model is supported by the
  data, compared to all other models
  under consideration.

An
introduction to the package can be found in the following
article:

- Pesigan, I. J. A., Cheung, S. F., Wu, H., Chang, F., & Leung, S. O. (2026).
  How plausible is my model? Assessing model plausibility of
  structural equation models using Bayesian posterior probabilities (BPP).
  *Behavior Research Methods, 58*(3), Article 73.
  https://doi.org/10.3758/s13428-025-02921-x

# Workflow

1. Fit an SEM model, the
    original model, as usual in `lavaan`.

2. Call `model_set()` on the output from
  Step 1. It will automatically do the
  following:

    - Enumerate the neighboring models
      of the original model.

    - Fit all the models and compute their
      BIC posterior probabilities (BPPs).

3. Examine the results by:

    - printing the output of `model_set()`, or

    - generating a graph using `model_graph()`.

# Example

This is a sample dataset,
`dat_serial_4_weak`,
with four variables:


``` r
library(modelbpp)
head(dat_serial_4_weak)
#>             x         m1         m2          y
#> 1  0.09107195  1.2632493  0.7823926  0.3252093
#> 2 -1.96063838 -0.7745526  0.2002867 -0.6379673
#> 3  0.20184014  0.2238152  0.2374072  0.4205998
#> 4 -2.25521708  0.1185732 -0.1727878  0.9320889
#> 5 -0.15321350  0.1509888  1.1251386  0.6892537
#> 6 -2.00640303 -0.1595208 -0.1553136 -0.2364792
```

## Step 1: Fit the Original Model

Fit this original model, a serial
mediation model, with one direct path,
from `x` to `y`:


``` r
library(lavaan)
#> This is lavaan 0.6-21
#> lavaan is FREE software! Please report any bugs.
mod1 <-
"
m1 ~ x
m2 ~ m1
y ~ m2 + x
"
fit1 <- sem(mod1, dat_serial_4_weak)
```

This the summary:


``` r
summary(fit1,
        fit.measures = TRUE)
#> lavaan 0.6-21 ended normally after 1 iteration
#> 
#>   Estimator                                         ML
#>   Optimization method                           NLMINB
#>   Number of model parameters                         7
#> 
#>   Number of observations                           100
#> 
#> Model Test User Model:
#>                                                       
#>   Test statistic                                 3.376
#>   Degrees of freedom                                 2
#>   P-value (Chi-square)                           0.185
#> 
#> Model Test Baseline Model:
#> 
#>   Test statistic                                39.011
#>   Degrees of freedom                                 6
#>   P-value                                        0.000
#> 
#> User Model versus Baseline Model:
#> 
#>   Comparative Fit Index (CFI)                    0.958
#>   Tucker-Lewis Index (TLI)                       0.875
#> 
#> Loglikelihood and Information Criteria:
#> 
#>   Loglikelihood user model (H0)               -216.042
#>   Loglikelihood unrestricted model (H1)       -214.354
#>                                                       
#>   Akaike (AIC)                                 446.084
#>   Bayesian (BIC)                               464.320
#>   Sample-size adjusted Bayesian (SABIC)        442.212
#> 
#> Root Mean Square Error of Approximation:
#> 
#>   RMSEA                                          0.083
#>   90 Percent confidence interval - lower         0.000
#>   90 Percent confidence interval - upper         0.232
#>   P-value H_0: RMSEA <= 0.050                    0.261
#>   P-value H_0: RMSEA >= 0.080                    0.628
#> 
#> Standardized Root Mean Square Residual:
#> 
#>   SRMR                                           0.050
#> 
#> Parameter Estimates:
#> 
#>   Standard errors                             Standard
#>   Information                                 Expected
#>   Information saturated (h1) model          Structured
#> 
#> Regressions:
#>                    Estimate  Std.Err  z-value  P(>|z|)
#>   m1 ~                                                
#>     x                 0.187    0.059    3.189    0.001
#>   m2 ~                                                
#>     m1                0.231    0.091    2.537    0.011
#>   y ~                                                 
#>     m2                0.341    0.089    3.835    0.000
#>     x                 0.113    0.052    2.188    0.029
#> 
#> Variances:
#>                    Estimate  Std.Err  z-value  P(>|z|)
#>    .m1                0.278    0.039    7.071    0.000
#>    .m2                0.254    0.036    7.071    0.000
#>    .y                 0.213    0.030    7.071    0.000
```



The fit is acceptable, though the RMSEA
is marginal (CFI = 0.958,
RMSEA = 0.083).

## Step 2: Call `model_set()`

Use `model_set()` to find the
neighboring models differ
from the target model by one on model
degrees of freedom, fit them, and compute
the BPPs:


``` r
out1 <- model_set(fit1)
```

## Step 3: Examine the Results

To examine the results, just print
the output:


``` r
out1
#> 
#> Call:
#> model_set(sem_out = fit1)
#> 
#> Number of model(s) fitted           : 7
#> Number of model(s) converged        : 7
#> Number of model(s) passed post.check: 7
#> 
#> The models (sorted by BPP):
#>             model_df df_diff Prior     BIC   BPP   cfi rmsea  srmr
#> original           2       0 0.143 464.320 0.323 0.958 0.083 0.050
#> drop: y~x          3      -1 0.143 464.340 0.320 0.848 0.129 0.095
#> add: y~m1          1       1 0.143 465.900 0.147 1.000 0.000 0.018
#> drop: m2~m1        3      -1 0.143 465.953 0.143 0.800 0.148 0.116
#> add: m2~x          1       1 0.143 468.575 0.038 0.939 0.142 0.046
#> drop: m1~x         3      -1 0.143 469.403 0.025 0.695 0.183 0.124
#> drop: y~m2         3      -1 0.143 473.291 0.004 0.577 0.216 0.133
#> 
#> Note:
#> - BIC: Bayesian Information Criterion.
#> - BPP: BIC posterior probability.
#> - model_df: Model degrees of freedom.
#> - df_diff: Difference in df compared to the original/target model.
#> - To show cumulative BPPs, call print() with 'cumulative_bpp = TRUE'.
#> - At least one model has fixed.x = TRUE. The models are not checked for
#>   equivalence.
#> - Since Version 0.1.6.3, the default ways to handle factor loadings
#>   have changed. Check the NEWS by news(package = 'modelbpp') to see how
#>   to reproduce results from previous versions.
```



The total number of models examined,
including the original model, is 7.

(Note: The total number of models
was 9 in previous version. Please
refer to the Note in the printout for
the changes.)

The BIC posterior probabilities
(BPPs) suggest that
the original model is indeed the most
probable model based on BPP. However,
the model with the direct path
dropped, `drop: y~x`, only
has slightly lower BPP
(0.320)

This suggests that, with equal prior
probabilities [@wu_simple_2020], the
support for the model with the direct
and without the direct path have similar
support from the data based on BPP.

Alternatively, we can use `model_graph()`
to visualize the BPPs and model relations
graphically:


``` r
graph1 <- model_graph(out1)
plot(graph1)
```

<div class="figure" style="text-align: center">
<img src="graph1-1.png" alt="plot of chunk graph1"  />
<p class="caption">plot of chunk graph1</p>
</div>

Each node (circle) represents one model.
The larger the BPP, the larger the node.

The arrow points from a simpler
model (a model with larger model *df*)
to a more complicated model (a model
with smaller model *df*). If two models
are connected by an arrow, then
one model can be formed from another model
by adding or removing one free parameter
(e.g., adding or removing one path).

## Repeat Step 2 with User Prior

In real studies, not all models are
equally probable before having data
(i.e., not all models have equal
prior probabilities). A researcher
fits the original model because

- its
  prior probability is higher than other
  models, at least other neighboring
  models (otherwise, it is not worthy
  of collecting data assess thi original
  model), but

- the prior probability
  is not so high to eliminate the need
  for collecting data to see how much it is
  supported by data.

Suppose we decide that the prior probability
of the original model is .50: probable, but
still needs data to decide whether it is
empirically supported

This can be done by setting `prior_sem_out`
to the desired prior probability when calling
`model_set()`:


``` r
out1_prior <- model_set(fit1,
                        prior_sem_out = .50)
```

The prior probabilities of all other
models are equal. Therefore, with
nine models and the prior of the target
model being .50, the prior probability
of the other eight model is (1 - .50) / 8
or .0625.

This is the printout:


``` r
out1_prior
#> 
#> Call:
#> model_set(sem_out = fit1, prior_sem_out = 0.5)
#> 
#> Number of model(s) fitted           : 7
#> Number of model(s) converged        : 7
#> Number of model(s) passed post.check: 7
#> 
#> The models (sorted by BPP):
#>             model_df df_diff Prior     BIC   BPP   cfi rmsea  srmr
#> original           2       0 0.500 464.320 0.741 0.958 0.083 0.050
#> drop: y~x          3      -1 0.083 464.340 0.122 0.848 0.129 0.095
#> add: y~m1          1       1 0.083 465.900 0.056 1.000 0.000 0.018
#> drop: m2~m1        3      -1 0.083 465.953 0.055 0.800 0.148 0.116
#> add: m2~x          1       1 0.083 468.575 0.015 0.939 0.142 0.046
#> drop: m1~x         3      -1 0.083 469.403 0.010 0.695 0.183 0.124
#> drop: y~m2         3      -1 0.083 473.291 0.001 0.577 0.216 0.133
#> 
#> Note:
#> - BIC: Bayesian Information Criterion.
#> - BPP: BIC posterior probability.
#> - model_df: Model degrees of freedom.
#> - df_diff: Difference in df compared to the original/target model.
#> - To show cumulative BPPs, call print() with 'cumulative_bpp = TRUE'.
#> - At least one model has fixed.x = TRUE. The models are not checked for
#>   equivalence.
#> - Since Version 0.1.6.3, the default ways to handle factor loadings
#>   have changed. Check the NEWS by news(package = 'modelbpp') to see how
#>   to reproduce results from previous versions.
```

If the prior of the target is set to .50,
then, taking into account both the prior
probabilities and the data, the target
model is strongly supported by the data.

This is the output of `model_graph()`:


``` r
graph1_prior <- model_graph(out1_prior)
plot(graph1_prior)
```

<div class="figure" style="text-align: center">
<img src="out1_prior-1.png" alt="plot of chunk out1_prior"  />
<p class="caption">plot of chunk out1_prior</p>
</div>

# Advanced Options

## More Neighboring Models

If desired, we can enumerate models
"farther away" from the target model.
For example, we can set the
maximum difference
in model *df* to 2, to include models
having two more or two less *df* than
the original model:


``` r
out1_df2 <- model_set(fit1,
                      df_change_add = 2,
                      df_change_drop = 2)
```

This is the printout. By default, when there
are more than 20 models, only the top 20
models on BPP will be printed:


``` r
out1_df2
#> 
#> Call:
#> model_set(sem_out = fit1, df_change_add = 2, df_change_drop = 2)
#> 
#> Number of model(s) fitted           : 14
#> Number of model(s) converged        : 14
#> Number of model(s) passed post.check: 14
#> 
#> The models (sorted by BPP):
#>                  model_df df_diff Prior     BIC   BPP   cfi rmsea  srmr
#> original                2       0 0.071 464.320 0.269 0.958 0.083 0.050
#> drop: y~x               3      -1 0.071 464.340 0.267 0.848 0.129 0.095
#> add: y~m1               1       1 0.071 465.900 0.122 1.000 0.000 0.018
#> drop: m2~m1             3      -1 0.071 465.953 0.119 0.800 0.148 0.116
#> drop: m2~m1;y~x         4      -2 0.071 465.973 0.118 0.690 0.160 0.149
#> add: m2~x               1       1 0.071 468.575 0.032 0.939 0.142 0.046
#> drop: m1~x              3      -1 0.071 469.403 0.021 0.695 0.183 0.124
#> drop: m1~x;y~x          4      -2 0.071 469.423 0.021 0.585 0.185 0.144
#> add: m2~x;y~m1          0       2 0.071 470.154 0.015 1.000 0.000 0.000
#> drop: m1~x;m2~m1        4      -2 0.071 471.036 0.009 0.536 0.196 0.160
#> drop: y~m2              3      -1 0.071 473.291 0.003 0.577 0.216 0.133
#> drop: y~m2;y~x          4      -2 0.071 474.819 0.001 0.422 0.218 0.170
#> drop: m2~m1;y~m2        4      -2 0.071 474.924 0.001 0.419 0.219 0.163
#> drop: m1~x;y~m2         4      -2 0.071 478.374 0.000 0.314 0.238 0.183
#> 
#> Note:
#> - BIC: Bayesian Information Criterion.
#> - BPP: BIC posterior probability.
#> - model_df: Model degrees of freedom.
#> - df_diff: Difference in df compared to the original/target model.
#> - To show cumulative BPPs, call print() with 'cumulative_bpp = TRUE'.
#> - At least one model has fixed.x = TRUE. The models are not checked for
#>   equivalence.
#> - Since Version 0.1.6.3, the default ways to handle factor loadings
#>   have changed. Check the NEWS by news(package = 'modelbpp') to see how
#>   to reproduce results from previous versions.
```

The number of models examined, including
the original model, is 14.

This is the output of `model_graph()`:


``` r
graph1_df2 <- model_graph(out1_df2,
                          node_label_size = .75)
plot(graph1_df2)
```

<div class="figure" style="text-align: center">
<img src="graph1_df2-1.png" alt="plot of chunk graph1_df2"  />
<p class="caption">plot of chunk graph1_df2</p>
</div>

Note: Due to the number of nodes,
`node_label_size` is used to reduce
the size of the labels.

## Excluding Some Parameters From the Search

When calling `model_set()`, users can
specify parameters that must be excluded
from the list to be added (`must_not_add`),
or must not be dropped (`must_not_drop`).

For example, suppose it is well
established that `m1~x` exists and should
not be dropped, we can exclude it
when calling `model_set()`


``` r
out1_no_m1_x <- model_set(fit1,
                          must_not_drop = "m1~x")
```

This is the output:


``` r
out1_no_m1_x
#> 
#> Call:
#> model_set(sem_out = fit1, must_not_drop = "m1~x")
#> 
#> Number of model(s) fitted           : 6
#> Number of model(s) converged        : 6
#> Number of model(s) passed post.check: 6
#> 
#> The models (sorted by BPP):
#>             model_df df_diff Prior     BIC   BPP   cfi rmsea  srmr
#> original           2       0 0.167 464.320 0.331 0.958 0.083 0.050
#> drop: y~x          3      -1 0.167 464.340 0.328 0.848 0.129 0.095
#> add: y~m1          1       1 0.167 465.900 0.150 1.000 0.000 0.018
#> drop: m2~m1        3      -1 0.167 465.953 0.147 0.800 0.148 0.116
#> add: m2~x          1       1 0.167 468.575 0.040 0.939 0.142 0.046
#> drop: y~m2         3      -1 0.167 473.291 0.004 0.577 0.216 0.133
#> 
#> Note:
#> - BIC: Bayesian Information Criterion.
#> - BPP: BIC posterior probability.
#> - model_df: Model degrees of freedom.
#> - df_diff: Difference in df compared to the original/target model.
#> - To show cumulative BPPs, call print() with 'cumulative_bpp = TRUE'.
#> - At least one model has fixed.x = TRUE. The models are not checked for
#>   equivalence.
#> - Since Version 0.1.6.3, the default ways to handle factor loadings
#>   have changed. Check the NEWS by news(package = 'modelbpp') to see how
#>   to reproduce results from previous versions.
```

The number of models reduced to 14.

This is the plot:


``` r
out1_no_m1_x <- model_graph(out1_no_m1_x)
plot(out1_no_m1_x)
```

<div class="figure" style="text-align: center">
<img src="out1_no_m1_x-1.png" alt="plot of chunk out1_no_m1_x"  />
<p class="caption">plot of chunk out1_no_m1_x</p>
</div>

## Models With Constraints

If the original model has equality
constraints, they will be included in
the search for neighboring models,
by default. That is, removing one
equality constraint between two models
is considered as a model with an
increase of 1 *df*.

## Recompute BPPs Without Refitting the Models

Users can examine the impact of the prior
probability of the original model without
refitting the models, by using
the output of `model_set()` as the
input, using the `model_set_out` argument:


``` r
out1_new_prior <- model_set(model_set_out = out1,
                            prior_sem_out = .50)
```

The results are identical to calling
`model_set()` with the original `lavaan`
output as the input:


``` r
out1_new_prior
#> 
#> Call:
#> model_set(model_set_out = out1, prior_sem_out = 0.5)
#> 
#> Number of model(s) fitted           : 7
#> Number of model(s) converged        : 7
#> Number of model(s) passed post.check: 7
#> 
#> The models (sorted by BPP):
#>             model_df df_diff Prior     BIC   BPP   cfi rmsea  srmr
#> original           2       0 0.500 464.320 0.741 0.958 0.083 0.050
#> drop: y~x          3      -1 0.083 464.340 0.122 0.848 0.129 0.095
#> add: y~m1          1       1 0.083 465.900 0.056 1.000 0.000 0.018
#> drop: m2~m1        3      -1 0.083 465.953 0.055 0.800 0.148 0.116
#> add: m2~x          1       1 0.083 468.575 0.015 0.939 0.142 0.046
#> drop: m1~x         3      -1 0.083 469.403 0.010 0.695 0.183 0.124
#> drop: y~m2         3      -1 0.083 473.291 0.001 0.577 0.216 0.133
#> 
#> Note:
#> - BIC: Bayesian Information Criterion.
#> - BPP: BIC posterior probability.
#> - model_df: Model degrees of freedom.
#> - df_diff: Difference in df compared to the original/target model.
#> - To show cumulative BPPs, call print() with 'cumulative_bpp = TRUE'.
#> - At least one model has fixed.x = TRUE. The models are not checked for
#>   equivalence.
#> - Since Version 0.1.6.3, the default ways to handle factor loadings
#>   have changed. Check the NEWS by news(package = 'modelbpp') to see how
#>   to reproduce results from previous versions.
```

## Many Neighboring Models

When a model has a lot of free parameters,
the number of neighboring models can be
large and it will take a long time to
fit all of them. Users can enable
parallel processing by setting `parallel`
to `TRUE` when calling `model_set()`.

## More Options

Please refer to the help page of
`model_set()` for options available.

# Further Information

For further information on other
functions,
please refer to their help pages.

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.model_set.R
\name{print.model_set}
\alias{print.model_set}
\title{Print a \code{model_set}-Class Object}
\usage{
\method{print}{model_set}(
  x,
  bic_digits = 3,
  bpp_digits = 3,
  sort_models = TRUE,
  max_models = 20,
  bpp_target = NULL,
  target_name = "original",
  more_fit_measures = c("cfi", "rmsea", "srmr"),
  fit_measures_digits = 3,
  short_names = FALSE,
  cumulative_bpp = FALSE,
  ...
)
}
\arguments{
\item{x}{A \code{model_set}-class object.}

\item{bic_digits}{The number of
decimal places to be displayed
for BIC.
Default is 3.}

\item{bpp_digits}{The number of
decimal places to be displayed
for BIC posterior probability
and prior probabilities.
Default is 3.}

\item{sort_models}{Whether the models
will be sorted by BIC posterior
probability.
Default is \code{TRUE}.}

\item{max_models}{The maximum number
of models to be printed. Default is
20.}

\item{bpp_target}{The desired
BIC probability. Used to compute
and print
the minimum prior probability
of the target model required to
achieve \code{bpp_target}. Default
is \code{NULL}.}

\item{target_name}{The name of the
target model as appeared in the
model list. Default is \code{"original"}.
Used if \code{bpp_target} is not \code{NULL}.}

\item{more_fit_measures}{Character
vector. To be passed to
\code{\link[lavaan:fitMeasures]{lavaan::fitMeasures()}}. Default
is \code{c("cfi", "rmsea", "srmr")}. Set it to
\code{NULL} to disable printing additional
fit measures.}

\item{fit_measures_digits}{The number of
decimal places to be displayed
for additional fit measures, if
requested. Default is 3.}

\item{short_names}{If \code{TRUE},
then simple short names will be
printed
along with full model names.
Default is \code{FALSE}. Short names
can be used when interpreting
the graph from \code{model_graph()}
if short names are used in the graph.}

\item{cumulative_bpp}{If \code{TRUE} and
the models are sorted by BPPs,
cumulative BPPs will be printed.
Default is \code{FALSE}.}

\item{...}{Optional arguments.
Ignored.}
}
\value{
\code{x} is returned invisibly.
Called for its side effect.
}
\description{
Print the content of
a \code{model_set}-class object.
}
\details{
It is the print method of the
output of \code{\link[=model_set]{model_set()}}.
}
\examples{

library(lavaan)

dat <- dat_path_model

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)

out <- model_set(fit)
out

}
\seealso{
A \code{model_set}-class object
is generated by \code{\link[=model_set]{model_set()}}.
}
\author{
Shu Fai Cheung \url{https://orcid.org/0000-0002-9871-9448}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.sem_outs.R
\name{print.sem_outs}
\alias{print.sem_outs}
\title{Print an \code{sem_outs}-Class Object}
\usage{
\method{print}{sem_outs}(x, max_models = 20, ...)
}
\arguments{
\item{x}{An \code{sem_outs}-class object.}

\item{max_models}{The maximum number
of models to be printed. Default
is 20.}

\item{...}{Optional arguments.
Ignored.}
}
\value{
\code{x} is returned invisibly.
Called for its side effect.
}
\description{
Print the content of
an \code{sem_outs}-class object.
}
\details{
The print method for the output
of \code{fit_many()}.
}
\examples{

library(lavaan)
dat <- dat_path_model
mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"
fit <- sem(mod, dat_path_model, fixed.x = TRUE)
mod_to_add <- get_add(fit)
out <- fit_many(mod_to_add, fit)
out
print(out, max_models = 1)

}
\seealso{
An \code{sem_outs}-class object
is generated by \code{fit_many()}.
}
\author{
Shu Fai Cheung \url{https://orcid.org/0000-0002-9871-9448}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dat_path_model_p06}
\alias{dat_path_model_p06}
\title{A Sample Dataset Based On a
Complex Path Model (For Testing)}
\format{
A data frame with six variables:
\describe{
\item{x1}{Predictor}
\item{x2}{Predictor}
\item{x3}{Predictor}
\item{y4}{Mediator}
\item{y5}{Mediator}
\item{y6}{Outcome}
}
}
\usage{
dat_path_model_p06
}
\description{
Generated from
a complex path model (n = 200).
}
\details{
The model used to generate
this dataset:

\if{html}{\out{<div class="sourceCode">}}\preformatted{y4 ~  x1 + x2 + x3
y5 ~  y4 + x1 + x2
y6 ~  y4 + y5 + x1 + x2 + x3
x1 ~~ x2 + x3
x2 ~~ x3
}\if{html}{\out{</div>}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_many.R
\name{fit_many}
\alias{fit_many}
\title{Fit a List of Models}
\usage{
fit_many(
  model_list,
  sem_out,
  original = NULL,
  parallel = FALSE,
  ncores = max(parallel::detectCores(logical = FALSE) - 1, 1),
  make_cluster_args = list(),
  progress = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{model_list}{A list of parameter
tables to be used by
\code{\link[lavaan:lavaan]{lavaan::lavaan()}} or \code{\link[=update]{update()}}.
Usually generated by \code{\link[=get_add]{get_add()}} or
\code{\link[=get_drop]{get_drop()}}.}

\item{sem_out}{The output from an
structural equation modeling
function. It currently supports
\link[lavaan:lavaan-class]{lavaan::lavaan} objects
only. Usually
the one used in \code{\link[=model_set]{model_set()}},
\code{\link[=get_add]{get_add()}} or
\code{\link[=get_drop]{get_drop()}} to generate the list of
models.}

\item{original}{String. If provided,
it should be a name of a model
in \code{model_list}, with which
differences in model degrees of
freedom will be computed for other
models. If \code{NULL}, the default,
then the model in \code{sem_out} will
be used to computed the differences
in model degrees of freedom. If \code{NA},
then differences in model \emph{df} will
not be computed.}

\item{parallel}{If \code{TRUE}, parallel
processing will be used to fit the
models. Default is \code{FALSE}.}

\item{ncores}{Numeric. The number of
CPU cores to be used if \code{parallel}
is \code{TRUE}.}

\item{make_cluster_args}{A list of
named arguments to be passed to
\code{parallel::makeCluster()}. Used by
advanced users to configure the
cluster if \code{parallel} is \code{TRUE}.
Default is \code{list()}.}

\item{progress}{Whether a progress
bar will be displayed, implemented
by the \code{pbapply} package. Default
is \code{TRUE}.}

\item{verbose}{Whether additional
messages will be displayed, such
as the expected processing time.
Default is \code{TRUE}.}
}
\value{
An object of the class
\code{sem_outs}, a list with the
following major elements:
\itemize{
\item \code{fit}: A named list of
\code{\link[lavaan:lavaan]{lavaan::lavaan()}} output objects or
\code{\link[=update]{update()}} for fitting a model with
the added parameters.
\item \code{change}: A numeric vector, of the
same length as \code{fit}. The change
in model \emph{df} for each fit compared
to the original model. A
positive number denotes one less
free parameter. A negative number
denotes one more free parameter or
one less constraint.
\item \code{converged}: A named vector of
boolean values, of the same length
as \code{fit}. Indicates whether each
fit converged or not.
\item \code{post_check}: A named vector of
boolean values, of the same length
as \code{fit}. Indicates whether the
solution of each fit is
admissible or not. Checked by
\code{\link[lavaan:lavInspect]{lavaan::lavInspect()}} with
the \code{what} argument set to
\code{"post.check"}.
}
}
\description{
Fit a list of models
to a dataset.
}
\details{
It receives a list of
models, defined by \code{lavaan} parameter
tables (usually generated by
\code{\link[=model_set]{model_set()}}, \code{\link[=get_add]{get_add()}} or
\code{\link[=get_drop]{get_drop()}}),
and fit them to a dataset stored
in a \code{lavaan}-class object.

This function is called by
\code{\link[=model_set]{model_set()}} and usually users do
not need to call it. It is exported
for advanced users.
}
\examples{

library(lavaan)
dat <- dat_path_model
mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"
fit <- sem(mod, dat_path_model, fixed.x = TRUE)
mod_to_add <- get_add(fit)
fit_many(mod_to_add, fit)

}
\author{
Shu Fai Cheung \url{https://orcid.org/0000-0002-9871-9448}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.partables.R
\name{print.partables}
\alias{print.partables}
\title{Print a \code{partables}-Class Object}
\usage{
\method{print}{partables}(x, max_tables = 10, ...)
}
\arguments{
\item{x}{A \code{partables}-class object.}

\item{max_tables}{The maximum number
of models to be printed.
Default is 10.}

\item{...}{Optional arguments.
Ignored.}
}
\value{
\code{x} is returned invisibly.
Called for its side effect.
}
\description{
Print the content of
a \code{partables}-class object.
}
\details{
The print method for the output
of \code{\link[=gen_models]{gen_models()}}, \code{\link[=get_add]{get_add()}},
and \code{\link[=get_drop]{get_drop()}}.
}
\examples{

library(lavaan)
dat <- dat_path_model
mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"
fit <- sem(mod, dat_path_model, fixed.x = TRUE)
mod_to_add <- get_add(fit)
mod_to_add
print(mod_to_add, max_tables = 1)
mod_to_drop <- get_drop(fit)
mod_to_drop
print(mod_to_drop, max_tables = 1)

}
\seealso{
\code{\link[=gen_models]{gen_models()}}, \code{\link[=get_add]{get_add()}},
and \code{\link[=get_drop]{get_drop()}}.
}
\author{
Shu Fai Cheung \url{https://orcid.org/0000-0002-9871-9448}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dat_sem}
\alias{dat_sem}
\title{A Sample Dataset Based On a
Structural Model (For Testing)}
\format{
An object of class \code{data.frame} with 250 rows and 16 columns.
}
\usage{
dat_sem
}
\description{
Generated from
a structural model with
latent variables (n = 250).
}
\details{
The model to be fitted:

\if{html}{\out{<div class="sourceCode">}}\preformatted{f1 =~ x1 + x2 + x3 + x4
f2 =~ x5 + x6 + x7 + x8
f3 =~ x9 + x10 + x11 + x12
f4 =~ x13 + x14 + x15 + x16
f3 ~ f1 + f2
f4 ~ f3
}\if{html}{\out{</div>}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dat_serial_4_weak}
\alias{dat_serial_4_weak}
\title{A Sample Dataset Based On a
Serial Mediation Model
With Weak Paths (For Testing)}
\format{
A data frame with four variables:
\describe{
\item{x}{Predictor}
\item{m1}{Mediator}
\item{m2}{Mediator}
\item{y}{Outcome}
}
}
\usage{
dat_serial_4_weak
}
\description{
Generated from
a serial mediation model (n = 100).
}
\details{
The model to be fitted:

\if{html}{\out{<div class="sourceCode">}}\preformatted{m1 ~ x
m2 ~ m1 + x
y  ~ m2 + m1 + x
}\if{html}{\out{</div>}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_set_combined.R
\name{model_set_combined}
\alias{model_set_combined}
\title{Two or More Hypothesized Models}
\usage{
model_set_combined(model_set_outputs, ...)
}
\arguments{
\item{model_set_outputs}{This must be
a named list of the outputs of
\code{\link[=model_set]{model_set()}}. The names will be used
as prefixes to name the models.}

\item{...}{Additional arguments to be
passed to \code{\link[=model_set]{model_set()}}.}
}
\value{
A \code{model_set}-class object, which is
simply an output of \code{\link[=model_set]{model_set()}}.
All methods and functions for the
output of \code{\link[=model_set]{model_set()}} will also
work on this object.
}
\description{
Combine the 'model_set()'
results of two or more hypothesis
models.
}
\details{
There are cases in which users
have more than one hypothesized model,
each with its own set of neighboring
models.

The function \code{\link[=model_set_combined]{model_set_combined()}}
let users combine the \code{\link[=model_set]{model_set()}}
results two or more hypothesized model.
Users can then compare the BPPs
of these hypothesized models, as well
as their neighboring models. Equivalent
models will be removed in the process.
}
\examples{

library(lavaan)

mod1 <-
"
x4 ~ x1
x7 ~ x4
"

mod2 <-
"
x1 ~ x4
x7 ~ x4
"

fit1 <- sem(mod1,
            HolzingerSwineford1939,
            fixed.x = FALSE)
fit2 <- sem(mod2,
            HolzingerSwineford1939,
            fixed.x = FALSE)

out1 <- model_set(fit1)
out2 <- model_set(fit2)

out1
out2

outb <- model_set_combined(
            list(fit1 = out1,
                 fit2 = out2))

outb

}
\seealso{
\code{\link[=model_set]{model_set()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_graph.R
\name{plot.model_graph}
\alias{plot.model_graph}
\title{Plot a Network of Models}
\usage{
\method{plot}{model_graph}(x, ...)
}
\arguments{
\item{x}{The output of
\code{\link[=model_graph]{model_graph()}}. (Named \code{x}
because it is required in the naming
of arguments of the \code{plot} generic
function.)}

\item{...}{Additional arguments,
passed to the \code{plot}-method of
an \code{igraph} object.}
}
\value{
\code{NULL}. Called for its
side effect.
}
\description{
Plot a network of models
generated by \code{model_graph()}.
}
\details{
This function is the plot
method of \code{model_graph} objects,
the output of
\code{\link[=model_graph]{model_graph()}}.

For now, it simply passes the object
to \code{plot}-method of an \code{igraph} object.
This function
is created for possible customization
of the plot in the future.
}
\examples{

library(lavaan)

dat <- dat_path_model

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)

out <- model_set(fit)
out

g <- model_graph(out)
plot(g)

}
\seealso{
\code{\link[=model_graph]{model_graph()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/partables_helpers.R
\name{partables_helpers}
\alias{partables_helpers}
\alias{identical_partables}
\alias{is_partable}
\alias{same_variables}
\alias{get_partables}
\alias{to_partables}
\title{Helper Functions For \code{partables}-Class Objects}
\usage{
identical_partables(object1, object2)

is_partable(x)

same_variables(x)

get_partables(x)

to_partables(...)
}
\arguments{
\item{object1}{A \code{lavaan} parameter
table or similar object.}

\item{object2}{A \code{lavaan} parameter
table or similar object.}

\item{x}{An object to be checked.}

\item{...}{The objects to be combined.}
}
\value{
The function \code{\link[=identical_partables]{identical_partables()}}
returns either \code{TRUE} or \code{FALSE}.

The function \code{\link[=is_partable]{is_partable()}}
returns either \code{TRUE} or \code{FALSE}.

The function \code{\link[=same_variables]{same_variables()}}
returns either \code{TRUE} or \code{FALSE}.

The function \code{\link[=get_partables]{get_partables()}}
returns a \code{partables}-class
object.

The function \code{\link[=to_partables]{to_partables()}}
returns a \code{partables}-class
object, created from the objects
supplied.
}
\description{
For tasks such as comparing two parameter
tables inside a \code{partables}-class object.
}
\details{
The function \code{\link[=identical_partables]{identical_partables()}} compare two
\code{lavaan} parameter tables and see whether they
are identical. (Adapted from a similar function
in the package \code{semhelpinghands}).

The function \code{\link[=is_partable]{is_partable()}} tries
to
check whether an object is "likely"
to be a parameter table that can be
used by \code{\link[lavaan:lavaan]{lavaan::lavaan()}} and its
wrappers, such as \code{\link[lavaan:sem]{lavaan::sem()}}.

Note that there is no guarantee the
the parameter table makes sense or
will not lead to error when fitted.
It can only check whether it has the
required columns.

The function \code{\link[=same_variables]{same_variables()}}
check whether all parameter labels
in a \code{partables}-class object use
the same observed variables.

The function \code{\link[=get_partables]{get_partables()}}
extract the \code{partable} object from
a \code{model_set}-class object.

The function \code{\link[=to_partables]{to_partables()}}
combine objects to create a
\code{partables}-class object.
The objects to be combined can be
a \code{lavaan}-class object (e.g.,
the output of \code{\link[lavaan:sem]{lavaan::sem()}})
or a parameter table of \code{lavaan}.
}
\examples{

library(lavaan)
mod1 <-
"
x3 ~ x1
x2 ~ x4
"
mod2 <-
"
x2 ~ x4
x3 ~ x1
"
fit1 <- sem(mod1, dat_path_model)
fit2 <- sem(mod2, dat_path_model)
pt1 <- parameterTable(fit1)
pt2 <- parameterTable(fit2)
identical_partables(pt1, pt2)


is_partable(pt1)


out <- model_set(fit1,
                 fit_models = FALSE)
same_variables(get_partables(out))



out <- model_set(fit1,
                 fit_models = FALSE)
get_partables(out)


fit1 <- sem(mod1, dat_path_model)
fit2 <- sem(mod2, dat_path_model)
pt1 <- parameterTable(fit1)
pt2 <- parameterTable(fit2)

to_partables(fit1, fit2)
to_partables(pt1, pt2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_graph.R
\name{model_graph}
\alias{model_graph}
\title{Generate a Graph of Models}
\usage{
model_graph(
  object,
  node_size_by_x = TRUE,
  x = NULL,
  node_size = 5,
  min_size = 5,
  max_size = 35,
  color_original = "lightblue",
  color_add = "burlywood1",
  color_drop = "lightgreen",
  color_others = "lightgrey",
  color_label = "black",
  node_label_size = 1,
  original = "original",
  drop_redundant_direct_paths = TRUE,
  label_arrow_by_df = NULL,
  arrow_label_size = 1,
  weight_arrows_by_df = c("inverse", "normal", "none"),
  arrow_min_width = 0.5,
  arrow_max_width = 2,
  progress = TRUE,
  short_names = FALSE,
  min_bpp_labelled = NULL,
  ...
)
}
\arguments{
\item{object}{Must be a
\code{model_set}-class object for now.}

\item{node_size_by_x}{Logical. Whether
node (vertex) sizes are determined
by a variable. Default is \code{TRUE}.
See \code{x} below on how size is
determined.}

\item{x}{If not \code{NULL}, it should be
a numeric vector of length equal to
the number of models. The node sizes
will be proportional to the values
of \code{x}, offset by \code{min_size}. If
\code{NULL}, the default, the BIC
posterior probabilities
stored in \code{object} will be retrieved.}

\item{node_size}{If \code{node_size_by_x}
is \code{FALSE}, this is the size for
all nodes.}

\item{min_size}{The minimum size
of a node. Default is 5.}

\item{max_size}{The maximum size
of a node. Default is 35.}

\item{color_original}{The color
of node of the original model.
Default is \code{"lightblue"}.}

\item{color_add}{The color of
the nodes of models formed by
adding one or more free parameters to
the original model.
Default is \code{"burlywood1"}.}

\item{color_drop}{The color of
the nodes of models formed by
dropping one or more free parameters
from the original model.
Default is \code{"lightgreen"}.}

\item{color_others}{The color
of other models not specified above.
Default is \code{"grey50"}.}

\item{color_label}{The color of the
text labels of the nodes. Default
is \code{"black"}.}

\item{node_label_size}{The size of
the labels of the nodes. Default is
1.}

\item{original}{String. The name
of the original model (target model).
Default is \code{"original"}.}

\item{drop_redundant_direct_paths}{Logical. Whether the redundant direct
path between two models. A direct path
is redundant if two models are also
connected through at least one
another model. Default is \code{TRUE}.}

\item{label_arrow_by_df}{If \code{TRUE},
then an arrow (edge) is always labelled
by the difference in model \emph{df}s.
If \code{FALSE}, then no arrows are
labelled. If \code{NULL}, then arrows are
labelled when not all differences
in model \emph{df}s are equal to one.
Default is \code{NULL}.}

\item{arrow_label_size}{The size of
the labels of the arrows (edges), if labelled.
Default is 1.}

\item{weight_arrows_by_df}{String.
Use if model \emph{df}
differences are stored.
If \code{"inverse"}, larger the
difference in model \emph{df}, \emph{narrower}
an arrow. That is, more similar two
models are, thicker the arrow. If
\code{"normal"}, larger the difference
in model \emph{df}, \emph{wider} an arrow.
If \code{"none"}, then arrow width is
constant, set to \code{arrow_max_width}.
Default is \code{"inverse"}.}

\item{arrow_min_width}{If
\code{weight_arrows_by_df} is not \code{"none"},
this is the minimum width of an
arrow.}

\item{arrow_max_width}{If
\code{weight_arrows_by_df} is not \code{"none"},
this is the maximum width of an
arrow. If \code{weight_arrows_by_df} is
\code{"none"}, this is the width of all
arrows.}

\item{progress}{Whether a progress
bar will be displayed for some
steps (e.g., checking for nested
relations). Default
is \code{TRUE}.}

\item{short_names}{If \code{TRUE} and
short model names are stored,
they will be used as model labels.
Please print the object with
\code{short_names = TRUE} to find the
corresponding full model names.}

\item{min_bpp_labelled}{If not \code{NULL},
this is the minimum BPP for a model
to be labelled. Models with BPP less
than this value will not be labelled.
Useful when the number of models
is large.}

\item{...}{Optional arguments. Not
used for now.}
}
\value{
A \code{model_graph}-class object that
can be used as as an \code{igraph}-object,
with a plot method (\code{\link[=plot.model_graph]{plot.model_graph()}})
with settings
suitable for plotting a network
of models with BIC posterior probabilities
computed.
}
\description{
Generate an 'igraph' object from a
'model_set' object.
}
\details{
It extracts the model list stored
in \code{object}, creates an adjacency
matrix, and then creates an \code{igraph}
object customized for visualizing
model relations.
\subsection{Construction of the Graph}{

This is the default way to construct
the graph when the model set is
automatically by \code{\link[=model_set]{model_set()}}.
\itemize{
\item Each model is connected by an
arrow, pointing from one model to
another model that

a. can be formed by adding one
or more free parameter, or

b. can be formed by releasing one
or more equality constraint
between two
parameters.

c. has nested relation with this
model as determined by the
method proposed by Bentler
and Satorra (2010), if the
models are not generated
internally.
}

That is, it points to a model with
more degrees of freedom (more
complicated),
and
is nested within that model in either
parameter sense or covariance sense.
\itemize{
\item By default, the size of the node
for each model is scaled by its
BIC posterior probability, if
available. See \emph{The Size of a Node}
below.
\item If a model is designated as the
original (target) model,
than he original model, the models
with
more degrees of freedom than the
original model,
and the models with fewer degrees
of freedom than the original models,
are colored differently.
\item The default layout is the Sugiyama
layout, with simpler models (models
with fewer degrees of freedom) on
the top. The lower a model is in
the network, the more the degrees
of freedom it has. This layout is
suitable for showing the nested
relations of the models. Models on
the same level (layer) horizontally
have the same model \emph{df}.
}

The output is an \code{igraph} object.
Users can customize it in any way
they want using functions from
the \code{igraph} package.

If a model has no nested relation
with all other model, it will not
be connected to other models.

If no model is named \code{original}
(default is \code{"original"}), then no
model is colored as the original model.
}

\subsection{User-Provided Models}{

If \code{object} contained one or more
user-provided models which are not
generated automatically by
\code{\link[=model_set]{model_set()}} or similar functions
(e.g., \code{\link[=gen_models]{gen_models()}}), then the
method by Bentler and Satorra (2010)
will be used to determine model
relations. Models connected by an
arrow has a nested relation based on
the NET method by Bentler and Satorra
(2010). An internal function inspired
by the \code{net} function from the
\code{semTools} package is used to
implement the NET method.
}

\subsection{The Size of a Node}{

When a model is scaled by \code{x},
which usually is the BIC posterior
probability, its size is determined
by:

\code{max_size * (x - min(x))/(max(x) - min(x)) + min_size}
}
}
\examples{

library(lavaan)

mod <-
"
m1 ~ x
y ~ m1
"

fit <- sem(mod, dat_serial_4, fixed.x = TRUE)

out <- model_set(fit)
out

g <- model_graph(out)
plot(g)

}
\references{
Bentler, P. M., & Satorra, A. (2010).
Testing model nesting and equivalence.
\emph{Psychological Methods, 15}(2),
111--123. \doi{10.1037/a0019625}
Asparouhov, T., & Muthén, B. (2019).
Nesting and Equivalence Testing for
Structural Equation Models.
\emph{Structural Equation Modeling: A Multidisciplinary Journal, 26}(2),
302--309. \doi{10.1080/10705511.2018.1513795}
}
\author{
Shu Fai Cheung \url{https://orcid.org/0000-0002-9871-9448}
The internal function for nesting
inspired by the \code{net} function
from the \code{semTools} package,
which was developed by
Terrence D. Jorgensen.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dat_cfa}
\alias{dat_cfa}
\title{A Sample Dataset Based On a
Confirmatory Factor Analysis Model
(For Testing)}
\format{
A data frame with six variables:
\describe{
\item{x1}{Indicator}
\item{x2}{Indicator}
\item{x3}{Indicator}
\item{x4}{Indicator}
\item{x5}{Indicator}
\item{x6}{Indicator}
}
}
\usage{
dat_cfa
}
\description{
Generated from
a confirmatory factor analysis
model (n = 200).
}
\details{
The model used to generate
this dataset:

\if{html}{\out{<div class="sourceCode">}}\preformatted{f1 =~ x1 + x2 + x3 + x5
f2 =~ x3 + x4 + x5 + x6
f1 ~~ f2
}\if{html}{\out{</div>}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_set.R
\name{model_set}
\alias{model_set}
\alias{gen_models}
\title{BIC Posterior Probabilities
of Neighboring Models}
\usage{
model_set(
  sem_out,
  partables = NULL,
  model_set_out = NULL,
  prior_sem_out = NULL,
  must_add = NULL,
  must_not_add = NULL,
  must_drop = NULL,
  must_not_drop = NULL,
  loadings_to_exclude_from_drop = c("single", "none", "all"),
  remove_constraints = TRUE,
  exclude_error_cov = TRUE,
  exclude_feedback = TRUE,
  exclude_xy_cov = TRUE,
  exclude_x_changed_to_y = TRUE,
  cross_add = c("pure_x", "pure_y"),
  cross_sets = NULL,
  df_change_add = 1,
  df_change_drop = 1,
  remove_duplicated = TRUE,
  fit_models = ifelse(!is.null(model_set_out$fit), FALSE, TRUE),
  compute_bpp = TRUE,
  original = "original",
  parallel = FALSE,
  ncores = max(parallel::detectCores(logical = FALSE) - 1, 1),
  make_cluster_args = list(),
  progress = TRUE,
  verbose = TRUE,
  skip_check_sem_out = FALSE,
  drop_equivalent_models = TRUE
)

gen_models(
  sem_out,
  must_add = NULL,
  must_not_add = NULL,
  must_drop = NULL,
  must_not_drop = NULL,
  loadings_to_exclude_from_drop = c("single", "none", "all"),
  remove_constraints = TRUE,
  exclude_error_cov = TRUE,
  exclude_feedback = TRUE,
  exclude_xy_cov = TRUE,
  exclude_x_changed_to_y = TRUE,
  cross_add = c("pure_x", "pure_y"),
  cross_sets = NULL,
  df_change_add = 1,
  df_change_drop = 1,
  remove_duplicated = TRUE,
  drop_equivalent_models = TRUE,
  progress = TRUE,
  output = c("partables", "model_set")
)
}
\arguments{
\item{sem_out}{It can be the output
from an
SEM function. Currently it supports
\link[lavaan:lavaan-class]{lavaan::lavaan} objects only.
If it is a named list of
\link[lavaan:lavaan-class]{lavaan::lavaan} objects, then
all arguments for model generation
will be ignored, and models will not
be refitted. Users need to ensure
that the models can be meaningfully
compared because they will not be
checked.}

\item{partables}{A \code{partables}-class
object, usually generated by
\code{\link[=get_add]{get_add()}} or \code{\link[=get_drop]{get_drop()}}. A named
list of parameter tables to be fitted
along with the original model in
\code{sem_out}.
If supplied, all arguments related
to identifying models will be ignored.
Default is \code{NULL}.}

\item{model_set_out}{If set to
the output of a previous call
to \code{\link[=model_set]{model_set()}} (a \code{model_set}-class
object), the list of stored models
will be used. All arguments related
to generate neighboring models will
be ignored.
If supplied, \code{sem_out} will also be
ignored and will be retrieved from
\code{model_set_out}, and \code{partables}
will also be ignored.
Default is \code{NULL}.}

\item{prior_sem_out}{The prior of the
model fitted in \code{sem_out}. Default
is \code{NULL}, and all models will have
equal prior probabilities.}

\item{must_add}{A character vector
of parameters, named in
\code{\link[lavaan:lavaan]{lavaan::lavaan()}} style (e.g.,
\code{"y ~ x"}), that must be added.
Default is \code{NULL}.}

\item{must_not_add}{A character
vector of parameters, named in
\code{\link[lavaan:lavaan]{lavaan::lavaan()}} style (e.g.,
\code{"x1 ~~ x1"}), that must not be
added. Default is \code{NULL}.}

\item{must_drop}{A character vector
of parameters, named in
\code{lavaan::lavaan()} style (e.g.,
\code{"y ~ x"}), that must be included.
Default is \code{NULL}.}

\item{must_not_drop}{A character
vector of parameters, named in
\code{\link[lavaan:lavaan]{lavaan::lavaan()}} style (e.g.,
\code{"x1 ~~ x1"}), that must not be
included. Default is \code{NULL}.}

\item{loadings_to_exclude_from_drop}{Whether factor loadings
will be excluded when deciding
parameters to drop. If \code{"single"},
then only "single" loadings (an indicator
loads on only one latent factor)
will be excluded (they will not be
dropped). If \code{"all"}, then
all factor loadings will be excluded
(no factor loadings will be dropped).
If \code{"none"}, then no loadings will
be excluded, and all free loadings
will be considered to be dropped. Be careful when using
\code{"none"} because the models may not
make sense. The settings in \code{must_drop}
and \code{must_not_drop} will override this
argument.}

\item{remove_constraints}{Whether
equality constraints will be
removed. Default is \code{TRUE}.}

\item{exclude_error_cov}{Exclude
error covariances of indicators.
Default is \code{TRUE}.}

\item{exclude_feedback}{Exclude
paths that will result in a feedback
loop. For example, if there is
path from \code{x} through \code{m} to \code{y},
then the path \code{x ~ y} will create
a feedback loop.  Default has been
changed to
\code{TRUE} since Version 0.1.3.5 because
feedback loops are usually
not included except when theoretically
justified. To reproduce results
based on previous version, set this
argument to \code{FALSE}.}

\item{exclude_xy_cov}{Exclude
covariance between two variables,
in which one has a path to another.
For example, if there is
path from \code{x} through \code{m} to \code{y},
then the covariance \code{x ~~ y},
which denotes the covariance between
\code{x} and the error term of \code{y}, will
be excluded if this argument is
\code{TRUE}. Default has been changed to
\code{TRUE} since Version 0.1.3.5 because
these covariances rarely are
interpretable. To reproduce results
based on previous version, set this
argument to \code{FALSE}.}

\item{exclude_x_changed_to_y}{Exclude
changes that make a "pure" \code{x}
variable to a \code{y} variable. For example,
adding the path \code{x2 ~ x1} to
the model \code{y ~ x1 + x2}.}

\item{cross_add}{A character vector
of whether and how cross-loadings
(an indicator loads on two or more
latent factors) will be added. If
\code{NULL}, no cross-loadings will be
added. If \code{"lav_x"} is in the vector,
then cross-loadings among "pure-x"
latent factors, latent
factors that do not regress on
any other variables, will be considered.
If \code{"lav_y"} is in the vector, then
cross-loadings among "pure-y" latent
factors, latent factors that regress
on at least one variables but they
themselves do not predict any other
variables, will be considered.
If \code{"user"} is in the vector, then
cross-loadings among latent factors listed
in \code{cross_sets} will be considered.}

\item{cross_sets}{A character vector
of latent variables for which cross-loadings
will be considered. Ignored if
\code{"user"} is not in \code{cross_add}.
Note that cross-loadings involving latent
variables which are mediators will
not be considered. To add them,
they must be specified explicitly
in \code{must_add}.}

\item{df_change_add}{How many degrees
of freedom (\emph{df}) away in the list.
All models with \emph{df} change less than
or equal to this number will be
included, taking into account
requirements set by other arguments.
Default is 1.}

\item{df_change_drop}{How many degrees
of freedom away in the list. All
models with \emph{df} change less than
or equal to this number will be
included, taking into account
requirements set by other arguments.
Default is 1.}

\item{remove_duplicated}{If \code{TRUE},
the default, duplicated models are
removed.}

\item{fit_models}{If \code{TRUE},
the
models will be fitted to the data,
usually stored in \code{sem_out}.
If \code{FALSE}, the models will be returned
as is, in the element \code{models}
of the output. If \code{model_set_out}
is set and models have been fitted,
then default is \code{FALSE}. Otherwise,
default is \code{TRUE}.}

\item{compute_bpp}{If \code{TRUE}, then BIC
posterior probabilities will be
computed. Default
is \code{TRUE}.}

\item{original}{String. The name of the
original (traget) model. Default is
\code{"original"}. Used if \code{prior_sem_out}
is unnamed and only has one value.}

\item{parallel}{If \code{TRUE}, parallel
processing will be used to fit the
models. Default is \code{FALSE}.}

\item{ncores}{Numeric. The number of
CPU cores to be used if \code{parallel}
is \code{TRUE}.}

\item{make_cluster_args}{A list of
named arguments to be passed to
\code{parallel::makeCluster()}. Used by
advanced users to configure the
cluster if \code{parallel} is \code{TRUE}.
Default is \code{list()}.}

\item{progress}{Whether a progress
bar will be displayed, implemented
by the \code{pbapply} package or by
\code{utils::txtProgressBar}. Default
is \code{TRUE}.}

\item{verbose}{Whether additional
messages will be displayed, such
as the expected processing time.
Default is \code{TRUE}.}

\item{skip_check_sem_out}{If \code{TRUE}
and \code{sem_out} is set, check
whether \code{sem_out} is of a supported
type (estimator is \code{"ML"} and
the model has only one group). If
not, an error will be raised.
Can be set to \code{FALSE} for experimenting
the functions on models not officially
supported.}

\item{drop_equivalent_models}{If
\code{TRUE}, the default, equivalent
models will be dropped in the final
output. This check can only be
conducted when no models are fitted
in \code{\link[lavaan:lavaan]{lavaan::lavaan()}} with
\code{fixed.x = TRUE} (which is the
default of \code{\link[lavaan:sem]{lavaan::sem()}}).}

\item{output}{If \code{"model_set"},
then the output is a \code{model_set}-class
object. If \code{"partables"}, the output
is a \code{partables}-class object.
Default is \code{partables}.}
}
\value{
The function \code{\link[=model_set]{model_set()}}
returns an object of the class
\code{model_set}, a list with the following
major elements:
\itemize{
\item \code{models}: A named list of parameter
tables. Each represent the models
identified.
\item \code{bic}: A numeric vector, of the
same length as \code{model}. The BIC of
each model.
\item \code{postprob}: A numeric vector, of
the same length as \code{model}. The
BIC posterior probability of each
model.
\item \code{fit}: A named list of
\code{\link[lavaan:lavaan]{lavaan::lavaan()}} output objects
or \code{\link[=update]{update()}} for fitting a model
with the added parameters, of the
same length as \code{model}.
\item \code{change}: A numeric vector, of the
same length as \code{model}. The change
in model df for each fit. A
positive number denotes one less
free parameter. A negative number
denotes one more free parameter or
one less constraint.
\item \code{converged}: A named vector of
boolean values, of the same length
as \code{model}. Indicates whether each
fit converged or not.
\item \code{post_check}: A named vector of
boolean values, of the same length
as \code{model}. Indicates whether the
solution of each fit is admissible
or not. Checked by
\code{\link[lavaan:lavInspect]{lavaan::lavInspect()}}.
}

The object returned by \code{\link[=gen_models]{gen_models()}}
depends on the argument \code{output}.
See the argument \code{output} for the
details
}
\description{
Identify neighboring
models, fit them, and return the
BIC posterior probabilities.
}
\details{
It computes the BIC
posterior probabilities of a set
of models by the method presented
in Wu, Cheung, and Leung (2020).

First, a list of model is identified
based on user-specified criteria.
By default, models differ from a fitted
model by one degree of freedom,
the 1-df-away \emph{neighboring} models,
will be found using \code{\link[=get_add]{get_add()}}
and \link{get_drop}.

Second, these models will be fitted
to the sample dataset, and their
BICs will be computed.

Third, their BIC posterior
probabilities will be computed
using their BICs. By default,
equal prior
probabilities for all the models
being fitted will be assumed in
the current version. This can be
changed by \code{prior_sem_out}.

The results can then be printed,
with the models sorted by descending
order of BIC posterior
probabilities. The results can
also be visualized using
\code{\link[=model_graph]{model_graph()}}.
}
\section{Functions}{
\itemize{
\item \code{model_set()}: Compute the BPPs of a list of models.
Can generate the models and/or fit the models. Can also
accept pregenerated models, or just update BPPs.

\item \code{gen_models()}: Generate a list of models (parameter tables).

}}
\examples{

library(lavaan)

dat <- dat_path_model

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)

out <- model_set(fit)
out

}
\references{
Wu, H., Cheung, S. F., & Leung, S. O.
(2020). Simple use of BIC to assess
model selection uncertainty: An
illustration using mediation and
moderation models.
\emph{Multivariate Behavioral Research},
\emph{55}(1), 1--16.
\doi{10.1080/00273171.2019.1574546}
}
\seealso{
\code{\link[=print.model_set]{print.model_set()}}
}
\author{
Shu Fai Cheung \url{https://orcid.org/0000-0002-9871-9448}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelbpp-package.R
\docType{package}
\name{modelbpp-package}
\alias{modelbpp}
\alias{modelbpp-package}
\title{modelbpp: Model BIC Posterior Probability}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Fits the neighboring models of a fitted structural equation model and assesses the model uncertainty of the fitted model based on BIC posterior probabilities (BPP), using the method presented in Wu, Cheung, and Leung (2020) \doi{10.1080/00273171.2019.1574546}. See Pesigan, Cheung, Wu, Chang, and Leung (2026) \doi{10.3758/s13428-025-02921-x} for an introduction to the package.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://sfcheung.github.io/modelbpp/}
  \item Report bugs at \url{https://github.com/sfcheung/modelbpp/issues}
}

}
\author{
\strong{Maintainer}: Shu Fai Cheung \email{shufai.cheung@gmail.com} (\href{https://orcid.org/0000-0002-9871-9448}{ORCID})

Authors:
\itemize{
  \item Shu Fai Cheung \email{shufai.cheung@gmail.com} (\href{https://orcid.org/0000-0002-9871-9448}{ORCID})
  \item Huiping Wu
  \item Shing On Leung (\href{https://orcid.org/0000-0002-0962-339X}{ORCID})
}

Other contributors:
\itemize{
  \item Ivan Jacob Agaloos Pesigan \email{r.jeksterslab@gmail.com} (\href{https://orcid.org/0000-0003-4818-8420}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/measurement_invariance.R
\name{measurement_invariance_models}
\alias{measurement_invariance_models}
\title{Measurement Invariance Models}
\usage{
measurement_invariance_models(
  cfa_out,
  max_free = 1,
  metric = TRUE,
  scalar = TRUE,
  progress = TRUE
)
}
\arguments{
\item{cfa_out}{The output of
\code{\link[lavaan:cfa]{lavaan::cfa()}}.}

\item{max_free}{The maximum number of
constraints to be released when
generating the partial invariance
models. For example, if set to 1,
then only the partial metric invariance
model only has at most one item allowed
to have different loadings across
group. Default is 1. If set to zero,
then no partial invariance models
will be generated.}

\item{metric}{Logical. If \code{TRUE},
the default, then metric invariance
model and its partial invariance
versions are generated.}

\item{scalar}{Logical. If \code{TRUE},
the default, then scalar invariance
model and its partial invariance
versions are generated.}

\item{progress}{Logical. If \code{TRUE},
the default, progress bars will be
displayed when fitting partial
invariance models.}
}
\value{
A list of \code{\link[lavaan:cfa]{lavaan::cfa()}}
output. The names are automatically
generated to indicate whether a model
is configural, metric, or scalar
invariance, or the item(s) without
between-group constraints on the
loadings (for partial metric
invariance) or intercepts (for
partial scalar invariance).
}
\description{
Generate metric and
scalar invariance models and their
partial invariance versions.
}
\details{
This a helper function to generate,
based on a multigroup confirmatory
factor analysis (CFA) model with
no between-group equality constraints,
the following models:
\itemize{
\item A metric invariance model (loadings
constrained to be equal across
groups).
\item A scalar invariance model (intercepts
and loadings constrained to be equal
across groups).
\item Partial invariance versions of the
previous two models, such as a
model with the loadings of all items,
except for one, constrained to be
equal across groups.
}

The models generated can then be used
in \code{model_set()} to compute BPPs.
\subsection{Requirements}{

The model used as the input needs to
be fitted with no between group
constrains, that is, it is a
configural invariance model. Although
not a must, it is advised to use
the default way to identify each
factor (that is, fixing a loading
to one).
}

\subsection{Implementation}{

This function simply use the
\code{group.partial} and \code{group.equal}
argument of \code{\link[lavaan:cfa]{lavaan::cfa()}} to
generate the models.
}
}
\examples{

library(lavaan)
# For illustration, only one factor is used,
# with one item from another factor added
# just to make the model not saturated.
HSmod <-
"
spatial =~ x1 + x2 + x3 + x4
"
fit_config <- cfa(model = HSmod,
                  data = HolzingerSwineford1939,
                  group = "school")
fit_mi <- measurement_invariance_models(fit_config)
names(fit_mi)
# Need to add 'skip_check_sem_out = TRUE' to use multigroup models.
out <- model_set(sem_out = fit_mi,
                 skip_check_sem_out = TRUE,
                 progress = FALSE,
                 parallel = FALSE)
print(out)

}
\seealso{
\code{\link[=model_set]{model_set()}}
}
\author{
Shu Fai Cheung \url{https://orcid.org/0000-0002-9871-9448}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_add.R
\name{get_add}
\alias{get_add}
\title{Models That Are Less Restricted}
\usage{
get_add(
  sem_out,
  must_add = NULL,
  must_not_add = NULL,
  remove_constraints = TRUE,
  exclude_error_cov = TRUE,
  exclude_feedback = FALSE,
  exclude_xy_cov = FALSE,
  exclude_x_changed_to_y = TRUE,
  cross_add = c("pure_x", "pure_y"),
  cross_sets = NULL,
  df_change = 1,
  model_id = NA,
  keep_correct_df_change = TRUE,
  remove_duplicated = TRUE,
  progress = FALSE
)
}
\arguments{
\item{sem_out}{The original model,
which is the output from an
structural equation modeling
function. Currently support
\link[lavaan:lavaan-class]{lavaan::lavaan} objects only.}

\item{must_add}{A character vector
of parameters, named in
\code{\link[lavaan:lavaan]{lavaan::lavaan()}} style (e.g.,
\code{"y ~ x"}), that must be added.
Default is \code{NULL}.}

\item{must_not_add}{A character
vector of parameters, named in
\code{\link[lavaan:lavaan]{lavaan::lavaan()}} style (e.g.,
\code{"x1 ~~ x1"}), that must not be
added. Default is \code{NULL}.}

\item{remove_constraints}{Whether
equality constraints will be
removed. Default is \code{TRUE}.}

\item{exclude_error_cov}{Exclude
error covariances of indicators.
Default is \code{TRUE}.}

\item{exclude_feedback}{Exclude
paths that will result in a feedback
loop. For example, if there is
path from \code{x} through \code{m} to \code{y},
then the path \code{x ~ y} will create
a feedback loop. Default is \code{FALSE}
for now, to maintain backward
compatibility.
Do not rely on the default value
because it will be changed to \code{TRUE}
in a future major version.}

\item{exclude_xy_cov}{Exclude
covariance between two variables,
in which one has a path to another.
For example, if there is
path from \code{x} through \code{m} to \code{y},
then the covariance \code{x ~~ y},
which denotes the covariance between
\code{x} and the error term of \code{y}, will
be excluded if this argument is
\code{TRUE}. Default is \code{FALSE} for now,
to maintain backward compatibility.
Do not rely on the default value
because it will be changed to \code{TRUE}
in a future major version.}

\item{exclude_x_changed_to_y}{Exclude
changes that make a "pure" \code{x}
variable to a \code{y} variable. For example,
adding the path \code{x2 ~ x1} to
the model \code{y ~ x1 + x2}.}

\item{cross_add}{A character vector
of whether and how cross-loadings
(an indicator loads on two or more
latent factors) will be added. If
\code{NULL}, no cross-loadings will be
added. If \code{"lav_x"} is in the vector,
then cross-loadings among "pure-x"
latent factors, latent
factors that do not regress on
any other variables, will be considered.
If \code{"lav_y"} is in the vector, then
cross-loadings among "pure-y" latent
factors, latent factors that regress
on at least one variables but they
themselves do not predict any other
variables, will be considered.
If \code{"user"} is in the vector, then
cross-loadings among latent factors listed
in \code{cross_sets} will be considered.}

\item{cross_sets}{A character vector
of latent variables for which cross-loadings
will be considered. Ignored if
\code{"user"} is not in \code{cross_add}.
Note that cross-loadings involving latent
variables which are mediators will
not be considered. To add them,
they must be specified explicitly
in \code{must_add}.}

\item{df_change}{How many degrees
of freedom (\emph{df}) away in the list.
All models with \emph{df} change less than
or equal to this number will be
included, taking into account
requirements set by other arguments.
Default is 1.}

\item{model_id}{The identification
number of the starting model.
Default is \code{NA}, no identification
number.}

\item{keep_correct_df_change}{Keep
only models with actual \emph{df} change
equal to expected \emph{df} change.}

\item{remove_duplicated}{If \code{TRUE},
the default, duplicated models are
removed.}

\item{progress}{Whether a progress
bar will be displayed, implemented
by the \code{pbapply} package. Default
is \code{FALSE}.}
}
\value{
An object of the class
\code{partables}, a named list of parameter
tables, each of them to be used by
\code{\link[lavaan:lavaan]{lavaan::lavaan()}} or \code{\link[=update]{update()}}
for fitting a model with the added
parameters.
}
\description{
Generate a list of
models with one or more fixed
parameter freed.
}
\details{
It generates a list of models
with one or more fixed parameter
freed (and the degrees of freedom, \emph{df},
increases by one or more). If a
model has equality constraints, models
with one or more of the constraints
between two free parameters
released will also be included.

Graphically, paths or covariances
are "added" to form the list of models.

The models to be included are
identified by
\code{\link[lavaan:modificationIndices]{lavaan::modificationIndices()}}.

The models will be checked by \code{lavaan}
to make sure that the decrease in
model degrees of freedom is of the
expected value.

This function is called by
\code{\link[=model_set]{model_set()}} and usually users do
not need to call it. It is exported
for advanced users.
}
\examples{

library(lavaan)
dat <- dat_path_model
mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"
fit <- sem(mod, dat_path_model, fixed.x = TRUE)
mod_to_add <- get_add(fit)
mod_to_add

}
\seealso{
\code{\link[=print.partables]{print.partables()}}
}
\author{
Shu Fai Cheung \url{https://orcid.org/0000-0002-9871-9448}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_drop.R
\name{get_drop}
\alias{get_drop}
\title{Models That Are More Restricted}
\usage{
get_drop(
  sem_out,
  must_drop = NULL,
  must_not_drop = NULL,
  loadings_to_exclude = c("single", "none", "all"),
  df_change = 1,
  model_id = NA,
  keep_correct_df_change = TRUE,
  remove_duplicated = TRUE,
  progress = FALSE
)
}
\arguments{
\item{sem_out}{The original model,
which is the output from an
structural equation modeling
function. Currently support
\link[lavaan:lavaan-class]{lavaan::lavaan} objects only.}

\item{must_drop}{A character vector
of parameters, named in
\code{lavaan::lavaan()} style (e.g.,
\code{"y ~ x"}), that must be included.
Default is \code{NULL}.}

\item{must_not_drop}{A character
vector of parameters, named in
\code{\link[lavaan:lavaan]{lavaan::lavaan()}} style (e.g.,
\code{"x1 ~~ x1"}), that must not be
included. Default is \code{NULL}.}

\item{loadings_to_exclude}{Whether factor loadings
will be excluded. If \code{"single"},
then only "single" loadings (an indicator
loads on only one latent factor)
will be excluded. If \code{"all"}, then
all factor loadings will be excluded.
If \code{"none"}, then no loadings will
be excluded. Be careful when using
\code{"none"} because the models may not
make sense. The settings in \code{must_drop}
and \code{must_not_drop} will override this
argument.}

\item{df_change}{How many degrees
of freedom away in the list. All
models with \emph{df} change less than
or equal to this number will be
included, taking into account
requirements set by other arguments.
Default is 1.}

\item{model_id}{The identification
number of the starting model.
Default is \code{NA}, no identification
number.}

\item{keep_correct_df_change}{Keep
only models with actual \emph{df} change
equal to expected \emph{df} change.}

\item{remove_duplicated}{If \code{TRUE},
the default, duplicated models are
removed.}

\item{progress}{Whether a progress
bar will be displayed, implemented
by the \code{pbapply} package. Default
is \code{FALSE}.}
}
\value{
An object of the class
\code{partables}, a named list of parameter
tables, each of them to be used by
\code{\link[lavaan:lavaan]{lavaan::lavaan()}} or \code{\link[=update]{update()}}
for fitting a model with the added
parameters.
}
\description{
Generate a list of
models with one or more free parameter
dropped (fixed to zero).
}
\details{
It generates a list of models
with one or more free parameters
dropped, that is, fixed to zero
(with degrees of freedom,
\emph{df}, increases by one or more).

All free parameters are included in
the pool of candidates, except for
those explicitly requested to be
kept.

The models will be checked by \code{lavaan}
to make sure that the increase in
model degrees of freedom is of the
expected value.

This function is called by
\code{\link[=model_set]{model_set()}} and usually users do
not need to call it. It is exported
for advanced users.
}
\examples{

library(lavaan)

dat <- dat_path_model
mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1 + x2
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)
mod_to_drop <- get_drop(fit)
mod_to_drop

}
\seealso{
\code{\link[=print.partables]{print.partables()}}
}
\author{
Shu Fai Cheung \url{https://orcid.org/0000-0002-9871-9448}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/min_bpp.R
\name{min_prior}
\alias{min_prior}
\title{Minimum Prior}
\usage{
min_prior(bic, bpp_target, target_name = "original")
}
\arguments{
\item{bic}{A named vector of BIC values
for a set of models.}

\item{bpp_target}{A value from zero
to 1. The desired BIC posterior
probability.}

\item{target_name}{The name of the
original model, as appeared in the
names of \code{bic}.}
}
\value{
A scalar. The required prior
probability.
}
\description{
Find the minimum prior
probability required to achieve
the desired BIC posterior
probability.
}
\details{
It assumes that all models other than
the original model have the same prior
probabilities.

This function is called by
\code{\link[=model_set]{model_set()}} or \code{\link[=print.model_set]{print.model_set()}}
and usually users do
not need to call it. It is exported
for advanced users.
}
\examples{

library(lavaan)

dat <- dat_path_model

mod <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fit <- sem(mod, dat_path_model, fixed.x = TRUE)

out <- model_set(fit)
min_prior(out$bic, bpp_target = .8)

}
\references{
Wu, H., Cheung, S. F., & Leung, S. O.
(2020). Simple use of BIC to assess
model selection uncertainty: An
illustration using mediation and
moderation models.
\emph{Multivariate Behavioral Research},
\emph{55}(1), 1--16.
\doi{10.1080/00273171.2019.1574546}
}
\seealso{
\code{\link[=model_set]{model_set()}} and
\code{\link[=print.model_set]{print.model_set()}}
}
\author{
Shu Fai Cheung \url{https://orcid.org/0000-0002-9871-9448}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/c_partables.R
\name{c.partables}
\alias{c.partables}
\alias{c.model_set}
\alias{partables_drop}
\alias{manipulate_partables}
\title{Manipulate Parameter Tables}
\usage{
\method{c}{partables}(...)

\method{c}{model_set}(...)

partables_drop(x, model_names = NULL)
}
\arguments{
\item{...}{An arbitrary number of
objects. All invalid objects (see
details) will be discarded. If an
object is named and is not
\code{partables} object, its name will be
used.}

\item{x}{A \code{partables}-class object.}

\item{model_names}{A character vector
of the names of models in a
\code{partables}-class object.}
}
\value{
A \code{partables}-class objects with
all the objects supplied combined
together. If an object is
a \code{lavaan}-class object, its
parameter table will be retrieved
by \code{\link[lavaan:parameterTable]{lavaan::parameterTable()}}.
If an object is a \code{model_set}-class
object, the stored \code{partables}-class
object will be retrieved.
}
\description{
Functions to manipulate
a \code{partables}-class object
}
\details{
The \code{partables}-class objects have
a \code{c()} method that can be used to
combine parameter tables.
Each object must be

a. a \code{partables}-class object,
b. a \code{model_set}-class object,
c. a \code{lavaan}-class object, or
d. a parameter table of
the class \code{lavaan.data.frame()},
usually generated by
\code{\link[lavaan:parameterTable]{lavaan::parameterTable()}}.
Other objects will be discarded.

Names will be used when combining
objects. If two objects have the same
names, then only the first one will be
retained. No warning message will be
issued. Users are encouraged to
explicitly name all objects carefully.

Note that, to invoke this method,
the first object must be a
\code{partables} object.

The \code{model_set} class also has a
\code{c}-method. It
will replace the first object by the stored
partables and then call the \code{c}-method of \code{partables} objects.

The function \code{\link[=partables_drop]{partables_drop()}} is
for dropping models from a
\code{partables}-class object.
}
\examples{

library(lavaan)

moda <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x1
ab := a*b
"

fita <- sem(moda, dat_path_model, fixed.x = TRUE)

outa <- model_set(fita,
                  progress = FALSE,
                  parallel = FALSE)

modb <-
"
x3 ~ a*x1 + b*x2
x4 ~ a*x2
ab := a*b
"

fitb <- sem(modb, dat_path_model, fixed.x = TRUE)

outb <- model_set(fitb,
                  progress = FALSE,
                  parallel = FALSE)

mod2 <-
"
x2 ~ 0*x3 + 0*x4
x1 ~ 0*x3
"
fit2 <- sem(mod2, dat_path_model)

mod3 <-
"
x2 ~ x3 + 0*x4
x1 ~ x3
"
fit3 <- sem(mod3, dat_path_model)

out <- c(outa$models, user2 = fit2, outb$models, user3 = fit3)
out

out2 <- c(outa, user2 = fit2, outb$models, user3 = fit3)
out2

out3 <- c(outa, user2 = fit2, outb, user3 = fit3)
out3

}
\author{
Shu Fai Cheung \url{https://orcid.org/0000-0002-9871-9448}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dat_path_model}
\alias{dat_path_model}
\title{A Sample Dataset Based on a
Path Model (For Testing)}
\format{
A data frame with four variables:
\describe{
\item{x1}{Predictor}
\item{x2}{Predictor}
\item{x3}{Mediator}
\item{x4}{Outcome}
}
}
\usage{
dat_path_model
}
\description{
Generated from the
a path model (n = 100).
}
\details{
The model used to generate
this dataset:

\if{html}{\out{<div class="sourceCode">}}\preformatted{x1 ~~ x2
x3 ~ x1 + x2
x4 ~ x3 + x1 + x2
}\if{html}{\out{</div>}}
}
\keyword{datasets}
